/*
 *	grail.h
 *
 *	GRAphical Interface Layer for building graphical interface elements. The
 *	name has been suggested by Erik BRUCHEZ and Daniel MARMIER; "Graal" had
 *	been first used, which is the french translation of "Grail".
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_GRAIL_H_
#define	_OPaC_GRAIL_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (grail.h);

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

#ifndef _OPaC_WINSTYLE_H_
#include <OPaC/winstyle.h>
#endif

#ifndef _OPaC_TEXTSTYLE_H_
#include <OPaC/textstyle.h>
#endif

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

#ifndef _OPaC_STRING_H_
#include <OPaC/string.h>
#endif

/****************************************************************/

/*
 *	Internal, private, machine dependent structures :
 */

struct OPaC_GrPort;								//	private display context
struct OPaC_GrPixmap;							//	private pixmap description
struct OPaC_GrWindow;							//	private window description
struct OPaC_GrFont;								//	private font description
struct OPaC_GrCursor;							//	private cursor definition
struct OPaC_ColorMan;							//	private color definition
struct OPaC_RemoteAppDesc;						//	private remote app definition

/****************************************************************/

/*
 *	Enumerate the available Grail Line Styles. These are
 *	the standard line styles needed by every graphical
 *	interface.
 *
 *	There exact representation may vary from system to
 *	system.
 */

OP_TYPEDEF (Card32, OPLineStyle);

#define	OP_LINE_SOLID		0					//	________
#define	OP_LINE_DOT			1					//	........
#define	OP_LINE_DASH		2					//	__ __ __

/****************************************************************/

/*
 *	Enumerate possible color models. This will be used by
 *	the widgets to decide how to draw themselves.
 */

OP_TYPEDEF (Card32, OPColorModel);
OP_TYPEDEF (Card32, OPGrayModel);

#define	OP_COLOR_MODEL_NONE		0				//	just black and white
#define	OP_COLOR_MODEL_POOR		1				//	few colors (less than 5 x 5 x 5)
#define	OP_COLOR_MODEL_MEDIUM	2				//	at least a 5 x 5 x 5 color cube
#define	OP_COLOR_MODEL_GOOD		3				//	12-bit and up (true color)

#define	OP_GRAY_MODEL_MONO		0				//	just black and white
#define	OP_GRAY_MODEL_POOR		1				//	about 2 extra gray shades
#define	OP_GRAY_MODEL_MEDIUM	2				//	at least 15 gray shades
#define	OP_GRAY_MODEL_GOOD		3				//	100 and more gray shades

/****************************************************************/

/*
 *	The OPaC Color Manager provides an interface to the
 *	color model used by the host. It supports both true
 *	and indexed color models.
 */

class OP_PUBLIC OPColorMan : public OPObject
{
	
protected:
	
	//	Each color in the palette (if any) is defined by its pixel value and RGB
	//	components. If it can be used for color matching, the match flag will be
	//	set to TRUE.
	
	struct PaletteEntry
	{
		OPColorPixel	pixel;					//	pixel value
		Bool			match_enable;			//	TRUE => may be used for matching
		Bool			gray;					//	TRUE => pure gray color
		Card8			r, g, b;				//	red, green and blue components
	};
	
	
	//	The system and the application may define some extra colors for specific
	//	usage (eg. hiliting a zone or coloring the cursor). These cannot be used
	//	with the standard color matching methods. They must be accessed through
	//	their name :
	
	struct NamedEntry
	{
		NamedEntry*		next;					//	pointer to next named color entry
		
		OPColorPixel	pixel;					//	color pixel value
		Count			index;					//	color index into palette (if any)
		OPAtomRef		name;					//	color name
		
		Card8			r, g, b;				//	red, green and blue components
	};
	
private:
	static OPColorManRef system_man;			//	system-wide default color manager
	
protected:
	OPaC_ColorMan*		info;					//	machine dependent information
	
	Count				depth;					//	depth in bits/pixel
	Count				pal_size;				//	palette size (0 if no palette available)
	Count				pal_gray;				//	palette gray elements
	Count				pal_cube;				//	palette color cube entries
	Count				pal_named;				//	palette named entries
	PaletteEntry*		pal_entry;				//	palette entries
	
	Count				cube_pos;				//	position of first color of cube in palette
	
	Count				named_num;				//	number of named colors
	Count				named_pos;				//	position of first named color in palette
	NamedEntry*			named_entry;			//	list of named entries
	
	OPColorModel		color_model;			//	color model
	OPGrayModel			gray_model;				//	gray model
	
	OPColorPixel		gray_mapping[256];		//	maps the 256 gray levels to some pixels
	
protected:
	OPColorMan ();
	virtual ~ OPColorMan ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
	virtual void ComputeNearestColor ();
	virtual OPColorPixel MakeTrueColor (Card8 r, Card8 g, Card8 b) const;
	
public:
	
	//	Color matching for standard drawing purposes and color matching for
	//	specific colors :
	
	virtual OPColorPixel FindColor (Real r, Real g, Real b) const;
	virtual OPColorPixel FindColor (Card8 r, Card8 g, Card8 b) const;
	virtual OPColorPixel FindColor (Real i) const;
	virtual OPColorPixel FindColor (Card8 i) const;
	virtual Bool FindColor (OPColorPixel p, RealOut r, RealOut g, RealOut b) const;
	virtual Bool FindColor (OPColorPixel p, Card8Out r, Card8Out g, Card8Out b) const;
	
	virtual Bool FindColor (OPAtomIn name, OPColorPixelOut pixel) const;
	virtual const OPColorPixel* ReturnGrayMapping () const;
	
	//	Named colors may be (re)defined by the application in a limited fashion.
	//	Colors with names beginning with "System:" are defined system-wide and
	//	shared among every application.
	
	virtual Bool DefineColor (OPAtomIn name, Real r, Real g, Real b);
	virtual Bool DefineColor (OPAtomIn name, Card8 r, Card8 g, Card8 b);
	
	//	It may be useful to know how many colors are available and if a "true"
	//	color model is in use :
	
	virtual Count CountColors () const;
	virtual Bool  IsTrueColor () const;
	
	virtual OPColorModel ReturnColorModel () const;
	virtual OPGrayModel  ReturnGrayModel () const;
	
	//	The Grail classes may need to access the private, machine
	//	dependent, information :
	
	OPaC_ColorMan* ReturnPrivateInfo () const { return this->info; }
	
	//	Usually, every window shares the system wide OPaC color manager. This
	//	contains the most used color samples.
	
	static const OPColorManRef& ReturnSystemColorMan ();
	static void InitSystemColorMan (OPColorManIn);
	static void KillSystemColorMan ();
	
	OP_DEFINE_CLASS (OPColorMan, OPObject);
};

/****************************************************************/

/*
 *	The OPaC Font Manager provides an interface to the
 *	underlying system dependent font database.
 */

class OP_PUBLIC OPFontMan : public OPObject
{
	
private:
	static OPFontManRef	system_man;				//	system-wide default font manager
	
protected:
	OPDataBaseRef		font_families;			//	font families database
	OPEntityModelRef	font_fam_model;			//	font family entity model
	OPEntityModelRef	font_def_model;			//	font definition entity model
	
protected:
	OPFontMan ();
	virtual ~ OPFontMan ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
public:
	virtual void GetFontFamiliesDB (OPDataBaseOut) const;
	
	static const OPFontManRef& ReturnSystemFontMan ();
	static void InitSystemFontMan (OPFontManIn);
	static void KillSystemFontMan ();
	
	OP_DEFINE_CLASS (OPFontMan, OPObject);
};

/****************************************************************/

/*
 *	The OPaC Graphic Abstraction Layer Font provides an
 *	interface to the fonts used by the underlying font
 *	machinery.
 */

class OPGrailFont : public OPObject
{
	
private:
	OPaC_GrFont*		info;					//	associated machine dependent info
	
	OPStringRef			name;					//	font name (family name only)
	OPStringRef			style;					//	font style (style name only)
	OPCoord				size;					//	font size in points
	
protected:
	OPGrailFont ();
	virtual ~ OPGrailFont ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
	virtual Bool RealizeFont ();
	
public:
	virtual Bool Initialise (OPStringIn name, OPStringIn style, OPCoord size);
	
	virtual void GetFontName (OPStringOut) const;
	virtual void GetFontStyle (OPStringOut) const;
	virtual OPCoord ReturnFontSize () const;
	
	virtual OPRelCoord ReturnAscender () const;
	virtual OPRelCoord ReturnDescender () const;
	virtual OPCoord ReturnWidth (const LatinChar* text) const;
	virtual OPCoord ReturnWidth (OPStringIn text) const;
	virtual OPCoord ReturnWidth (OPStringIn text, Count start, Count length) const;
	
	virtual OPaC_GrFont* ReturnPrivateInfo () const { return this->info; }
	
	OP_DEFINE_CLASS (OPGrailFont, OPObject);
};

/****************************************************************/

/*
 *	The Grail Port class is the base class used to interface
 *	any graphical operation.
 *	
 *	It can be compared with a GRAFPORT (Apple) or a GDI HDC,
 *	depending on your beliefs :-).
 *	
 *	The Grail Port is used as the base class for graphical
 *	device drivers, such as GrailWindow and GrailPixmap.
 */

class OP_PUBLIC OPGrailPort : public OPObject
{
	
protected:
	OPaC_GrPort*	info;						//	associated machine dependent info
	
	OPAppRef		application;				//	"this" is owned by "application"
	
	OPColorManRef	color_man;					//	associated color manager
	OPFontManRef	font_man;					//	associated font manager
	OPAdornerRef	adorner;					//	associated adorner
	
	OPCoord			rect_x1, rect_y1;			//	drawing rectangle top left corner
	OPCoord			rect_x2, rect_y2;			//	drawing rectangle bottom right corner
	
	OPCoord			clip_x1, clip_y1;			//	clipping rectangle top left corner
	OPCoord			clip_x2, clip_y2;			//	clipping rectangle bottom right corner
	
	OPLineStyle		line_style;					//	current line style
	OPColorPixel	fore_color;					//	foreground color
	OPColorPixel	back_color;					//	background color
	OPGrailFontRef	font;						//	current font
	
	Bool			is_ready;					//	TRUE => context initialised
	Bool			is_drawing;					//	TRUE => draw in progress
	
	Bool			use_transparency;			//	TRUE => use transparency
	OPColorPixel	transparent_color;			//	transparent color (if any)
	
protected:
	OPGrailPort ();
	virtual ~ OPGrailPort ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	virtual void InitApp (OPAppIn app);
	virtual void InitAdorner (OPAdornerIn adorner);
	
	
	//	The drawing should always be enclosed between a "Begin Draw" and an "End
	//	Draw" invocation. This will set up a local origin at [x1;y1] and clip
	//	everything to the specified rectangle. The rectangle will be marked as
	//	dirty and refreshed ASAP.
	
	virtual void BeginDraw (OPRectIn box, OPRelCoord margin = 0);
	virtual void BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	virtual void EndDraw ();
	virtual void Flush ();
	
	
	//	Clipping functions used to define the clipping rectangle. No other, more
	//	complex, area can be defined using Grail.
	
	virtual void SetClipRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void SetClipRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void ResetClipRect ();
	
	virtual void GetClipRect (OPRectOut box) const;
	virtual void GetClipRect (OPCoordOut x1, OPCoordOut y1,
							  OPCoordOut x2, OPCoordOut y2) const;
	
	
	//	The fundamental graphical operations allow to draw horizontal and vertical
	//	lines very quickly. They also handle rectangles. Support for circles and
	//	more complex graphical shapes might be added in the future, as soon as
	//	the need arises :
	
	virtual void DrawDot (OPCoord x, OPCoord y);
	virtual void DrawDot (OPCoord x, OPCoord y, OPColorPixel color);
	
	virtual void DrawLineDX (OPCoord x, OPCoord y, OPCoord dx);
	virtual void DrawLineDY (OPCoord x, OPCoord y, OPCoord dy);
	virtual void DrawLineToX (OPCoord x1, OPCoord y1, OPCoord x2);
	virtual void DrawLineToY (OPCoord x1, OPCoord y1, OPCoord y2);
	
	virtual void DrawRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void DrawRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void FillRect (OPRectIn box, OPRelCoord margin = 0);
	virtual void FillRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	//	The hair lines, like the horizontal and vertical ones, include the end
	//	points. They are only used to draw style-less lines.
	
	virtual void DrawHairLine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy);
	virtual void DrawHairLineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	virtual void DrawAALine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy);
	virtual void DrawAALineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2);
	
	//	It is possible to draw some simple horizontal text for the user interface.
	//	If you need more complex methods, look for ODD :
	
	virtual void ShowText (OPCoord x, OPCoord y,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0,
						   OPCoord dx = 0);
	
	virtual void ShowText (OPRectIn box,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0,
						   OPRelCoord margin = 0);
	
	virtual void ShowText (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
						   OPStringIn text,
						   Count start = 0, Count len = 0, Card8 pos = 0);
	
	
	//	Raster operations are also very useful for the user interface. Here come
	//	the related methods :
	
	virtual void ScrollRect (OPRectIn box, OPRelCoord move_x, OPRelCoord move_y,
							 OPRelCoord margin = 0);
	
	virtual void ScrollRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
							 OPRelCoord move_x, OPRelCoord move_y);
	
	virtual void CopyRect (OPGrailPortIn source,
						   OPCoord source_x, OPCoord source_y,
						   OPRectIn dest_box, OPRelCoord dest_margin = 0);
	
	virtual void CopyRect (OPGrailPortIn source,
						   OPCoord source_x, OPCoord source_y,
						   OPCoord dest_x1, OPCoord dest_y1,
						   OPCoord dest_x2, OPCoord dest_y2);
	
	virtual void SetBitmap (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy,
							const Card8* bits);
	
	virtual	void SetGrayPixels (OPCoord x, OPCoord y, OPCoord dx,
								const Card8* intensity);
	
	virtual void SetColorPixels (OPCoord x, OPCoord y, OPCoord dx,
								 const OPColorPixel* pixels);
	
	virtual	void SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx,
							   const Card8* red,
							   const Card8* green,
							   const Card8* blue);
	
	//	Graphical context manipulation :
	
	virtual void SetForeColor (OPColorPixel fore_color);
	virtual void SetBackColor (OPColorPixel back_color);
	virtual void SetLineStyle (OPLineStyle line_style);
	virtual void SetFont (OPGrailFontIn font);
	
	virtual void SetTransparency (OPColorPixel transparent_color);
	virtual void SetTransparent (Bool ok = TRUE);
	
	virtual const OPColorManRef& ReturnColorMan () const;
	virtual const OPFontManRef& ReturnFontMan () const;
	virtual const OPAdornerRef& ReturnAdorner ();
	
	//	Compute the pixel co-ordinates of a line (kind of Bresenham algorithm
	//	with optional error information which can be used for anti-aliasing).
	
	virtual void ComputeLineDXDots (Card32 x, Card32 y, Card32 dx, Int32 dy,
									Card32* y_array,
									Card8* y_error = 0,
									Real error = 0.0);
	
	virtual void ComputeLineDYDots (Card32 x, Card32 y, Int32 dx, Card32 dy,
									Card32* x_array,
									Card8* x_error = 0,
									Real error = 0.0);
	
	//	Internal use only :
	
	OPaC_GrPort* ReturnPrivatePortInfo () const	{ return this->info; }
	
	OP_DEFINE_CLASS (OPGrailPort, OPObject);
};

/****************************************************************/

/*
 *	The Grail Pixmap class provides an easy way to draw off-
 *	screen,	eg. in order to implement double buffering.
 *
 *	It is also possible to load a pixmap from a pixmap file.
 *
 *	NIY:	Handling the pixmap's depth !
 */

class OP_PUBLIC OPGrailPixmap : public OPGrailPort
{
	
protected:
	OPaC_GrPixmap*	pixmap;						//	associated machine dependent info
	OPCoord			dx, dy;						//	pixmap width
	Count			depth;						//	pixmap depth (1/2/4/8/16/24)
	Card8**			rgb_cache[3];				//	array of cached pixels
	Bool			is_cache_enabled;			//	TRUE => cache is enabled
	
protected:
	OPGrailPixmap ();
	virtual ~ OPGrailPixmap ();
	
	virtual void StoreInstance (OPStorageIn storage);			/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);			/* OVERRIDDEN */
	
	virtual Bool SystemGetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* red, Card8* green, Card8* blue);
	virtual Bool AccessCacheRGBLine (int line, Card8*& r, Card8*& g, Card8*& b);
	virtual void CacheRGBLine (int line, const Card8* r, const Card8* g, const Card8* b);
	virtual Bool FindRGBLine (int line, Card8* r, Card8* g, Card8* b);
	
public:
	virtual void BeginDraw (OPRectIn box, OPRelCoord = 0);		/* OVERRIDDEN */
	virtual void BeginDraw (OPCoord, OPCoord, OPCoord, OPCoord);/* OVERRIDDEN */
	virtual void EndDraw ();									/* OVERRIDDEN */
	
	virtual void Initialise (OPGrailPixmapIn window, OPCoord dx, OPCoord dy);
	virtual void Initialise (Count depth, OPCoord dx, OPCoord dy);
	
	void  GetSize (OPCoordOut dx, OPCoordOut dy) const	{ dx = this->dx; dy = this->dy; }
	Count ReturnDepth () const							{ return this->depth; }
	
	virtual Bool GetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* red, Card8* green, Card8* blue);
	virtual void EnableCache (Bool enable = TRUE);
	virtual void FlushCache ();
	
	OPaC_GrPixmap* ReturnPrivatePixmapInfo () const		{ return this->pixmap; }
	
	OP_DEFINE_CLASS (OPGrailPixmap, OPGrailPort);
};

/****************************************************************/

/*
 *	The Grail Window class provides a mean to create a new
 *	window and its associated pixmap buffer.
 */

class OP_PUBLIC OPGrailWindow : public OPGrailPixmap
{
	
	friend class OPGrailPixmap;					//	OPGrailPixmap may access this->window
	
protected:
	OPaC_GrWindow*	window;						//	machine dependent window info
	
	Bool			is_active;					//	TRUE => active window
	Bool			is_open;					//	TRUE => open window
	Bool			is_reduced;					//	TRUE => reduced window
	
	OPRect			bounding_box;				//	bounding box	
	OPWindowRef		view;						//	OPaC "view" of the window
	OPWindowStyle	style;						//	window style
	
protected:
	OPGrailWindow ();
	virtual ~ OPGrailWindow ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void BeginDraw (OPRectIn box, OPRelCoord = 0);		/* OVERRIDDEN */
	virtual void BeginDraw (OPCoord, OPCoord, OPCoord, OPCoord);/* OVERRIDDEN */
	virtual void EndDraw ();									/* OVERRIDDEN */
	virtual void Flush ();										/* OVERRIDDEN */
	
	
	//	Initialise a window with a fixed style. The name might not always be
	//	used, but it should always be available :
	
	virtual void Initialise (OPWindowIn view,
							 OPWindowStyle style,
							 OPStringIn name,
							 OPCoord x1, OPCoord y1,
							 OPCoord x2, OPCoord y2);
	
	virtual void Initialise (OPWindowIn view,
							 OPWindowStyle style,
							 OPStringIn name,
							 OPRectIn box, OPRelCoord margin = 0);
	
	virtual void ChangeName (OPStringIn name);
	
	virtual void Open ();
	virtual void Reduce ();
	virtual void Close ();
	virtual void ActivateAsFrontWindow (Bool front = TRUE);
	
	virtual void SysChangedActive (Bool);
	virtual void SysChangedOpen (Bool);
	virtual void SysChangedVisible (Bool);
	virtual void SysChangedReduced (Bool);
	
	virtual Bool IsActive () const;
	
	//	The window may be moved explicitely by the program. This should only be
	//	used to reflect dragging...
	
	virtual void SetOrigin (OPCoord x, OPCoord y);
	virtual void GetOrigin (OPCoordOut x, OPCoordOut y) const;
	virtual void SetSize (OPCoord x, OPCoord y);
	virtual void GetSize (OPCoordOut x, OPCoordOut y) const;
	
	virtual void UserChangedGeometry ();
	
	virtual void GetFrameSize (OPCoordOut xl, OPCoordOut yt,
							   OPCoordOut xr, OPCoordOut yb);
	
	static void GetFrameSize (OPWindowStyle style,
							  OPCoordOut xl, OPCoordOut yt,
							  OPCoordOut xr, OPCoordOut yb);
	
	//	Called by the window manager :
	
	virtual void Redraw ();
	virtual void RedrawIcon ();
	
	//	Interrogation methods :
	
	void GetWindow (OPWindowOut w) const		 { w = this->view; }
	OPaC_GrWindow* ReturnPrivateWindowInfo () const { return this->window; }

	static Bool FindOwnWindow (OPCoord, OPCoord, OPGrailWindowOut wdo, Count index = 0);
	static Bool FindOPaCWindow (OPCoord, OPCoord, OPaC_RemoteAppDesc* desc, Count index = 0);
	
	OP_DEFINE_CLASS (OPGrailWindow, OPGrailPixmap);
};

/****************************************************************/

/*
 *	The Grail Cursor can be used to change the graphic
 *	cursor's shape.
 */

class OP_PUBLIC OPGrailCursor : public OPObject
{
	
protected:
	OPaC_GrCursor*		cursor;					//	machine dependent cursor info
	
	OPGrailPixmapRef	foreground;				//	foreground for cursor
	OPGrailPixmapRef	background;				//	background for cursor
	
	Int16				x_foreground;			//	[x] position in foreground pixmap
	Int16				y_foreground;			//	[y] position in foreground pixmap
	
	Int16				x_background;			//	[x] position in background pixmap
	Int16				y_background;			//	[y] position in background pixmap
	
	Int16				dx, dy;					//	size of cursor in pixmaps
	Int16				hot_x, hot_y;			//	hotspot position relative to cursor
	
protected:
	OPGrailCursor ();
	virtual ~ OPGrailCursor ();
	
	virtual void StoreInstance (OPStorageIn storage);
	virtual void RestoreInstance (OPStorageIn storage);
	
public:
	virtual void Initialise (OPCoord hotx, OPCoord hoty,
							 OPCoord dx, OPCoord dy,
							 OPGrailPixmapIn fg, OPCoord xf, OPCoord yf,
							 OPGrailPixmapIn bg, OPCoord xb, OPCoord yb);
	
	virtual void Activate () const;
	virtual void Hide () const;
	virtual void ResetToNormal () const;
	
	OP_DEFINE_CLASS (OPGrailCursor, OPObject);
};

/****************************************************************/

extern OPGrailPixmapRef OP_PUBLIC OPaC_LoadGIF (OPStringIn name, OPMemZone* zone);

#endif /* _OPaC_GRAIL_H_ */
