/*
 *	gridlist.h
 *
 *	
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_SCRLIST_H_
#define	_OPaC_SCRLIST_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (scrlist.h);

/****************************************************************/

#ifndef _OPaC_CELLGRID_H_
#include <OPaC/cellgrid.h>
#endif

#ifndef _OPaC_CELL_H_
#include <OPaC/cell.h>
#endif

/****************************************************************/

/*
 *	The scroll list may get its data through some callback
 *	functions, when the container configuration does not
 *	provide enough flexibility. Following callbacks are
 *	supported :
 *	
 *	  -	Text callback, fills in a string with the text to
 *		display in the specified cell.
 *	
 *	  - View callback, fills in a view which will be added
 *		to the specified cell (several views/strings items
 *		are possible for each cell).
 *	
 *	  -	Cell callback, fills in a cell with subviews and
 *		other elements.
 *	
 *	The callbacks should return TRUE if the data could be
 *	properly defined.
 */

typedef Bool (*OPListTextCallBack)(OPScrollList* list, OPObjectIn arg, Count line, Count col, OPStringInOut text);
typedef Bool (*OPListViewCallBack)(OPScrollList* list, OPObjectIn arg, Count line, Count col, OPViewInOut view);
typedef Bool (*OPListCellCallBack)(OPScrollList* list, OPObjectIn arg, Count line, OPCellInOut cell);

/****************************************************************/

/*
 *	The scroll list is nothing but a cell grid which changes
 *	the cell's contents when some scrolling is performed.
 */

class OP_PUBLIC OPScrollList : public OPCellGrid
{
	
	friend Bool ListTextCallBack (void*, Count, OPStringInOut);
	friend Bool ListViewCallBack (void*, Count, OPViewInOut);
	friend Bool ListCellCallBack (void*, Count, OPCellInOut);
	
private:
	struct ListDataDef;
	
protected:
	OPScrollerRef		scroller;						//	(vertical) scroller
	OPCoord				margin;							//	margin around the border
	
	OPCellAlignment		def_alignment[OP_CELL_CTRL];	//	default cell alignment
	OPCoord				def_left_space[OP_CELL_CTRL];	//	default left space to cell
	OPCoord				def_right_space[OP_CELL_CTRL];	//	default right space to cell
	
	Count				num_top;						//	current # of topmost cell
	Count				num_total;						//	current # of cells (total count)
	
	ListDataDef*		data[OP_CELL_CTRL];				//	data definition for scroll list
	Bool*				active_hilite;					//	actively hilited cells
	
	Count				refresh_top;					//	first cell to refresh
	Count				refresh_num;					//	# of cells to refresh
	
	Count				select_num;						//	# of selected cells
	
	Bool				select_on_click;				//	TRUE => clicking selects elements
	Bool				select_multiple_ok;				//	TRUE => allows multiple selections
	Bool				select_at_least_one;			//	TRUE => at least one selection

	OPArrayRef			contents;						//	data associated to list contents
	
protected:
	OPScrollList ();
	virtual ~ OPScrollList ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void ChangeGeometry (OPRectIn);				/* OVERRIDDEN */
	virtual void UpdateScrollerPos (OPRectIn);
	
	virtual void VisibleRatioChanged ();
	virtual void ScrollCells (Int32 amount);
	virtual void ScrollRange (Card32 top, Card32 num, Int32 amount);
	
	virtual void MarkRefreshAll ();
	virtual void MarkRefreshRange (Card32 top, Card32 num);
	virtual void RefreshCells ();

	virtual void NewDataDef (Count index);
	virtual void DeleteDataDef (Count index);
	virtual void DeleteAllDataDef ();

public:
	virtual void Delete ();								/* OVERRIDDEN */
	
	//	The cells in the scroll list can contain up to OP_CELL_CTRL items
	//	horizontally: the position of the items is specified by 'index'.
	
	virtual Bool SetAlignment (Count index, OPCellAlignment);
	virtual Bool SetLeftSpace (Count index, OPCoord);
	virtual Bool SetRightSpace (Count index, OPCoord);
	
	virtual Bool SetContainer (Count index, OPGenericContainerIn, OPAtomIn query, OPStringIn fmt = OPStringNull);
	virtual Bool SetQueryList (Count index, OPObjectIn root, OPGenericContainerIn queries, OPStringIn fmt = OPStringNull);
	virtual Bool SetExtraction (Count index, OPArrayIn extraction);
	
	virtual Bool SetCallbackText (Count index, OPListTextCallBack callback, OPObjectIn arg);
	virtual Bool SetCallbackView (Count index, OPListViewCallBack callback, OPObjectIn arg);
	virtual Bool SetCallbackCell (OPListCellCallBack callback, OPObjectIn arg);
	
	//	The scroll list provides a simple way to store informations
	//	which can be used by the callbacks :
	
	virtual Bool SetContents (Count line, OPObjectIn data);
	virtual Bool GetContents (Count line, OPObjectOut data) const;
	
	//	Methods used to specify which cells should be displayed by the
	//	scroll list :
	
	virtual Bool SetScroller (OPScrollerIn scroller);
	virtual Bool SetTotalCount (Count max);
	
	virtual void SetActiveHilite (Count index, Bool yes);
	virtual Bool ReturnActiveHilite (Count index) const;
	
	virtual void SetSelectOnClick (Bool yes);
	virtual void SetSelectMultipleOk (Bool yes);
	virtual void SetSelectAtLeastOne (Bool yes);
	virtual Bool ReturnSelectOnClick () const;
	virtual Bool ReturnSelectMultipleOk () const;
	virtual Bool ReturnSelectAtLeastOne () const;
	
	virtual Bool FindActiveHilite (CountOut index) const;
	virtual Bool FindActiveHilite (CountOut index, Count nth) const;
	
	virtual Bool SetVisiblePos (Real relative);
	virtual Real ReturnVisiblePos () const;
	virtual Real ReturnVisibleRatio () const;
	
	virtual Bool CenterCell (Count pos, Real relative);
	
	//	Observers exported for run-time binding :
	
	virtual void HandleScroll (OPObjectIn object, Card32 id);
	virtual void HandleCellClick (OPObjectIn src, Card32 id);

	//	Overridden methods :

	virtual Bool SetArraySize (Count w, Count h);						/* OVERRIDDEN */
	virtual Bool Detect (OPCoord x, OPCoord y, OPViewOut);				/* OVERRIDDEN */
	virtual void Draw (OPGrailPortIn, OPRectIn, OPDrawOp = OP_DRAW_ALL);/* OVERRIDDEN */
	virtual void SetCell (Count h, Count v, OPControlIn);				/* OVERRIDDEN */
	
	OP_DEFINE_CLASS (OPScrollList, OPCellGrid);
};

#endif /* _OPaC_SCRLIST_H_ */

