/*
 *	pubmacro.h
 *
 *	This exports the macros which will be used to publish information
 *	about queriable data.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_PUBMACRO_H_
#define	_OPaC_PUBMACRO_H_

#ifndef _OPaC_OBJECT_H_
#include <OPaC/object.h>
#endif

OP_INTERFACE (pubmacro.h);

/****************************************************************/

/*
 *	The accessors and modifiers are defined as pointers to
 *	member functions, one for each supported type :
 */

typedef Bool (*OPaC_ReadAccessBool)(const OPObject*, BoolOut);
typedef Bool (*OPaC_ReadAccessInt32)(const OPObject*, Int32Out);
typedef Bool (*OPaC_ReadAccessCard32)(const OPObject*, Card32Out);
typedef Bool (*OPaC_ReadAccessReal)(const OPObject*, RealOut);
typedef Bool (*OPaC_ReadAccessOPObject)(const OPObject*, OPObjectOut);

typedef Bool (*OPaC_WriteAccessBool)(OPObject*, BoolIn);
typedef Bool (*OPaC_WriteAccessInt32)(OPObject*, Int32In);
typedef Bool (*OPaC_WriteAccessCard32)(OPObject*, Card32In);
typedef Bool (*OPaC_WriteAccessReal)(OPObject*, RealIn);
typedef Bool (*OPaC_WriteAccessOPObject)(OPObject*, OPObjectIn);

typedef Bool   (*OPaC_FunctionBool)(const OPObject*);
typedef Int32  (*OPaC_FunctionInt32)(const OPObject*);
typedef Card32 (*OPaC_FunctionCard32)(const OPObject*);
typedef Real   (*OPaC_FunctionReal)(const OPObject*);
typedef OPObjectRef (*OPaC_FunctionOPObject)(const OPObject*);

/*
 *	The accessors and modifiers are grouped together into
 *	two unions in order to reduce storage space :
 */

union OPaC_PubDataRdAcc
{
	OPaC_ReadAccessBool		 accessBool;
	OPaC_ReadAccessInt32	 accessInt32;
	OPaC_ReadAccessCard32	 accessCard32;
	OPaC_ReadAccessReal		 accessReal;
	OPaC_ReadAccessOPObject	 accessOPObject;
	
	OPaC_FunctionBool		 funcBool;
	OPaC_FunctionInt32		 funcInt32;
	OPaC_FunctionCard32		 funcCard32;
	OPaC_FunctionReal		 funcReal;
	OPaC_FunctionOPObject	 funcOPObject;
};

union OPaC_PubDataWrAcc
{
	OPaC_WriteAccessBool	 accessBool;
	OPaC_WriteAccessInt32	 accessInt32;
	OPaC_WriteAccessCard32	 accessCard32;
	OPaC_WriteAccessReal	 accessReal;
	OPaC_WriteAccessOPObject accessOPObject;
};

/****************************************************************/

/*
 *	Data can be published thanks to the following macros :
 *
 *	- OP_PUBLISH_BITSET_INFO (Class, "name", info);
 *	- OP_PUBLISH_MODIFIER (Class, "name", Get, Set, "desc.", BaseType, RealType);
 *	- OP_PUBLISH_MODIFUNC (Class, "name", Func, Set, "desc.", BaseType, RealType);
 *	- OP_PUBLISH_ACCESSOR (Class, "name", Get, "desc.", BaseType, RealType);
 *	- OP_PUBLISH_SETVALUE (Class, "name", Set, "desc.", BaseType, RealType);
 *	- OP_PUBLISH_FUNCTION (Class, "name", Func, "desc.", BaseType, RealType);
 *
 *	The parameters are defined as following :
 *
 *	- Class is the class name for which the publishing is done.
 *	- "name" is the name of the data field.
 *	- Get/Set/Func is the method for getting, setting or retrieving
 *	  the field.
 *	- "desc." is a short description of the field.
 *	- BaseType is one of Bool/Int32/Card32/Real/OPObject, compatible
 *	  with the type of the field.
 *	- RealType is the real type of the field.
 */

#define	OP_PUBLISH_BITSET_INFO(c,name,b)						\
	OP_PUBLISH_BITSET_INFO_1(c,name,b,__LINE__)

#define	OP_PUBLISH_MODIFIER(c,name,get,set,desc,type,rtyp)		\
	OP_PUBLISH_MODIFIER_1(c,name,get,set,desc,type,rtyp,__LINE__)

#define	OP_PUBLISH_MODIFUNC(c,name,fun,set,desc,type,rtyp)		\
	OP_PUBLISH_MODIFUNC_1(c,name,fun,set,desc,type,rtyp,__LINE__)

#define	OP_PUBLISH_ACCESSOR(c,name,get,desc,type,rtyp)			\
	OP_PUBLISH_ACCESSOR_1(c,name,get,desc,type,rtyp,__LINE__)

#define	OP_PUBLISH_SETVALUE(c,name,set,desc,type,rtyp)			\
	OP_PUBLISH_SETVALUE_1(c,name,set,desc,type,rtyp,__LINE__)

#define	OP_PUBLISH_FUNCTION(c,name,fun,desc,type,rtyp)			\
	OP_PUBLISH_FUNCTION_1(c,name,fun,desc,type,rtyp,__LINE__)

/****************************************************************/

#define	OP_PUBLISH_BITSET_INFO_1(c,name,b,l)					\
	OP_PUBLISH_BITSET_INFO_2(c,name,b,l)

#define	OP_PUBLISH_MODIFIER_1(c,name,get,set,desc,type,rtyp,l)	\
	OP_PUBLISH_MODIFIER_2(c,name,get,set,desc,type,rtyp,l)

#define	OP_PUBLISH_MODIFUNC_1(c,name,fun,set,desc,type,rtyp,l)	\
	OP_PUBLISH_MODIFUNC_2(c,name,fun,set,desc,type,rtyp,l)

#define	OP_PUBLISH_ACCESSOR_1(c,name,get,desc,type,rtyp,l)		\
	OP_PUBLISH_ACCESSOR_2(c,name,get,desc,type,rtyp,l)

#define	OP_PUBLISH_SETVALUE_1(c,name,set,desc,type,rtyp,l)		\
	OP_PUBLISH_SETVALUE_2(c,name,set,desc,type,rtyp,l)

#define	OP_PUBLISH_FUNCTION_1(c,name,fun,desc,type,rtyp,l)		\
	OP_PUBLISH_FUNCTION_2(c,name,fun,desc,type,rtyp,l)

/****************************************************************/

#define	OP_PUBLISH_BITSET_INFO_2(c,name,b,l)					\
	OP_PUBLISH__B0(c,l,name);									\
	notify->PublishDataEx (pdd, b, sizeof (b));					\
	OP_PUBLISH__E(c,l)

/****************************************************************/

#define	OP_PUBLISH_MODIFIER_2(c,name,get,set,desc,type,rtyp,l)	\
	OP_PUBLISH__RDACC(c,l,type,rtyp,get);						\
	OP_PUBLISH__WRACC(c,l,type,rtyp,set);						\
	OP_PUBLISH__B(c,l,name,desc,rtyp);							\
	pdd.is_read_ok  = TRUE;										\
	pdd.is_write_ok = TRUE;										\
	pdd.is_accessor = TRUE;										\
	pdd.is_function = FALSE;									\
	pdd.rd.access##type = __rd_acc_##c##_##l;					\
	pdd.wr.access##type = __wr_acc_##c##_##l;					\
	OP_PUBLISH__E(c,l)

#define	OP_PUBLISH_MODIFUNC_2(c,name,fun,set,desc,type,rtyp,l)	\
	OP_PUBLISH__RDFUN(c,l,type,rtyp,fun);						\
	OP_PUBLISH__WRACC(c,l,type,rtyp,set);						\
	OP_PUBLISH__B(c,l,name,desc,rtyp);							\
	pdd.is_read_ok  = TRUE;										\
	pdd.is_write_ok = TRUE;										\
	pdd.is_accessor = TRUE;										\
	pdd.is_function = TRUE;										\
	pdd.rd.func##type   = __rd_fun_##c##_##l;					\
	pdd.wr.access##type = __wr_acc_##c##_##l;					\
	OP_PUBLISH__E(c,l)

#define	OP_PUBLISH_ACCESSOR_2(c,name,get,desc,type,rtyp,l)		\
	OP_PUBLISH__RDACC(c,l,type,rtyp,get);						\
	OP_PUBLISH__B(c,l,name,desc,rtyp);							\
	pdd.is_read_ok  = TRUE;										\
	pdd.is_write_ok = FALSE;									\
	pdd.is_accessor = TRUE;										\
	pdd.is_function = FALSE;									\
	pdd.rd.access##type = __rd_acc_##c##_##l;					\
	pdd.wr.access##type = 0;									\
	OP_PUBLISH__E(c,l)

#define	OP_PUBLISH_SETVALUE_2(c,name,set,desc,type,rtyp,l)		\
	OP_PUBLISH__WRACC(c,l,type,rtyp,set);						\
	OP_PUBLISH__B(c,l,name,desc,rtyp);							\
	pdd.is_read_ok  = FALSE;									\
	pdd.is_write_ok = TRUE;										\
	pdd.is_accessor = TRUE;										\
	pdd.is_function = FALSE;									\
	pdd.rd.access##type = 0;									\
	pdd.wr.access##type = __wr_acc_##c##_##l;					\
	OP_PUBLISH__E(c,l)

#define	OP_PUBLISH_FUNCTION_2(c,name,fun,desc,type,rtyp,l)		\
	OP_PUBLISH__RDFUN(c,l,type,rtyp,fun);						\
	OP_PUBLISH__B(c,l,name,desc,rtyp);							\
	pdd.is_read_ok  = TRUE;										\
	pdd.is_write_ok = FALSE;									\
	pdd.is_accessor = FALSE;									\
	pdd.is_function = TRUE;										\
	pdd.rd.func##type   = __rd_fun_##c##_##l;					\
	pdd.wr.access##type = 0;									\
	OP_PUBLISH__E(c,l)

/****************************************************************/

#define OP_PUBLISH__WRACC(c,l,type,realtype,set)				\
static Bool														\
__wr_acc_##c##_##l (OPObject* obj, type##In data)				\
{																\
	c* instance = (c*)(obj);									\
	return instance->set (data);								\
}																\
class OPObject

#define OP_PUBLISH__RDACC(c,l,type,realtype,get)				\
static Bool														\
__rd_acc_##c##_##l (const OPObject* obj, type##Out data)		\
{																\
	const c* instance = (const c*)(obj);						\
	realtype##Ref out;											\
	Bool ok = instance->get (out);								\
	data = out;													\
	return ok;													\
}																\
class OPObject

#define OP_PUBLISH__RDFUN(c,l,type,realtype,fun)				\
static type##Ref												\
__rd_fun_##c##_##l (const OPObject* obj)						\
{																\
	const c* instance = (const c*)(obj);						\
	return instance->fun ();									\
}																\
class OPObject

#define	OP_PUBLISH__B0(c,l,name)								\
static void														\
__collect_##c##_##l (OPRuntime* notify)							\
{																\
	OPaC_PubDataDef pdd;										\
	pdd.runtime     = & OP_RUNTIME (c);							\
	pdd.var_name    = name

#define	OP_PUBLISH__B(c,l,name,desc,type)						\
	OP_PUBLISH__B0(c,l,name);									\
	pdd.var_desc    = desc;										\
	pdd.var_type    = type##TypeDef

#define	OP_PUBLISH__E(c,l)										\
	notify->PublishData (pdd);									\
}																\
static OPaC_PubDataExport										\
__export_##c##_##l (__collect_##c##_##l)

/****************************************************************/

enum OPaC_BitsetDataType
{
	OP_BITSET_DATA_MASK,
	OP_BITSET_DATA_BIT,
	OP_BITSET_DATA_VALUE,
};

struct OPaC_BitsetDataDef
{
	Card32				value;					//	value (either mask/bit/value)
	const LatinChar*	name;					//	name of this element
	OPaC_BitsetDataType	type;					//	type (either mask/bit/value)
};

/****************************************************************/

/*
 *	The public data declaration needs some special structures
 *	which won't be used directly by the programmer, but will
 *	be generated by the OP_PUBLISH_xxx macros.
 */

struct OPaC_PubDataDef
{
	const OPRuntimeRef*	runtime;				//	runtime description
	const LatinChar*	var_name;				//	variable name
	const LatinChar*	var_desc;				//	variable description
	OPTypeDef			var_type;				//	variable type
	Bool				is_read_ok;				//	TRUE => may read from variable
	Bool				is_write_ok;			//	TRUE => may write to variable
	Bool				is_accessor;			//	TRUE => uses an accessor
	Bool				is_function;			//	TRUE => uses a function
	OPaC_PubDataRdAcc	rd;						//	pointer to read accessor
	OPaC_PubDataWrAcc	wr;						//	pointer to write modifier
};

typedef void (*OPaC_PubDataInFunc)(OPRuntime*);

/*
 *	The published data export class will record the individual
 *	data definitions which are then handled by OPRuntime.
 */

class OP_PUBLIC OPaC_PubDataExport
{
	friend class OPRuntime;
	
private:
	static OPaC_PubDataExport* head;			//	first published data
	static OPaC_PubDataExport* tail;			//	last published data
	
private:
	OPaC_PubDataExport*	next;					//	next published data in global list
	OPaC_PubDataInFunc	init;					//	published data initialization function
	
public:
	OPaC_PubDataExport (OPaC_PubDataInFunc init)
		{ (this->tail ? this->tail->next : this->head) = this;
		  this->init = init; this->next = 0; this->tail = this; }
};

/****************************************************************/

#endif /* _OPaC_PUBMACRO_H_ */
