/*
 *	types.h
 *
 *	This file contains the main types used throughout the OPaC project files
 *	and also some useful macros for asserting some conditions.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef	_OPaC_TYPES_H_
#define	_OPaC_TYPES_H_

#ifndef _OPaC_COMPILER_H_
#include <OPaC/compiler.h>
#endif

OP_INTERFACE (types.h);

/****************************************************************/

/*
 *	The simple types have an associated "identifier" which is
 *	simpler to use than the OPTypeDef structure.
 */

typedef Card8 OPSimpleType;

#define	OP_TYPE_NONE	((OPSimpleType)(0))		//	no type (NULL)
#define	OP_TYPE_BOOL	((OPSimpleType)(1))		//	boolean
#define	OP_TYPE_INT		((OPSimpleType)(2))		//	32-bit signed integer
#define	OP_TYPE_CARD	((OPSimpleType)(3))		//	32-bit unsigned integer
#define	OP_TYPE_ENUM	((OPSimpleType)(4))		//	enumeration
#define	OP_TYPE_REAL	((OPSimpleType)(5))		//	real, floating point

#define	OP_TYPE_OBJECT	((OPSimpleType)(100))	//	any kind of object
#define	OP_TYPE_TEXT	((OPSimpleType)(101))	//	either OPString or OPAtom
#define	OP_TYPE_ARRAY	((OPSimpleType)(102))	//	an array of objects

#define	OP_TYPE_USER	((OPSimpleType)(120))	//	user type 120..169

#define	OP_TYPE_LSTRING	((OPSimpleType)(198))	//	a LatinChar string
#define	OP_TYPE_USTRING	((OPSimpleType)(199))	//	a UnicodeChar string

//	User defined types :

#define	OP_TYPE_STRUCT	((OPSimpleType)(200))	//	structure (complex)
#define	OP_TYPE_UNION	((OPSimpleType)(201))	//	union (complex)
#define	OP_TYPE_OTHER	((OPSimpleType)(202))	//	special complex type

#define	OP_TYPE_ANY		((OPSimpleType)(255))	//	any type -- not specified

/****************************************************************/

/*
 *	It is interesting to know the data types at run-time.
 *	That's why OPaC provides OP_TYPExxx macros below. It
 *	needs some room to record data about them :
 */

struct OPaC_TypeDef
{
	const char*			type;					//	data type name
	OPaC_TypeDef*		derived_from;			//	data type which it has been derived from
	OPSimpleType		simple;					//	data type ID (if it is a simple one)
	const void*			runtime;				//	pointer to runtime description (if any)
	unsigned long		size;					//	data type size
};

typedef const OPaC_TypeDef*	OPTypeDef;

typedef const OPaC_TypeDef&	OPaC_TypeDefIn;
typedef OPaC_TypeDef&		OPaC_TypeDefOut;
typedef OPaC_TypeDef&		OPaC_TypeDefInOut;

typedef OPTypeDef			OPTypeDefIn;
typedef OPTypeDef&			OPTypeDefOut;
typedef OPTypeDef&			OPTypeDefInOut;


/*
 *	The OP_TYPE macro declares an already defined type so
 *	that it will be known by the run-time library.
 */

#define	OP_TYPE(T)			OP_TYPE_X(T,OP_PUBLIC)
#define	APP_TYPE(T)			OP_TYPE_X(T,OP_APP)

#define	OP_TYPE_X(T,t)											\
																\
t extern OPaC_TypeDef	T##_type;								\
t extern OPTypeDef	T##TypeDef;									\
																\
typedef T			T##In;										\
typedef T&			T##Out;										\
typedef T&			T##InOut;									\
																\
typedef T*			T##Ptr;										\
typedef T##Ptr		T##PtrIn;									\
typedef T##Ptr&		T##PtrOut;									\
typedef T##Ptr&		T##PtrInOut

#define	OP_IMPLEMENT_TYPE(T,X)									\
OPaC_TypeDef T##_type = { #T, 0, X, 0, sizeof (T) };			\
OPTypeDef	 T##TypeDef = & T##_type


/*
 *	The OP_TYPEENUM macro is similar to OP_TYPE, but for
 *	enumerated types.
 */

#define	OP_TYPEENUM(T)		OP_TYPEENUM_X(T,OP_PUBLIC)
#define	APP_TYPEENUM(T)		OP_TYPEENUM_X(T,OP_APP)

#define	OP_TYPEENUM_X(T,t)										\
																\
t extern OPaC_TypeDef T##_type;									\
t extern OPTypeDef	T##TypeDef;									\
																\
typedef T			T##In;										\
typedef T&			T##Out;										\
typedef T&			T##InOut;									\
																\
typedef T*			T##Ptr;										\
typedef T##Ptr		T##PtrIn;									\
typedef T##Ptr&		T##PtrOut;									\
typedef T##Ptr&		T##PtrInOut

#define	OP_IMPLEMENT_TYPEENUM(T)								\
OPaC_TypeDef T##_type =	{ #T, 0, OP_TYPE_ENUM, 0, sizeof (T) };	\
OPTypeDef	 T##TypeDef = & T##_type

/*
 *	The OP_TYPESTRUCT macro is similar to OP_TYPEENUM, but
 *	for class and structure types.
 */

#define	OP_TYPECLASS(T)			OP_TYPECLASS_X(T,OP_PUBLIC)
#define	OP_TYPESTRUCT(T)		OP_TYPESTRUCT_X(T,OP_PUBLIC)
#define	APP_TYPECLASS(T)		OP_TYPECLASS_X(T,OP_APP)
#define	APP_TYPESTRUCT(T)		OP_TYPESTRUCT_X(T,OP_APP)

#define	OP_TYPECLASS_X(T,t)		OP_TYPESTRUCT_X(T,t)
#define	OP_TYPESTRUCT_X(T,t)									\
																\
t extern OPaC_TypeDef T##_type;									\
t extern OPTypeDef	T##TypeDef;									\
																\
typedef const T&	T##In;										\
typedef T&			T##Out;										\
typedef T&			T##InOut;									\
																\
typedef T*			T##Ptr;										\
typedef T##Ptr		T##PtrIn;									\
typedef T##Ptr&		T##PtrOut;									\
typedef T##Ptr&		T##PtrInOut

#define	OP_IMPLEMENT_TYPESTRUCT(T)								\
OPaC_TypeDef T##_type =	{ #T, 0, OP_TYPE_STRUCT, 0, sizeof (T) }; \
OPTypeDef	 T##TypeDef = & T##_type

/*
 *	The OP_TYPEDEF macro defines a type as would typedef
 *	do; it will be known by the run-time library.
 */

#define	OP_TYPEDEF(B,T)		OP_TYPEDEF_X(B,T,OP_PUBLIC)
#define	APP_TYPEDEF(B,T)	OP_TYPEDEF_X(B,T,OP_APP)

#define	OP_TYPEDEF_X(B,T,t)										\
																\
typedef B T;													\
																\
t extern OPaC_TypeDef T##_type;									\
t extern OPTypeDef	T##TypeDef;									\
																\
typedef T			T##In;										\
typedef T&			T##Out;										\
typedef T&			T##InOut;									\
																\
typedef T*			T##Ptr;										\
typedef T##Ptr		T##PtrIn;									\
typedef T##Ptr&		T##PtrOut;									\
typedef T##Ptr&		T##PtrInOut

#define	OP_IMPLEMENT_TYPEDEF(B,T,X)								\
OPaC_TypeDef T##_type =	{ #T, & B##_type, X, 0, sizeof (T) };	\
OPTypeDef	 T##TypeDef = & T##_type

/****************************************************************/

#ifndef _OPaC_SYS_H_
#include <OPaC/sys.h>
#endif

/****************************************************************/

//	Publish the machine independent signed and unsigned integer types
//	which will be widely used throughout the OPaC project :

OP_TYPE (Int8);					typedef	Int8	Int8Ref;
OP_TYPE (Int16);				typedef	Int16	Int16Ref;
OP_TYPE (Int32);				typedef	Int32	Int32Ref;
OP_TYPE (Int64);				typedef	Int64	Int64Ref;

OP_TYPE (Card8);				typedef	Card8	Card8Ref;
OP_TYPE (Card16);				typedef	Card16	Card16Ref;
OP_TYPE (Card32);				typedef	Card32	Card32Ref;
OP_TYPE (Card64);				typedef	Card64	Card64Ref;

//	Also publish real as some precise (80-bit or more) floating point
//	type which is used for this project, instead of double, etc. :

OP_TYPE (Real);					typedef	Real	RealRef;

//	Simple boolean definition. Be sure this will take as few space as
//	possible. Enums may take a lot of room...

OP_TYPEDEF (Card8,	Bool);		typedef	Bool	BoolRef;

#undef	FALSE
#undef	TRUE

#define	FALSE	((Bool)(0))
#define	TRUE	((Bool)(1))


//	Some character related definitions. Whenever you store characters,
//	you should either choose LatinChar or UnicodeChar.

typedef	char		LatinChar;

OP_TYPE (LatinChar);
OP_TYPEDEF (Card16,	UnicodeChar);


//	The following types should be defined in <OPaC/grail.h>, but they
//	are also needed in other contexts :

OP_TYPEDEF (Card32, OPColorPixel);				//	colored pixel representation
OP_TYPEDEF (Real,   OPCoord);					//	coordinate is a reasonable quantity
OP_TYPEDEF (Real,   OPRelCoord);				//	relative coordinate is signed


//	A comparison will yield 4 possible results, which have been coded
//	as #defines rather than an enum. This way, the programmer may just
//	manipulate ints.

#define	OP_COMP_THIS_BEFORE		(-1)
#define	OP_COMP_THIS_SAME		( 0)
#define	OP_COMP_THIS_AFTER		( 1)


//	The counter has a fixed size. This is for portability considerations
//	rather than speed.

OP_TYPEDEF (Card32,	Count);


//	Byte and Card8 are synonyms, but Card8 may be preferred by some people.

OP_TYPEDEF (Card8, Byte);

/****************************************************************/

//	Bit definition (used in bitsets). Replaces the ugly 0x0008000 and
//	other kind of constant definitions.

#define	OP_BIT(x)	(((Card32)(1)) << x)


//	Error and warning functions & macros, used mainly for debugging.

extern OP_PUBLIC volatile void OPaC_ErrorAbort (const LatinChar*, int, const LatinChar*);
extern OP_PUBLIC void OPaC_ErrorPrint (const LatinChar*, int, const LatinChar*);

#define	OP_ABORT(x)												\
	OPaC_ErrorAbort (x, __LINE__, __FILE__)

#define	OP_ASSERT(x)											\
	(x) ? (void)(0) /* ok, no error */							\
		: (OPaC_ErrorAbort ("Failed assertion: " #x, __LINE__, __FILE__), (void)(0))

#define	OP_WARN_IF(x)											\
	(x) ? (void)(0) /* ok, no error */							\
		: (OPaC_ErrorPrint ("Warning: " #x, __LINE__, __FILE__), (void)(0))

#define	OP_POSSIBLE_PROBLEM(x)									\
	do { } while (0)

/****************************************************************/

/*
 *	Every header file should export its class declarations
 *	before it includes any other file.
 */

#define	OP_IMPORT_CLASS(c)	OP_EXPORT_CLASS(c)

#define	APP_EXPORT_CLASS(c)	OP_EXPORT_CLASS(c)
#define	APP_IMPORT_CLASS(c)	OP_IMPORT_CLASS(c)

#define OP_EXPORT_CLASS(c)										\
																\
class c;														\
class c##Ref;													\
class c##Array;													\
																\
typedef const c##Ref& c##In;									\
typedef c##Ref& c##Out;											\
typedef c##Ref& c##InOut;										\
																\
typedef const c##Array& c##ArrayIn;								\
typedef c##Array& c##ArrayOut;									\
typedef c##Array& c##ArrayInOut

/****************************************************************/

#endif /* _OPaC_TYPES_H_ */
