/*
 *	view.h
 *
 *	Base View class definition. The view class is the base class for every
 *	visual interface element (widget).
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _OPaC_VIEW_H_
#define	_OPaC_VIEW_H_

#ifndef _OPaC_OPEXPORT_H_
#include <OPaC/opexport.h>
#endif

OP_INTERFACE (view.h);

#ifndef _OPaC_ATOM_H_
#include <OPaC/atom.h>
#endif

/****************************************************************/

/*
 *	Each view has some capabilities and has some internal
 *	state. It is represented by the following bitset :
 */

OP_TYPEDEF (Card32, OPViewState);				//	it is a bitset, therefore not an enum

#define	OP_VIEW_IS_ENABLED	OP_BIT (0)			//	TRUE => view may get events
#define	OP_VIEW_IS_ALIVE	OP_BIT (1)			//	TRUE => view may become active
#define	OP_VIEW_IS_FROZEN	OP_BIT (2)			//	TRUE => view is frozen -- NIY
#define	OP_VIEW_IS_HOT		OP_BIT (3)			//	TRUE => view is hot -- NIY
#define	OP_VIEW_IS_ACTIVE	OP_BIT (4)			//	TRUE => view has keyboard focus
#define	OP_VIEW_IS_DRAGGING	OP_BIT (5)			//	TRUE => view is the current drag source
#define	OP_VIEW_IS_DROPPING	OP_BIT (6)			//	TRUE => view is the current drop destination
#define	OP_VIEW_IS_DISABLED	OP_BIT (7)			//	TRUE => view has been globally disabled
#define	OP_VIEW_IS_CONNECT	OP_BIT (8)			//	TRUE => view is the current connection destination

#define	OP_VIEW_OK_DRAG		OP_BIT (16)			//	TRUE => view accepts dragging
#define	OP_VIEW_OK_DROP		OP_BIT (17)			//	TRUE => view accepts dropping
#define	OP_VIEW_OK_DETECT	OP_BIT (18)			//	TRUE => view accepts detection
#define	OP_VIEW_OK_MAIN		OP_BIT (19)			//	TRUE => view accepts to become main responder
#define	OP_VIEW_OK_UI_EDIT	OP_BIT (20)			//	TRUE => view accepts user interface editing
#define	OP_VIEW_OK_UI_MOVE	OP_BIT (21)			//	TRUE => view accepts user interface moving

/****************************************************************/

OP_TYPEDEF (Card32, OPDrawOp);					//	it is a bitset

#define	OP_DRAW_NONE		((OPDrawOp)(0x00))	//	draw nothing
#define	OP_DRAW_OBJECT		((OPDrawOp)(0x01))	//	draw the object (foreground)
#define	OP_DRAW_BACKGROUND	((OPDrawOp)(0x02))	//	draw the background
#define	OP_DRAW_ALL			((OPDrawOp)(0x03))	//	draw everything (background + object)

/****************************************************************/

/*
 *	The drag & drop protocol defines what kind of operation is
 *	going on. The following are supported :
 *
 *	- User is dragging an object in order to drop it (real drag
 *	  and drop).
 *
 *	- User is dragging a connection point.
 */

enum OPDragDropProtocol
{
	OP_DDP_DRAGnDROP		= 0x0000,			//	drag & drop :
	OP_DDP_DRAGnDROP_CHECK	= 0x0001,			//	- check if valid target
	OP_DDP_DRAGnDROP_ENTER	= 0x0002,			//	- notify cursor entered responder
	OP_DDP_DRAGnDROP_LEAVE	= 0x0003,			//	- notify cursor left responder
	OP_DDP_DRAGnDROP_MOTION	= 0x0004,			//	- notify cursor moved over responder
	OP_DDP_DRAGnDROP_CANCEL	= 0x0005,			//	- operation cancelled
	OP_DDP_DRAGnDROP_COMMIT	= 0x0006,			//	- operation committed
	OP_DDP_DRAGnDROP_SUBST	= 0x0007,			//	- substitute in place of real target
	
	OP_DDP_CONNECT			= 0x0010,			//	connection :
	OP_DDP_CONNECT_CHECK	= 0x0011,			//	- check if valid target
	OP_DDP_CONNECT_ENTER	= 0x0012,			//	- notify cursor entered responder
	OP_DDP_CONNECT_LEAVE	= 0x0013,			//	- notify cursor left responder
	OP_DDP_CONNECT_MOTION	= 0x0014,			//	- notify cursor moved over responder
	OP_DDP_CONNECT_CANCEL	= 0x0015,			//	- operation cancelled
	OP_DDP_CONNECT_COMMIT	= 0x0016,			//	- operation committed
	OP_DDP_CONNECT_SUBST	= 0x0017,			//	- substitute in place of real target
	
	OP_DDP_UI_EDIT			= 0x0020,			//	dynamic user interface edition:
	OP_DDP_UI_EDIT_CHECK	= 0x0021,			//	- check if valid target
	/*
	 *
	 */
	OP_DDP_UI_EDIT_COMMIT	= 0x0026,			//	- operation committed
	OP_DDP_UI_EDIT_SUBST	= 0x0027			//	- substitute in place of real target
};

OP_TYPEENUM (OPDragDropProtocol);

/****************************************************************/

#ifndef _OPaC_RECT_H_
#include <OPaC/rect.h>
#endif

/****************************************************************/

/*
 *	Base class for every responder object; a View object is
 *	basically a Responder... This means that the user interface
 *	is crowded with responders :-)
 */

class OP_PUBLIC OPResponder : public OPObject
{
	
private:
	OPResponderRef		next_responder;			//	link with next responder
	
protected:
	OPResponder ();
	virtual ~ OPResponder ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
public:
	virtual void SetNextResponder (OPResponderIn next);
	virtual Bool GetNextResponder (OPResponderOut next) const;
	
	virtual Bool GetView (OPViewOut view);
	
	//	Any responder may become the Main Responder. The Main Responder
	//	will get the key events first and will usually be the active
	//	responder in a window.
	
	virtual Bool WouldAcceptMain () const;
	virtual void BecomeMain (Bool clicked);
	virtual void ResignMain ();
	
	//	The HandleEvent should not be overridden, since it does the
	//	dispatching into the more specific event handling methods.
	//	Only OPWindow is supposed to override this method !
	
	virtual void DispatchEvent (OPEventIn event);
	
	//	The following methods will make the event handling easier :
	
	virtual void KeyDown (OPEventIn event);
	virtual void KeyChange (OPEventIn event);
	virtual void KeyUp (OPEventIn event);
	
	virtual void ButtonPress (OPEventIn event);
	virtual void ButtonRelease (OPEventIn event);
	
	virtual void MotionNotify (OPEventIn event);
	virtual void DragNotify (OPEventIn event);
	
	virtual void Tick ();
	virtual void Miscellanous (OPEventIn event);
	
	//	The following method cannot be called by the dispatching
	//	method. It will be called by the window or application.
	
	virtual void EnterNotify (OPEventIn event);
	virtual void LeaveNotify (OPEventIn event);
	
	OP_DEFINE_CLASS (OPResponder, OPObject);
};

/****************************************************************/

/*
 *	The OPaC View class describes a graphical element used
 *	in the user interface. It specifies a hierarchy and a
 *	geometry.
 */

class OP_PUBLIC OPView : public OPResponder
{
		
	struct OP_PUBLIC ViewChildren
	{
		OPViewRef		head;					//	first child
		OPViewRef		tail;					//	last child
	};
	
	struct OP_PUBLIC ViewSiblings
	{
		OPViewRef		next;					//	next sibling
		OPViewRef		prev;					//	previous sibling
	};
	
protected:
	static OPAtomRef	atom_look_man_changed;	//	atom "LookManChanged" (used by Broadcast)
	
	OPWindowRef			window;					//	link to containing window
	OPViewRef			parent;					//	link to parent view
	OPSpringBoardRef	spring;					//	link to associated spring curve

	ViewSiblings		sibling;				//	link to sibling views
	ViewChildren		child;					//	link to children views (subviews)
	ViewSiblings		active_sibling;			//	link to next/previous sibling to activate
	ViewChildren		active_child;			//	link to first/last child to activate
	OPRect				box;					//	window relative containing box
	OPViewState			state;					//	current state of the view (bitset)
	Int32				tag;					//	tag (can be used freely by programmer)

	Bool				no_draw_back;			//	don't draw the background of the view
	
protected:
	OPView ();
	virtual ~ OPView ();
	
	virtual void StoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	virtual void RestoreInstance (OPStorageIn storage);	/* OVERRIDDEN */
	
	virtual void ChangeGeometry (OPRectIn box);
	virtual void SetEnableState (Bool ok, Bool broadcast = TRUE);
	virtual void RemoveActive ();
	
public:
	virtual void Delete ();								/* OVERRIDDEN */
	virtual Bool GetView (OPViewOut view);				/* OVERRIDDEN */
	virtual Bool WouldAcceptMain () const;				/* OVERRIDDEN */
	virtual void BecomeMain (Bool clicked);				/* OVERRIDDEN */
	virtual void ResignMain ();							/* OVERRIDDEN */

	virtual void MakeMainResponder ();
	
	//	The view has some complex relationship with other views, an associated
	//	geometry and lots of attributes. Here are the accessors :
	
	virtual Bool SetWindow (OPWindowIn window);
	virtual Bool SetParent (OPViewIn parent);
	virtual Bool SetSpringBoard (OPSpringBoardIn spring);

	virtual Bool WouldAcceptChild (OPViewIn child) const;
	virtual void NotifyInsertChild (OPViewIn child);
	virtual void NotifyRemoveChild (OPViewIn child);
	
	virtual void ReorderFirst ();
	virtual void ReorderLast ();
	virtual void ReorderBefore (OPViewIn which_view);
	virtual void ReorderAfter (OPViewIn which_view);
	
	virtual void SetBox (OPRectIn box);
	virtual void SetOrigin (OPCoord x, OPCoord y);
	virtual void SetSize (OPCoord dx, OPCoord dy);
	virtual void ChangeOrigin (OPRelCoord dx, OPRelCoord dy);
	virtual void ChangeSize (OPRelCoord dx, OPRelCoord dy);
	
	virtual void SetFirstActive (OPViewIn view);
	virtual void SetLastActive (OPViewIn view);
	virtual void SetActiveBefore (OPViewIn view);
	virtual void SetActiveAfter (OPViewIn view);
	virtual Bool SetUserTag (Int32 tag);
	
	virtual Bool GetWindow (OPWindowOut window) const;
	virtual Bool GetParent (OPViewOut view) const;
	virtual Bool GetSpringBoard (OPSpringBoardOut spring) const;
	
	virtual Bool GetNextSibling (OPViewOut view) const;
	virtual Bool GetPrevSibling (OPViewOut view) const;
	virtual Bool GetFirstChild (OPViewOut view) const;
	virtual Bool GetLastChild (OPViewOut view) const;
	
	virtual void GetOrigin (OPCoordOut x, OPCoordOut y) const;
	virtual void GetSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetMinSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual void GetMaxSize (OPCoordOut dx, OPCoordOut dy) const;
	virtual const OPRect& ReturnBox () const;
	virtual Bool GetBox (OPRectOut box) const;
	
	virtual Bool GetFirstActive (OPViewOut view) const;
	virtual Bool GetLastActive (OPViewOut view) const;
	virtual Bool GetNextActive (OPViewOut view) const;
	virtual Bool GetPrevActive (OPViewOut view) const;
	virtual void FindNextActive (OPViewOut view) const;
	virtual void FindPrevActive (OPViewOut view) const;
	virtual Int32 ReturnUserTag () const;
	
	//	Access to the view's state. Only some flags are publicly modifiable.
	//	The others will have to be modified with some specialized method.
	
	virtual void SetFlagOkDetect (Bool ok);
	virtual void SetFlagOkUIEdit (Bool ok);
	virtual void SetFlagOkUIMove (Bool ok);
	virtual void SetFlagOkMain (Bool ok);
	virtual void SetFlagOkDrag (Bool ok);
	virtual void SetFlagOkDrop (Bool ok);
	
	virtual void SetEnable (Bool ok);
	virtual Bool ReturnEnable () const;
	
	virtual void SetGlobalEnable (Bool mode);
	virtual Bool ReturnGlobalEnable () const;
	
	virtual void ChangeState (OPViewState state, Bool set);
	virtual Bool SetState (OPViewState state);
	virtual OPViewState ReturnState () const;
	
	//	Miscellanous methods :
	
	virtual Bool FindUserTag (Int32 tag, OPViewOut view);
	virtual Bool Detect (OPCoord x, OPCoord y, OPViewOut view);
	virtual Bool Detect (OPRectIn box);
	
	virtual Bool IsMainResponder () const;
	virtual void ParentResizing (OPCoord dx, OPCoord dy);
	virtual void ChildChangedSize (OPViewIn child);
	virtual void Broadcast (OPAtomIn message, OPObjectIn value);
	
	//	Draw the entire view or just part of it. This will always draw into
	//	a buffer before being dumped to the screen (this results in slower
	//	initial drawing but gets rid of some ugly glitches and speeds up the
	//	redrawing operations significantly).
	
	virtual void Draw (OPGrailPortIn port, OPDrawOp what = OP_DRAW_ALL);
	virtual void Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what = OP_DRAW_ALL);
	virtual void DrawOverlay (OPGrailPortIn port, OPRectIn rect);
	virtual void Display ();
	virtual void Display (OPRectIn rect);
	
	//	Special accessor methods for frequently used semi-global variables :
	
	virtual const OPGrailPortRef& ReturnGrailPort () const;
	virtual const OPLookManRef& ReturnLookMan () const;
	virtual const OPAdornerRef& ReturnAdorner () const;
	
	//	Drag & drop protocol handling :
	
	virtual Bool DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp);
	
	OP_DEFINE_CLASS (OPView, OPResponder);
};

/****************************************************************/

#endif /* _OPaC_VIEW_H_ */
