/*
 *	sd_event.cxx
 *
 *	Portable event interface. Implementation for X11
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 08.02.98

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <string.h>
#include <stdio.h>

#include "sd_interface.h"

#include <OPaC/new.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/window.h>
#include <OPaC/app.h>

/****************************************************************/

#define	OP_TIMER_TICK	(50)			//	timer tick interval [ms]

OP_IMPLEMENT_CLASS (OPEvent);

/****************************************************************/


OPEventRef OPEvent::current = 0;


/****************************************************************/


static int capture_count = 0;
static bool quit_demand = FALSE;

// Stuff for call to "select" used for timer interrupts

static int          fileDescriptor; // needed by select()
static fd_set       readFlag;       // idem
static fd_set       readFdSelect;   // idem
static int          nfds;           // idem
static int          status;         // idem

static timeval timeOut;


// Returns a pointer to an OPaC window given a X window

static OPGrailWindow* ReturnOPwinFromXwin(Window xwindow)
{
	unsigned long addr;

	unsigned char *inst;
	Atom xinst_atom;
	Atom xtype_atom;
	int format;
	unsigned long nitems;
	unsigned long bytes_after;
	
	xinst_atom = XInternAtom(xdpy, OP_XPROP_WIN_INST, False);

  // The adress of the OPGrailWindow structure is stored in a window property

	XGetWindowProperty(xdpy, xwindow,
										 xinst_atom,
										 0, 8,
										 0,
										 XA_STRING,
										 &xtype_atom,
										 &format,
										 &nitems,
										 &bytes_after,
										 (unsigned char**)&inst);

  OP_ASSERT(xtype_atom != None);  // It must be an OPaC window

	inst[nitems]=0;
	sscanf(inst, "%x", &addr);

	XFree(inst);

	return (OPGrailWindow*)addr;
}


// Fills in buttons and modifiers fields according to state field of X event

void
SetOPEventState(Card32& buttons, Card32& modifiers, unsigned int state)
{
	buttons = modifiers = 0;

	if( (state & Button1Mask) ) buttons |= OP_EV_BUTTON_LEFT ;
	if( (state & Button2Mask) ) buttons |= OP_EV_BUTTON_MIDDLE ;
	if( (state & Button3Mask) ) buttons |= OP_EV_BUTTON_RIGHT ;
	if( (state & Button4Mask) ) buttons |= OP_EV_BUTTON_EXTRA ;
	if( (state & Button5Mask) ) buttons |= OP_EV_BUTTON_EXTRA ;			

	if( (state & ShiftMask)   ) modifiers |= OP_EV_SHIFT   ;
	if( (state & ControlMask) ) modifiers |= OP_EV_CONTROL ;
	if( (state & LockMask)    ) modifiers |= OP_EV_CAPS_LOCK ;
	if( (state & Mod1Mask)    ) modifiers |= OP_EV_META_1 ;
	if( (state & Mod2Mask)    ) modifiers |= OP_EV_META_2 ;
	if( (state & Mod3Mask)    ) modifiers |= OP_EV_META_3 ;
	if( (state & Mod4Mask)    ) modifiers |= OP_EV_META_4 ;
	if( (state & Mod5Mask)    ) modifiers |= OP_EV_META_5 ;
}
			
// Fills in the OPKey structure and the name of key used

void
HandleKeyEvent( OPKeyOut key, OPAtomRef name, XKeyEvent& xkey)
{

	LatinChar buffer[16];
	LatinChar c;
	KeySym sym;
	XComposeStatus status_in_out;

	key.type = 0;
	key.value = 0;
	key.code = 0;

	// Let's check out what kinda key it is

	XLookupString(&xkey, &c, 1, &sym, &status_in_out); 

	// Character keys

	if( sym >= XK_space && sym <= XK_ydiaeresis ) { // sym is in LATIN1
		key.type = OP_KEY_ALPHANUM;

		if( sym >= XK_a && sym <= XK_z )  // sym is lower case
			key.code = (Card8) (XKeysymToString(sym - XK_a + XK_A)[0]);
		else
			key.code = (Card8) (XKeysymToString(sym)[0]);
		
		key.value = (Card16) c;
	}
	else {

		// Common keys
		key.value = 0;

		switch( sym ) { 
		case XK_BackSpace: name = "BACKSPACE"; key.type  = OP_KEY_BACKSPACE; break;
		case XK_Clear:	   name = "CLEAR";     key.type  = OP_KEY_SPECIAL;   break;
		case XK_Delete:    name = "DELETE";    key.type  = OP_KEY_DELETE;    break;
		case XK_Escape:	   name = "ESCAPE";    key.type  = OP_KEY_SPECIAL;   break;
		case XK_Execute:   name = "EXECUTE";   key.type  = OP_KEY_SPECIAL;   break;
		case XK_Help:      name = "HELP";      key.type  = OP_KEY_SPECIAL;   break;
		case XK_Insert:    name = "INSERT";    key.type  = OP_KEY_SPECIAL;   break;
		case XK_Num_Lock:  name = "NUM LOCK";  key.type  = OP_KEY_SPECIAL;   break;
		case XK_Pause:	   name = "PAUSE";     key.type  = OP_KEY_SPECIAL;   break;
		case XK_Print:	   name = "PRINT";     key.type  = OP_KEY_SPECIAL;   break;
		case XK_Return:	   name = "RETURN";    key.type  = OP_KEY_RETURN;    break;
		case XK_Select:	   name = "SELECT";    key.type  = OP_KEY_SPECIAL;   break;
		case XK_KP_Separator:name = "SEPARATOR";key.type  = OP_KEY_SPECIAL;  break;
		case XK_space:	   name = "SPACE";     key.value = key.code = ' ';   break;
		case XK_Tab:	     name = "TAB";       key.type  = OP_KEY_TAB;       break;
		
			//	Cursor keys, with value specifying the kind of cursor move which has
			//	been asked for :
		
		case XK_Prior: name = "PAGE UP";   
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_PAGE_UP;   break;
		case XK_Next:  name = "PAGE DOWN"; 
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_PAGE_DOWN; break;
		case XK_End:   name = "END";       
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_END;       break;
		case XK_Home:  name = "HOME";      
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_HOME;      break;
		case XK_Left:  name = "LEFT";      
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_LEFT;      break;
		case XK_Up:    name = "UP";        
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_UP;        break;
		case XK_Right: name = "RIGHT";     
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_RIGHT;     break;
		case XK_Down:  name = "DOWN";      
			key.type = OP_KEY_CURSOR; key.code = OP_CURSOR_DOWN;      break;
		
			//	Numeric pad key, which are considered to be normal keys by OPaC
		
		case XK_KP_0: case XK_KP_1: case XK_KP_2: case XK_KP_3: case XK_KP_4:
		case XK_KP_5: case XK_KP_6: case XK_KP_7: case XK_KP_8: case XK_KP_9:
			key.value = key.code = sym - XK_KP_0;
			break;
		
		case XK_KP_Multiply: key.value = key.code = '*'; break;
		case XK_KP_Add:      key.value = key.code = '+'; break;
		case XK_KP_Subtract: key.value = key.code = '-'; break;
		case XK_KP_Decimal:  key.value = key.code = '.'; break;
		case XK_KP_Divide:   key.value = key.code = '/'; break;

      // Function keys
		
		case XK_F1:  case XK_F2:  case XK_F3:  case XK_F4:
		case XK_F5:  case XK_F6:  case XK_F7:  case XK_F8:
		case XK_F9:  case XK_F10: case XK_F11: case XK_F12:
		case XK_F13: case XK_F14: case XK_F15: case XK_F16:
		case XK_F17: case XK_F18: case XK_F19: case XK_F20:
		case XK_F21: case XK_F22: case XK_F23: case XK_F24:

			if ( (sym - XK_F1 + 1) > 9) {
				buffer[0] = 'F';
				buffer[1] = (LatinChar)((sym - XK_F1 + 1) / 10) + '0';
				buffer[2] = (LatinChar)((sym - XK_F1 + 1) % 10) + '0';
				buffer[3] = 0;
			} else {
				buffer[0] = 'F';
				buffer[1] = (LatinChar)(sym - XK_F1 + 1) + '0';
				buffer[2] = 0;
			}
			
			key.type  = OP_KEY_FUNCTION;
			key.code  = sym - XK_F1 + 1;
			name = buffer;
			break;

      // An unknown key

		default: 
			key.type = OP_KEY_NONE;
		}
	}
}


/*
 *	Hand the event capture from the current window to the
 *	specified one. This may be useful if some dragging op
 *	creates a new window and needs to give it the focus.
 */

void
OPEvent::HandOverToWindow (OPWindowIn window)
{
	OP_MEMBER (HandOverToWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	OP_ASSERT (capture_count);
	OP_ASSERT (window.IsValid ());

	OPGrailWindowRef gwdo = window->ReturnGrailPort();
	Window xwindow = gwdo->ReturnPrivateWindowInfo()->xwindow;

	XGrabPointer(xdpy, 
							 xwindow,
							 0,
							 ButtonPressMask |
							 ButtonReleaseMask |
							 PointerMotionMask,
							 GrabModeAsync, GrabModeAsync,
							 None,
							 None, CurrentTime);

  XFlush(xdpy);
}

void
OPEvent::GrabWindowEvents (OPWindowIn window, Bool mode)
{
	OP_METHOD (GrabWindowEvents, 2);
	OP_ARG (1, OPWindowIn, window);
	OP_ARG (2, Bool, mode);
	
}

/****************************************************************/

/*
 *	Initialize an event record. In fact, there should be
 *	just one such record in any OPaC application.
 */

OPEvent::OPEvent ()
{
	OP_METHOD (OPEvent, 0);
	
	this->type    = OP_EVENT_NONE;
	this->id      = 0;
	
	this->window  = 0;
	
	this->x       = 0;
	this->y       = 0;
	this->buttons = 0;
	
	this->change  = 0;
	this->data    = 0;
	this->name    = "";
	this->code    = 0;
	
	this->key.type      = 0;
	this->key.value     = 0;
	this->key.code      = 0;
	this->key.modifiers = 0;
	
	this->is_dead = FALSE;

	// Stuff for call to "select" used for timer interrupts

	fileDescriptor = XConnectionNumber(xdpy);
  FD_ZERO(&readFlag);
  FD_SET(fileDescriptor,&readFlag);
  nfds = 0;
  if (fileDescriptor >= nfds) nfds = fileDescriptor + 1;

}



/*
 *	Destroy the event : this will free the references to
 *	the associated window, data and key name.
 */

OPEvent::~OPEvent ()
{
	OP_METHOD (~OPEvent, 0);
	
	this->window = 0;
	this->data   = 0;
	this->name   = "";
}



/*
 *	Is is not possible to store an event. NIY.
 */

void
OPEvent::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to save OPEvent");
}



/*
 *	Is is not possible to restore an event. NIY.
 */

void
OPEvent::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Trying to restore OPEvent");
}



/*
 *	Wait for an event and fill in the event record. This
 *	should only be called by the OPApp class.
 *
 *	Returns FALSE when no more event can be delivered by
 *	the system (the application has been killed).
 */

Bool
OPEvent::WaitForEvent ()
{
 	OP_METHOD (WaitForEvent, 0);

	static Card32 chg_buttons   = 0;
	static Card32 chg_modifiers = 0;
	
	XEvent xevent; // The current X Window event
	OPWindowRef win;  // The OPWindow where the event occurs
	
	// Trap a quit demand

	this->is_dead = quit_demand;
	this->type    = OP_EVENT_NONE;
	
	while( (this->is_dead == FALSE)
				 && (this->type == OP_EVENT_NONE)) {
		
		// The event record does not contain valid information 
		// other than the window, the mouse state & position and
		// the keyboard modifiers.

		this->id++;
		this->change = 0;
		this->data   = 0;
		this->name   = (OPAtom *)0;
		this->code   = 0;

		// Temp vars

		LatinChar* key_name = 0;

		Card32 buttons;
		Card32 modifiers;

		Bool event_waiting;
			
		while(TRUE) {
			readFdSelect = readFlag;

			timeOut.tv_sec = 0;
			timeOut.tv_usec = OP_TIMER_TICK*1000;
			
			event_waiting = XPending(xdpy);
			
			// wait for an X Window event or timeout elapsed			

			if(!event_waiting)
				status = select(nfds,&readFdSelect,NULL,NULL,&timeOut);
			else
				status = 1;

			if(!status) // timeout reached
				OPApp::CurrentApp()->HandleTick (OP_TIMER_TICK); // Send tick to App
			
			if(status || event_waiting) {
				XNextEvent(xdpy, &xevent); // Let's get it
				break;
			} 
		}

		// Let's check the type of the event and do what we've to do

		switch(xevent.type) {

		case xButtonPress:
			this->type = OP_EVENT_PRESS;
			ReturnOPwinFromXwin(xevent.xbutton.window)->GetWindow(win);
			this->window = win;
			this->x = (OPRelCoord) xevent.xbutton.x;
			this->y = (OPRelCoord) xevent.xbutton.y;
			
			capture_count++;
			SetOPEventState(buttons, modifiers, xevent.xbutton.state);
			this->buttons = buttons;
			this->key.modifiers = modifiers;

			switch( xevent.xbutton.button ) {
			case Button1:
				this->change |= OP_EV_BUTTON_LEFT;
				this->buttons |= OP_EV_BUTTON_LEFT;  break;
			case Button2:
				this->change |= OP_EV_BUTTON_MIDDLE;
				this->buttons |= OP_EV_BUTTON_MIDDLE;  break;		
			case Button3:
				this->change |= OP_EV_BUTTON_RIGHT;
				this->buttons |= OP_EV_BUTTON_RIGHT;  break;		
			default:
				this->change |= OP_EV_BUTTON_EXTRA;
				this->buttons |= OP_EV_BUTTON_EXTRA;  break;					 
			} 
			break;

		case xButtonRelease:
			this->type = OP_EVENT_RELEASE;
			ReturnOPwinFromXwin(xevent.xbutton.window)->GetWindow(win);
			this->window = win;
			this->x = (OPRelCoord) xevent.xbutton.x;
			this->y = (OPRelCoord) xevent.xbutton.y;

			XUngrabPointer(xdpy, CurrentTime);

			capture_count--;
			SetOPEventState(buttons, modifiers, xevent.xbutton.state);
			this->buttons = buttons;
			this->key.modifiers = modifiers;

			switch( xevent.xbutton.button ) {
			case Button1:
				this->change |= OP_EV_BUTTON_LEFT;
				this->buttons &= ~ OP_EV_BUTTON_LEFT; break;
			case Button2:
				this->change |= OP_EV_BUTTON_MIDDLE;
				this->buttons &= ~ OP_EV_BUTTON_MIDDLE; break;
			case Button3:
				this->change |= OP_EV_BUTTON_RIGHT;
				this->buttons &= ~ OP_EV_BUTTON_RIGHT; break;
			default:
				this->change |= OP_EV_BUTTON_EXTRA;
				this->buttons &= ~ OP_EV_BUTTON_EXTRA; break;
			} 
			break;

		case KeyPress:
			this->type = OP_EVENT_KEY_DOWN;
			ReturnOPwinFromXwin(xevent.xkey.window)->GetWindow(win);
			this->window = win;
			this->x = (OPRelCoord) xevent.xkey.x;
			this->y = (OPRelCoord) xevent.xkey.y;

			SetOPEventState(buttons, modifiers, xevent.xkey.state);
			this->buttons = buttons;
			this->key.modifiers = modifiers;			
			this->change = chg_modifiers ^ modifiers;
			chg_modifiers = modifiers;

			HandleKeyEvent( this->key, this->name, xevent.xkey );

			break;

		case KeyRelease:
			this->type = OP_EVENT_KEY_UP;
			ReturnOPwinFromXwin(xevent.xkey.window)->GetWindow(win);
			this->window = win;
			this->x = (OPRelCoord) xevent.xkey.x;
			this->y = (OPRelCoord) xevent.xkey.y;

			SetOPEventState(buttons, modifiers, xevent.xkey.state);
			this->buttons = buttons;
			this->key.modifiers = modifiers;			
			this->change = chg_modifiers ^ modifiers;
			chg_modifiers = modifiers;
			
			HandleKeyEvent( this->key, this->name, xevent.xkey );

			break;
			 
		case xMotionNotify:
			// Flush all pending motion events from the queue, drastically 
			// improves performance of drag'n'drop
			
			while( XCheckWindowEvent(xdpy, xevent.xmotion.window, 
															 PointerMotionMask,
															 &xevent));

			this->type = OP_EVENT_MOTION;
			ReturnOPwinFromXwin(xevent.xmotion.window)->GetWindow(win);		
			this->window = win;
			this->x = (OPRelCoord) xevent.xmotion.x;
			this->y = (OPRelCoord) xevent.xmotion.y;

			SetOPEventState(buttons, modifiers, xevent.xmotion.state);
			this->buttons = buttons;
			this->key.modifiers = modifiers;
						 
			break;


		case Expose:
			// shortcut to repaint the window if it has been exposed
			while( XCheckWindowEvent(xdpy, xevent.xexpose.window, 
															 ExposureMask,
															 &xevent));

			ReturnOPwinFromXwin(xevent.xexpose.window)->Redraw();
      break;		

		case ConfigureNotify:
			ReturnOPwinFromXwin(xevent.xconfigure.window)->UserChangedGeometry();
			break;
      
    case FocusIn:
      ReturnOPwinFromXwin(xevent.xfocus.window)->SysChangedActive(TRUE);
      break;

    case FocusOut:
      ReturnOPwinFromXwin(xevent.xfocus.window)->SysChangedActive(FALSE);
      break;

    case ClientMessage:      
      Atom xwm_delete_window = XInternAtom(xdpy, "WM_DELETE_WINDOW", False);

      if(xevent.xclient.data.l[0] == xwm_delete_window)
        PostQuitEvent();
    }
  }

  return this->is_dead ? FALSE : TRUE;
}

/*
 *	Dispatch the event to the window and let it do the job
 *	for us.
 */

void
OPEvent::DispatchEvent()
{
	OP_METHOD (DispatchEvent, 0);

	if (this->window.IsValid ()) {
		this->window->DispatchEvent (OPEvent::current);
	}
}

/****************************************************************/

/*
 *	Send a quit event to ourself, so that we can exit that
 *	bloody event loop :-).
 */

void
OPEvent::PostQuitEvent ()
{
	OP_METHOD (PostQuitEvent, 0);
	quit_demand = TRUE;
}



/*
 *	Initialize the event-loop related resources (eg. the
 *	timer).
 */

void
OPEvent::Initialise ()
{
	OP_MEMBER (Init, 0);
	
	OPEvent::current = OP_NEW_FROM_ZONE (OPaC_sys_zone, OPEvent);
	
}

/*
 *	Clean-up before exiting the application. This should
 *	undo what the "OPaC_InitEvent" method set up.
 */

void
OPEvent::Kill ()
{
	OP_MEMBER (Kill, 0);
	
	OPEvent::current = 0;
	
}
