/*
 *	sd_grail.cxx
 *
 *	This file implements the MS-Windows Win32 interface to the
 *	graphical API.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

/****************************************************************/

#include <math.h>

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/sd/grail.h>

/****************************************************************/

OP_MORE_IMPLEMENTATION (OPGrailPort);

/****************************************************************/

#define	BUFFER_LENGTH		400
#define	MAX_LINE_LENGTH		2000

static LatinChar g_buffer_array[BUFFER_LENGTH];
static Card32    g_pos_array[MAX_LINE_LENGTH];
static Card8     g_err_array[MAX_LINE_LENGTH];

/****************************************************************/

/*
 *	Initialise an empty Grail Port.
 */

void
OPaC_GrailPort_Initialise (OPaC_GrPort*& info, OPColorPixel f, OPColorPixel b)
{
	OP_FUNCTION (OPaC_GrailPort_Initialise, 3);
	OP_ARG (1, OPaC_GrPort*&, info);
	OP_ARG (2, OPColorPixel, f);
	OP_ARG (3, OPColorPixel, b);
	
	info = OP_ALLOC_FROM_ZONE (OPaC_sys_zone, OPaC_GrPort);
	
	info->hdc    = NULL;						//	no handle to any display context
	info->hrgn   = CreateRectRgn (0,0,1,1);		//	dummy clipping region (for later reuse)
	info->hpen   = NULL;						//	default no pen
	info->hbrush = NULL;						//	default no brush
	info->hfont  = NULL;						//	default no font
	info->hpdel  = FALSE;						//	TRUE => must delete pen
	info->hbdel  = FALSE;						//	TRUE => must delete brush
	info->hfdel  = FALSE;						//	TRUE => must delete font
	
	info->fgcol  = (COLORREF)(f);
	info->bgcol  = (COLORREF)(b);
	info->style  = PS_SOLID;
	
	OP_ASSERT (info->hrgn != 0);
}


/*
 *	Free a Grail Port.
 */

void
OPaC_GrailPort_Kill (OPaC_GrPort*& info)
{
	OP_FUNCTION (OPaC_GrailPort_Kill, 1);
	OP_ARG (1, OPaC_GrPort*&, info);
	OP_ASSERT (info->hdc == NULL);
		
	if (info->hrgn) DeleteRgn (info->hrgn);
	if (info->hpdel) DeletePen (info->hpen);
	if (info->hbdel) DeleteBrush (info->hbrush);
//	if (info->hfdel) DeleteFont (info->hfont);
	
	OP_FREE_FROM_ZONE (OPaC_sys_zone, info);
}


/*
 *	Start drawing into the Grail Port. This does not setup
 *	a clipping region...
 */

void
OPaC_GrailPort_BeginDraw (OPaC_GrPort* info, int x1, int y1, int x2, int y2,
						  OPaC_ColorMan* cman)
{
	OP_FUNCTION (OPaC_GrailPort_BeginDraw, 6);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, int, x1);
	OP_ARG (3, int, y1);
	OP_ARG (4, int, x2);
	OP_ARG (5, int, y2);
	OP_ARG (6, OPaC_ColorMan*, cman);
	
//	SetRectRgn (info->hrgn, x1, y1, x2 + 1, y2 + 1);
//	SelectClipRgn (info->hdc, info->hrgn);
	
	if (cman->hpal) {
		SelectPalette (info->hdc, cman->hpal, FALSE);
	}
	
	SetBkMode (info->hdc, TRANSPARENT);
	RealizePalette (info->hdc);
}


/*
 *	Finish drawing. This may be used to free the resources
 *	allocated by .._BeginDraw
 */

void
OPaC_GrailPort_EndDraw (OPaC_GrPort* info)
{
	OP_FUNCTION (OPaC_GrailPort_EndDraw, 1);
	OP_ARG (1, OPaC_GrPort*, info);
}



/*
 *	Flush the graphical output. This forces the drawing
 *	operations to be executed. This will not display the
 *	result on the screen, because of the double buffering.
 */

void
OPaC_GrailPort_Flush (OPaC_GrPort* info)
{
	OP_FUNCTION (OPaC_GrailPort_Flush, 1);
	OP_ARG (1, OPaC_GrPort*, info);
}



/*
 *	Define a clipping rectangle. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetClipRect (OPaC_GrPort* info, int x1, int y1, int x2, int y2)
{
	OP_FUNCTION (OPaC_GrailPort_SetClipRect, 5);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, int, x1);
	OP_ARG (3, int, y1);
	OP_ARG (4, int, x2);
	OP_ARG (5, int, y2);
	
	SetRectRgn (info->hrgn, x1, y1, x2 + 1, y2 + 1);
	SelectClipRgn (info->hdc, info->hrgn);
}



/*
 *	Define the foreground color. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetForeColor (OPaC_GrPort* info, OPColorPixel pixel, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetForeColor, 3);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPColorManIn, cman);
	
	//	Make the brush...
	
	if (cman->CountColors () > 20) {
		
		//	Create a logical brush (solid color, no ugly WINDOWS
		//	created pattern !
		
		LOGBRUSH lb;
		HBRUSH hbr, hbr_old;
		
		lb.lbStyle = BS_SOLID;
		lb.lbColor = (COLORREF)(pixel);
		lb.lbHatch = 0;
		
		hbr     = CreateBrushIndirect (&lb);
		hbr_old = SelectBrush (info->hdc, hbr);
		
		SetTextColor (info->hdc, (COLORREF)(pixel));
		
		if (info->hbdel) DeleteBrush (info->hbrush);
		
		info->hbrush = hbr;
		info->hbdel  = TRUE;
		
	} else {
		
		//	We can't rely on WINDOWS to produce nice patterns for the missing
		//	gray levels. Therefore, we just use the stock patterns.
		
		int y = ( GetRValue (pixel) * 3
				+ GetGValue (pixel) * 6
				+ GetBValue (pixel) ) / 256;
		
		HBRUSH hbr;
		
		if (y < 2)
			hbr = GetStockBrush (BLACK_BRUSH);
		else if (y < 4)
			hbr = GetStockBrush (DKGRAY_BRUSH);
		else if (y < 6)
			hbr = GetStockBrush (GRAY_BRUSH);
		else if (y < 8)
			hbr = GetStockBrush (LTGRAY_BRUSH);
		else
			hbr = GetStockBrush (WHITE_BRUSH);
		
		SetTextColor (info->hdc, (COLORREF)(pixel));
		SelectBrush (info->hdc, hbr);
		
		if (info->hbdel) DeleteBrush (info->hbrush);
		
		info->hbrush = hbr;
		info->hbdel  = FALSE;
	}
}


/*
 *	Define the background color. This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetBackColor (OPaC_GrPort* info, OPColorPixel pixel, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetBackColor, 3);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPColorManIn, cman);
	
	SetBkColor (info->hdc, (COLORREF)(pixel));
}


/*
 *	Define the line style.  This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetLineStyle (OPaC_GrPort* info, OPColorPixel pixel, OPLineStyle style, OPColorManIn cman)
{
	OP_FUNCTION (OPaC_GrailPort_SetLineStyle, 4);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPColorPixel, pixel);
	OP_ARG (3, OPLineStyle, style);
	OP_ARG (4, OPColorManIn, cman);
	
	const int pen_style[] = { PS_SOLID, PS_DOT, PS_DASH };
	LOGBRUSH  brush       = { BS_SOLID, (COLORREF)(pixel), 0 };
	int       pen_s       = pen_style[style];
	
	if ( (info->fgcol == pixel)
	  && (info->hpen) ) {
		info->style = pen_s;
		return;
	}
	
	//	Since Win32 does not draw dashed and dotted lines the way I want, I
	//	use a solid brush... and LineDDA to select the proper pixels.
	
	HPEN hpen      = ExtCreatePen (PS_COSMETIC | PS_SOLID, 1, &brush, 0, 0);
	HPEN hpen_old  = SelectPen (info->hdc, hpen);
	
	if (info->hpdel) DeletePen (info->hpen);
	
	info->hpen  = hpen;
	info->hpdel = TRUE;
	info->fgcol = pixel;
	info->style = pen_s;
}


/*
 *	Following ugly code is needed since there is no way to draw the
 *	proper lines using GDI. We have to determine ourself which pixel
 *	to turn on in dotted and dashed lines.
 */

struct paramLineDDA
{
	OPaC_GrPort*	info;
	int				count;
};

static void CALLBACK
funcLineDDADot (int x, int y, LPARAM data)
{
	paramLineDDA* pdata = (paramLineDDA*)(data);
	
	if (pdata->count == 0) {
		SetPixelV (pdata->info->hdc, x, y, pdata->info->fgcol);
		pdata->count = 1;
	} else {
		pdata->count = 0;
	}
}

static void CALLBACK
funcLineDDADash (int x, int y, LPARAM data)
{
	paramLineDDA* pdata = (paramLineDDA*)(data);
	
	if (pdata->count < 2) {
		SetPixelV (pdata->info->hdc, x, y, pdata->info->fgcol);
		pdata->count++;
	} else {
		pdata->count++;
		if (pdata->count == 4) {
			pdata->count = 0;
		}
	}
}

static void
DrawLine (OPaC_GrPort* info, int x1, int y1, int x2, int y2)
{
	if (info->style == PS_SOLID) {
		MoveToEx (info->hdc, x1, y1, NULL);
		LineTo (info->hdc, x2, y2);
	} else if (info->style == PS_DOT) {
		paramLineDDA param = { info, 0 };
		LineDDA (x1, y1, x2, y2, funcLineDDADot, (LPARAM)(&param));
	} else if (info->style == PS_DASH) {
		paramLineDDA param = { info, 0 };
		LineDDA (x1, y1, x2, y2, funcLineDDADash, (LPARAM)(&param));
	}
}

static void
DrawRect (OPaC_GrPort* info, int x1, int y1, int x2, int y2)
{
	if (info->style == PS_SOLID) {
		MoveToEx (info->hdc, x1, y1, NULL);
		LineTo (info->hdc, x2, y1);
		LineTo (info->hdc, x2, y2);
		LineTo (info->hdc, x1, y2);
		LineTo (info->hdc, x1, y1);
	} else if (info->style == PS_DOT) {
		paramLineDDA param = { info, 0 };
		LineDDA (x1, y1, x2, y1, funcLineDDADot, (LPARAM)(&param));
		LineDDA (x2, y1, x2, y2, funcLineDDADot, (LPARAM)(&param));
		LineDDA (x2, y2, x1, y2, funcLineDDADot, (LPARAM)(&param));
		LineDDA (x1, y2, x1, y1, funcLineDDADot, (LPARAM)(&param));
	} else if (info->style == PS_DASH) {
		paramLineDDA param = { info, 0 };
		LineDDA (x1, y1, x2, y1, funcLineDDADash, (LPARAM)(&param));
		LineDDA (x2, y1, x2, y2, funcLineDDADash, (LPARAM)(&param));
		LineDDA (x2, y2, x1, y2, funcLineDDADash, (LPARAM)(&param));
		LineDDA (x1, y2, x1, y1, funcLineDDADash, (LPARAM)(&param));
	}
}


/*
 *	Define the current font.  This will be called only
 *	between a ..BeginDraw and a ..EndDraw.
 */

void
OPaC_GrailPort_SetFont (OPaC_GrPort* info, OPaC_GrFont* font_info)
{
	OP_FUNCTION (OPaC_GrailPort_SetFont, 2);
	OP_ARG (1, OPaC_GrPort*, info);
	OP_ARG (2, OPaC_GrFont*, font_info);
	
	HFONT hfont = (font_info && font_info->hfont)
	              ? font_info->hfont
	              : GetStockFont (ANSI_FIXED_FONT);
	
//	if (info->hfdel) DeleteFont (info->hfont);
	
	SelectFont (info->hdc, hfont);
	
	info->hfont = hfont;
	info->hfdel = FALSE;		//	never delete the font (allocated by font object)
}



/****************************************************************
 *																*
 *																*
 *			GRAIL PORT CLASS : TARGET SPECIFIC METHODS			*
 *																*
 *																*
 ****************************************************************/


/*
 *	Draw just one pixel.
 */

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y)
{
	OP_METHOD (DrawDot, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	SetPixelV (this->info->hdc, (int)(x), (int)(y), (COLORREF)(this->fore_color));
}

void
OPGrailPort::DrawDot (OPCoord x, OPCoord y, OPColorPixel color)
{
	OP_METHOD (DrawDot, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPColorPixel, color);
	
	SetPixelV (this->info->hdc, (int)(x), (int)(y), (COLORREF)(color));
}


/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line length of 0 won't draw anything !
 */

void
OPGrailPort::DrawLineDX (OPCoord x, OPCoord y, OPCoord dx)
{
	OP_METHOD (DrawLineDX, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (dx != 0) {
		DrawLine (this->info, (int)(x), (int)(y), (int)(x+dx), (int)(y));
	}
}

void
OPGrailPort::DrawLineDY (OPCoord x, OPCoord y, OPCoord dy)
{
	OP_METHOD (DrawLineDY, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (dy != 0) {
		DrawLine (this->info, (int)(x), (int)(y), (int)(x), (int)(y+dy));
	}
}



/*
 *	Draw a line (segment), using one of the predefined dash
 *	styles. The segment is always hair thin... The last dot
 *	will be drawn.
 *
 *	A line with both extremities at the same location will
 *	draw just a single dot.
 */

void
OPGrailPort::DrawLineToX (OPCoord x1, OPCoord y1, OPCoord x2)
{
	OP_METHOD (DrawLineToX, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (x2 >= x1) {
		DrawLine (this->info, (int)(x1), (int)(y1), (int)(x2+1), (int)(y1));
	} else {
		DrawLine (this->info, (int)(x1), (int)(y1), (int)(x2-1), (int)(y1));
	}
}

void
OPGrailPort::DrawLineToY (OPCoord x1, OPCoord y1, OPCoord y2)
{
	OP_METHOD (DrawLineToY, 3);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if (y2 >= y1) {
		DrawLine (this->info, (int)(x1), (int)(y1), (int)(x1), (int)(y2+1));
	} else {
		DrawLine (this->info, (int)(x1), (int)(y1), (int)(x1), (int)(y2-1));
	}
}


/*
 *	Draw rectangular box. The bottom left corner will be
 *	drawn at [x2;y2]. This might not be the case on every
 *	system, so be careful to implent this the OPaC way.
 */

void
OPGrailPort::DrawRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (DrawRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	::DrawRect (this->info, (int)x1, (int)y1, (int)x2, (int)y2);
}

void
OPGrailPort::DrawRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	::DrawRect (this->info, (int)x1, (int)y1, (int)x2, (int)y2);
}



/*
 *	Fill a rectangular box. The surface drawn is exactly
 *	the same as with DrawRect. Beware, this is usually not
 *	the case on other systems !!!
 *
 *	BE CAREFUL TO IMPLEMENT THIS METHOD THE RIGHT WAY !
 */

void
OPGrailPort::FillRect (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (FillRect, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	static HPEN pen = NULL;
	if (pen == NULL) pen = GetStockPen (NULL_PEN);
	
	HPEN old_pen = SelectPen (this->info->hdc, pen);
	Rectangle (this->info->hdc, (int)(x1), (int)(y1), (int)(x2 + 2), (int)(y2 + 2));
	SelectPen (this->info->hdc, old_pen);
}

void
OPGrailPort::FillRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (FillRect, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	static HPEN pen = NULL;
	if (pen == NULL) pen = GetStockPen (NULL_PEN);
	
	HPEN old_pen = SelectPen (this->info->hdc, pen);
	Rectangle (this->info->hdc, (int)(x1), (int)(y1), (int)(x2 + 2), (int)(y2 + 2));
	SelectPen (this->info->hdc, old_pen);
}


/*
 *	Draw a line segment. The segment is always hair thin and
 *	is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawHairLine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawHairLine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawHairLineTo (x, y, x+dx, y+dy);
}


void
OPGrailPort::DrawHairLineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawHairLineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	
	if (fabs (x2-x1) >= fabs (y2-y1)) {
		
		//	Mostly horizontal line...
		
		if (x1 > x2) {
			OPCoord xt = x2;
			OPCoord yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Card32  x    = (Card32)(x1);
		Card32  y    = (Card32)(y1);
		Card32  dx   = (Card32)(x2-x1);
		Int32   dy   = (Int32)(y2-y1);
		Card32* dots = g_pos_array;
		
		if (dx == 0) {
			SetPixelV (this->info->hdc, x, y, (COLORREF)(this->fore_color));
			return;
		}
		
		OP_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDXDots (x, y, dx, dy, dots, 0, 0.5);
		
		do {
			SetPixelV (this->info->hdc, x++, *dots++, (COLORREF)(this->fore_color));
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line...
		
		if (y1 > y2) {
			OPCoord xt = x2;
			OPCoord yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Card32  x    = (Card32)(x1);
		Card32  y    = (Card32)(y1);
		Int32   dx   = (Int32)(x2-x1);
		Card32  dy   = (Card32)(y2-y1);
		Card32* dots = g_pos_array;
		
		if (dy == 0) {
			SetPixelV (this->info->hdc, x, y, (COLORREF)(this->fore_color));
			return;
		}
		
		OP_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDYDots (x, y, dx, dy, dots, 0, 0.5);
		
		do {
			SetPixelV (this->info->hdc, *dots++, y++, (COLORREF)(this->fore_color));
		} while (dy--);
	}
}


/*
 *	Draw a line segment with anti-aliasing. The segment is always
 *	hair thin and is not dashed... The last dot will be drawn.
 */

void
OPGrailPort::DrawAALine (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy)
{
	OP_METHOD (DrawAALine, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPCoord, dy);
	
	this->DrawAALineTo (x, y, x+dx, y+dy);
}

void
OPGrailPort::DrawAALineTo (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (DrawAALineTo, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	
	if (fabs (x2-x1) >= fabs (y2-y1)) {
		
		//	Mostly horizontal line...
		
		if (x1 > x2) {
			OPCoord xt = x2;
			OPCoord yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Card32  x     = (Card32)(x1);
		Card32  y     = (Card32)(y1);
		Card32  dx    = (Card32)(x2-x1);
		Int32   dy    = (Int32)(y2-y1);
		Card32* dots  = g_pos_array;
		Card8*  error = g_err_array;
		
		if (dx == 0) {
			SetPixelV (this->info->hdc, x, y, (COLORREF)(this->fore_color));
			return;
		}
		
		OP_ASSERT (dx < MAX_LINE_LENGTH);
		
		this->ComputeLineDXDots (x, y, dx, dy, dots, error);
		
		do {
			Card8    i_on  = *error++;
			Card8    i_off = 255 - i_on;
			COLORREF c_on  = PALETTERGB (i_on,  i_on,  i_on);
			COLORREF c_off = PALETTERGB (i_off, i_off, i_off);
			SetPixelV (this->info->hdc, x, *dots, c_on);
			if (i_on) SetPixelV (this->info->hdc, x, *dots + 1, c_off);
			x++;
			dots++;
		} while (dx--);
		
	} else {
		
		//	Mostly vertical line...
		
		if (y1 > y2) {
			OPCoord xt = x2;
			OPCoord yt = y2;
			x2 = x1; y2 = y1;
			x1 = xt; y1 = yt;
		}
		
		Card32  x     = (Card32)(x1);
		Card32  y     = (Card32)(y1);
		Int32   dx    = (Int32)(x2-x1);
		Card32  dy    = (Card32)(y2-y1);
		Card32* dots  = g_pos_array;
		Card8*  error = g_err_array;
		
		if (dy == 0) {
			SetPixelV (this->info->hdc, x, y, (COLORREF)(this->fore_color));
			return;
		}
		
		OP_ASSERT (dy < MAX_LINE_LENGTH);
		
		this->ComputeLineDYDots (x, y, dx, dy, dots, error);
		
		do {
			Card8    i_on  = *error++;
			Card8    i_off = 255 - i_on;
			COLORREF c_on  = PALETTERGB (i_on,  i_on,  i_on);
			COLORREF c_off = PALETTERGB (i_off, i_off, i_off);
			SetPixelV (this->info->hdc, *dots, y, c_on);
			if (i_on) SetPixelV (this->info->hdc, *dots + 1, y, c_off);
			y++;
			dots++;
		} while (dy--);
	}
}


/*
 *	Draw a piece of text, which will be extracted from a
 *	string starting at character "start" and up to "length"
 *	characters. The way to position the text is specified
 *	by "pos" and "dx" is used if aligned positioning has
 *	been requested.
 */

void
OPGrailPort::ShowText (OPCoord x, OPCoord y, OPStringIn text,
					   Count start, Count length,
					   Card8 pos,
					   OPCoord dx)
{
	OP_METHOD (ShowText, 7);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPStringIn, text);
	OP_ARG (4, Count, start);
	OP_ARG (5, Count, length);
	OP_ARG (6, Card8, pos);
	OP_ARG (7, OPCoord, dx);
	OP_ASSERT (this->is_drawing == TRUE);
	
	Count len = text->CountCharacters ();
	
	if (start >= len) {
		
		//	Do not draw anything, since the start of the text in the string is
		//	past the end of the string.
		
		return;
	}
	
	if ( (len < (start + length))
	  || (length == 0) ) {
		
		//	User is asking for a too long piece of text, longer than it really
		//	can be.
		
		length = len - start;
	}
	
	LatinChar* buffer_alloc = 0;
	LatinChar* buffer = 0;
	
	if (length < BUFFER_LENGTH) {
		
		//	Use the buffer from the stack.
		
		buffer = g_buffer_array;
		
	} else {
		
		//	Use the buffer from the temporary memory.
		
		buffer = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_temp_zone, LatinChar, length);
		buffer_alloc = buffer;
	}
	
	text->ExtractStringCopy (buffer, start, length);
	
	const int hpos[] = { TA_LEFT, TA_CENTER, TA_RIGHT, 0 };
	const int vpos[] = { TA_BASELINE, TA_TOP, 0, TA_BOTTOM };
	
	int hi = (pos & OP_TEXT_HPOS_MASK);
	int vi = (pos & OP_TEXT_VPOS_MASK);
	
	if (hi == OP_TEXT_HPOS_ALIGN) {
		
		//	Text should be aligned on both margins.
		//	NIY
		
		hi = OP_TEXT_HPOS_CENTER;
	}
	
	if (vi == OP_TEXT_VPOS_CENTER) {
		
		SIZE size;
		GetTextExtentPoint32 (this->info->hdc, buffer, (int)(length), &size);
		
		y -= (OPCoord)(size.cy) / 2;
		vi = OP_TEXT_VPOS_TOP;
	}
	
	if (hi == OP_TEXT_HPOS_CENTER) {
		
		SIZE size;
		GetTextExtentPoint32 (this->info->hdc, buffer, (int)(length), &size);
		
		x -= (OPCoord)(size.cx) / 2;
		hi = OP_TEXT_HPOS_LEFT;
	}
	
	SetTextAlign (this->info->hdc, hpos[hi >> OP_TEXT_HPOS_SHIFT] | vpos[vi >> OP_TEXT_VPOS_SHIFT] | TA_NOUPDATECP);
	TextOut (this->info->hdc, (int)(x), (int)(y), buffer, (int)(length));
	
	if (buffer_alloc) {
		
		//	Free the memory if we need to.
		
		OP_FREE_FROM_ZONE (OPaC_temp_zone, buffer_alloc);
	}
}



/*
 *	Draw a piece of text in the specified rectangle. This
 *	uses the moreShowText
 */

void
OPGrailPort::ShowText (OPRectIn box, OPStringIn text,
					   Count start, Count length, Card8 pos, OPRelCoord margin)
{
	OP_METHOD (ShowText, 6);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPStringIn, text);
	OP_ARG (3, Count, start);
	OP_ARG (4, Count, length);
	OP_ARG (5, Card8, pos);
	OP_ARG (6, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ShowText (x1, y1, x2, y2, text, start, length, pos);
}

void
OPGrailPort::ShowText (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
					   OPStringIn text,
					   Count start, Count length,
					   Card8 pos)
{
	OP_METHOD (ShowText, 8);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPStringIn, text);
	OP_ARG (6, Count, start);
	OP_ARG (7, Count, length);
	OP_ARG (8, Card8, pos);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( (x2 <= x1) || (y2 <= y1) ) {
		return;
	}
	
	OPCoord x = x1;
	OPCoord y = y1;
	
	switch (pos & OP_TEXT_VPOS_MASK) {
		
		case OP_TEXT_VPOS_TOP:
			y = y1;
			break;
		
		case OP_TEXT_VPOS_CENTER:
			y = (y1 + y2 + 1) / 2;
			break;
			
		case OP_TEXT_VPOS_BOTTOM:
			y = y2;
			break;
	}
	
	switch (pos & OP_TEXT_HPOS_MASK) {
		
		case OP_TEXT_HPOS_LEFT:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_CENTER:
			x = (x1 + x2 + 1) / 2;
			break;
		
		case OP_TEXT_HPOS_ALIGN:
			x = x1;
			break;
		
		case OP_TEXT_HPOS_RIGHT:
			x = x2;
			break;
	}
	
	this->ShowText (x, y, text, start, length, pos, x2-x1+1);
}



/*
 *	Scroll a rectangular area of the Grail Port. This will
 *	not erase the invalidated parts of the picture.
 */

void
OPGrailPort::ScrollRect (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2,
						 OPRelCoord move_x, OPRelCoord move_y)
{
	OP_METHOD (ScrollRect, 6);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	OP_ARG (5, OPRelCoord, move_x);
	OP_ARG (6, OPRelCoord, move_y);
	OP_ASSERT (this->is_drawing == TRUE);
	
	RECT scroll_rect = { (int)(x1), (int)(y1), (int)(x2 + 1), (int)(y2 + 1) };
	RECT clip_rect   = { (int)(this->rect_x1), (int)(this->rect_y1),
						 (int)(this->rect_x2 + 1), (int)(this->rect_y2 + 1) };
	
	ScrollDC (this->info->hdc, (int)(move_x), (int)(move_y),
			  &scroll_rect, &clip_rect, 0, 0);
}

void
OPGrailPort::ScrollRect (OPRectIn box, OPRelCoord move_x, OPRelCoord move_y,
						 OPRelCoord margin)
{
	OP_METHOD (ScrollRect, 4);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, move_x);
	OP_ARG (3, OPRelCoord, move_y);
	OP_ARG (4, OPRelCoord, margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(box.DX ()) < (2*margin))
	  || ((OPRelCoord)(box.DY ()) < (2*margin)) ) {
		return;
	}
	
	OPCoord x1 = box.X1 () + margin;
	OPCoord y1 = box.Y1 () + margin;
	OPCoord x2 = box.X2 () - margin;
	OPCoord y2 = box.Y2 () - margin;
	
	this->ScrollRect (x1, y1, x2, y2, move_x, move_y);
}



/*
 *	Copy a rectangle from the source port to the current
 *	port, into the specified box, from the source position
 *	given by its top/left corner.
 *
 *	NB: This will copy the pixels, including the one of
 *		the bottom right corner. This is still the same
 *		policy as with the rectangle.
 */

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPCoord dest_x1, OPCoord dest_y1,
					   OPCoord dest_x2, OPCoord dest_y2)
{
	OP_METHOD (CopyRect, 7);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPCoord, dest_x1);
	OP_ARG (5, OPCoord, dest_y1);
	OP_ARG (6, OPCoord, dest_x2);
	OP_ARG (7, OPCoord, dest_y2);
	OP_ASSERT (this->is_drawing == TRUE);
	
	OPGrailPort* src = source.ReturnPtr ();
	
	if (this->use_transparency) {
		
		if (FALSE /*::GetDeviceCaps (this->info->hdc, CAPS1) & C1_TRANSPARENT*/) {
			
			//	Special transparency background mode is supported by this display
			//	driver. Hope most do it, since the pixel by pixel operation is
			//	very slow...
			
			int      old_mode = SetBkMode (this->info->hdc, NEWTRANSPARENT);
			COLORREF rgb_back = SetBkColor (this->info->hdc, this->transparent_color);
			
			//	Actual blt is a simple source copy; transparency is automatic.
			
			BitBlt (this->info->hdc,
					(int)(dest_x1), (int)(dest_y1),
					(int)(dest_x2 - dest_x1 + 1), (int)(dest_y2 - dest_y1 + 1),
					src->info->hdc,
					(int)(source_x), (int)(source_y),
					SRCCOPY);
			
			SetBkColor (this->info->hdc, rgb_back);
			SetBkMode (this->info->hdc, old_mode);
			return;
			
		} else {
		    
			int dx = (int)(dest_x2 - dest_x1 + 1);
			int dy = (int)(dest_y2 - dest_y1 + 1);
			int ox = (int)(dest_x1);
			int oy = (int)(dest_y1);
			int sx = (int)(source_x);
			int sy = (int)(source_y);
			
			HDC      hdc_mask = CreateCompatibleDC (src->info->hdc);
			HBITMAP  hbm_mask = CreateBitmap (dx, dy, 1, 1, NULL);
			COLORREF rgb_back = SetBkColor (src->info->hdc, this->transparent_color);
			COLORREF rgb_bk = SetBkColor (this->info->hdc, RGB(255, 255, 255));
			COLORREF rgb_tx = SetTextColor (this->info->hdc, RGB(0, 0, 0));
			
			//	Create the monochrome mask based on the transparent color...
			
			SelectObject (hdc_mask, hbm_mask);
			
			BitBlt (hdc_mask, 0, 0, dx, dy, src->info->hdc, sx, sy, SRCCOPY);
			BitBlt (this->info->hdc, ox, oy, dx, dy, src->info->hdc, sx, sy, SRCINVERT);
			BitBlt (this->info->hdc, ox, oy, dx, dy, hdc_mask, 0, 0, SRCAND);
			BitBlt (this->info->hdc, ox, oy, dx, dy, src->info->hdc, sx, sy, SRCINVERT);
			
//			SelectObject (hdc_mask, NULL);
			DeleteObject (hbm_mask);
			DeleteDC (hdc_mask);
			
			SetBkColor (src->info->hdc, rgb_back);
			SetBkColor (this->info->hdc, rgb_bk);
			SetTextColor (this->info->hdc, rgb_tx);
			
			return;
		}
		
		//	If transparent drawing is required, do it pixel by pixel. Doing it by
		//	hand is not fast but at least it works... Should consider using MaskBlt.
		
		int dy = (int)(dest_y1);
		int sy = (int)(source_y);
		int h  = (int)(dest_y2 - dest_y1 + 1);
		
		while (h--) {
			int dx = (int)(dest_x1);
			int sx = (int)(source_x);
			int w  = (int)(dest_x2 - dest_x1 + 1);
			while (w--) {
				
				OPColorPixel pixel = GetPixel (src->info->hdc, sx, sy);
				
				if (pixel != this->transparent_color) {
					SetPixelV (this->info->hdc, dx, dy, pixel);
				}
				
				sx++;
				dx++;
			}
			sy++;
			dy++;
		}

		return;
	}
	
	//	BitBlt needs to operate on compatible bitmaps !!! Why ?
	
	BitBlt (this->info->hdc,
			(int)(dest_x1), (int)(dest_y1),
			(int)(dest_x2 - dest_x1 + 1), (int)(dest_y2 - dest_y1 + 1),
			src->info->hdc,
			(int)(source_x), (int)(source_y),
			SRCCOPY);
}

void
OPGrailPort::CopyRect (OPGrailPortIn source,
					   OPCoord source_x, OPCoord source_y,
					   OPRectIn dest_box, OPRelCoord dest_margin)
{
	OP_METHOD (CopyRect, 5);
	OP_ARG (1, OPGrailPortIn, source);
	OP_ARG (2, OPCoord, source_x);
	OP_ARG (3, OPCoord, source_y);
	OP_ARG (4, OPRectIn, dest_box);
	OP_ARG (5, OPRelCoord, dest_margin);
	OP_ASSERT (this->is_drawing == TRUE);
	
	if ( ((OPRelCoord)(dest_box.DX ()) < (2*dest_margin))
	  || ((OPRelCoord)(dest_box.DY ()) < (2*dest_margin)) ) {
		return;
	}
	
	OPCoord x1 = dest_box.X1 () + dest_margin;
	OPCoord y1 = dest_box.Y1 () + dest_margin;
	OPCoord x2 = dest_box.X2 () - dest_margin;
	OPCoord y2 = dest_box.Y2 () - dest_margin;
	
	this->CopyRect (source, source_x, source_y, x1, y1, x2, y2);
}



/*
 *	Set a line of pixels with the specified pixel values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetColorPixels (OPCoord x, OPCoord y, OPCoord dx,
							 const OPColorPixel* pixels)
{
	OP_METHOD (SetGrayPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const OPColorPixel*, pixels);
	OP_ASSERT (this->is_drawing == TRUE);
	
	Count delta = (Count)(dx);

	while (delta--) {
		SetPixelV (this->info->hdc, (int)(x), (int)(y), (COLORREF)(*pixels++));
		x += 1;
	}
}



/*
 *	Set a line of pixels with the specified gray values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetGrayPixels (OPCoord x, OPCoord y, OPCoord dx,
							const Card8* intensity)
{
	OP_METHOD (SetGrayPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, cont Card8*, intensity);
	OP_ASSERT (this->is_drawing == TRUE);
	
	this->SetRGBPixels (x, y, dx, intensity, intensity, intensity);
}



/*
 *	Set a line of pixels with the specified RGB values. This
 *	is not very fast, but provides an easy way to build some
 *	pixel maps yourself.
 */

void
OPGrailPort::SetRGBPixels (OPCoord x, OPCoord y, OPCoord dx,
						   const Card8* red,
						   const Card8* green,
						   const Card8* blue)
{
	OP_METHOD (SetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, const Card8*, red);
	OP_ARG (5, const Card8*, green);
	OP_ARG (6, const Card8*, blue);
	OP_ASSERT (this->is_drawing == TRUE);
	
	Count delta = (Count)(dx);

	if (delta > 4) {
		
		//	If there are enough bytes to set, use the DIB interface, since the
		//	overhead will be smaller than the sum of the individual calls made
		//	by the simple "SetPixel" method.

		BITMAPINFO info;
		
		info.bmiHeader.biSize          = sizeof (info.bmiHeader);
		info.bmiHeader.biWidth         = (LONG)(delta);
		info.bmiHeader.biHeight        = -1;
		info.bmiHeader.biPlanes        = 1;
		info.bmiHeader.biBitCount      = 24;
		info.bmiHeader.biCompression   = BI_RGB;
		info.bmiHeader.biSizeImage     = 0;
		info.bmiHeader.biXPelsPerMeter = 0;
		info.bmiHeader.biYPelsPerMeter = 0;
		info.bmiHeader.biClrUsed       = 0;
		info.bmiHeader.biClrImportant  = 0;
		
		Card32 width = (Card32)(delta);
		Card8* rgb   = OP_ALLOC_ARRAY (Card8, 3*width);

		if (rgb) {
			
			Card8* ptr = rgb;
			
			while (width--) {
				*ptr++ = *blue++;
				*ptr++ = *green++;
				*ptr++ = *red++;
			}
			
			SetDIBitsToDevice (this->info->hdc, (int)(x), (int)(y), (DWORD)(delta), (DWORD)(1),
							   0, 0, 0, 1, rgb, &info, DIB_RGB_COLORS);
			
			OP_FREE (rgb);
			return;
		}
	}
	
	//	Same remark as for the monochrome version of SetPixels.
	
	while (delta--) {
		SetPixelV (this->info->hdc, (int)(x), (int)(y),
				   PALETTERGB (red[0], green[0], blue[0]));
		x++; red++; green++; blue++;
	}
}


/*
 *	Set a block of pixels with the specified monochrome values.
 *	The pixels set to "1" will appear as black on screen.
 */

void
OPGrailPort::SetBitmap (OPCoord x, OPCoord y, OPCoord dx, OPCoord dy, const Card8* bits)
{
	OP_METHOD (SetBitmap, 5);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ARG (5, cont Card8*, bits);
	OP_ASSERT (this->is_drawing == TRUE);
	
	struct {
		BITMAPINFO info;			//	contains one RGBQUAD element
		RGBQUAD    quad[1];			//	additional one needed (a total of two are required)
	} i;
	
	i.info.bmiHeader.biSize          = sizeof (i.info.bmiHeader);
	i.info.bmiHeader.biWidth         = (LONG)(dx);
	i.info.bmiHeader.biHeight        = -(LONG)(dy);
	i.info.bmiHeader.biPlanes        = 1;
	i.info.bmiHeader.biBitCount      = 1;
	i.info.bmiHeader.biCompression   = BI_RGB;
	i.info.bmiHeader.biSizeImage     = 0;
	i.info.bmiHeader.biXPelsPerMeter = 0;
	i.info.bmiHeader.biYPelsPerMeter = 0;
	i.info.bmiHeader.biClrUsed       = 0;
	i.info.bmiHeader.biClrImportant  = 0;
	
	const RGBQUAD rgb_black = { 0x00, 0x00, 0x00, 0x00 };
	const RGBQUAD rgb_white = { 0xFF, 0xFF, 0xFF, 0x00 };
	
	i.info.bmiColors[0] = rgb_white;
	i.info.bmiColors[1] = rgb_black;
	
	SetDIBitsToDevice (this->info->hdc,
					   (int)(x), (int)(y),			//	origin of destination
					   (DWORD)(dx), (DWORD)(dy),	//	source rectangle size
					   0, 0,						//	origin of source
					   0,							//	first scan line
					   (DWORD)(dy),					//	number of scan lines
					   bits,						//	bitmap (DIB bits)
					   &i.info,						//	bitmap info for color mapping
					   DIB_RGB_COLORS);				//	color table contains RGB values
}


