/*
 *	adorner.cxx
 *
 *	The adorner is used to help drawing parts of the interface,
 *	in collaboration with the Look Manager.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>

OP_IMPLEMENTATION (cell.h);
OP_IMPLEMENT_CLASS (OPAdorner);

/****************************************************************/

/*
 *	Initialise an adorner instance. This will also handle
 *	the first, global, initialization.
 */

OPAdorner::OPAdorner ()
{
	OP_METHOD (OPAdorner, 0);
	
	this->colors = 0;
}


/*
 *	Destroy an adorner instance. This will free the color
 *	table which had been built by "Initialise".
 */

OPAdorner::~OPAdorner ()
{
	OP_METHOD (~OPAdorner, 0);
	
	if (this->colors) {
		OP_FREE (this->colors);
	}
}


/*
 *	The adorner can't be stored, since the associated port
 *	can't either.
 */

void
OPAdorner::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Cannot store Adorner");
}


/*
 *	The adorner can't be restored, since the associated port
 *	can't either.
 */

void
OPAdorner::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	OP_ABORT ("Cannot restore Adorner");
}


void
OPAdorner::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->port = 0;
	
	super::Delete ();
}


void
OPAdorner::Initialise (OPGrailPortIn port)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ASSERT (port.IsValid ());
	OP_ASSERT (this->colors == 0);
	
	this->port    = port;
	this->lookman = OPLookMan::Current ();
	this->colman  = port->ReturnColorMan ();
	this->colors  = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, OPColorPixel, OP_COLOR_NUM);
	
	Count index = 0;
	
	for (;;) {
		OPLookColor color = this->lookman->FindNthColor (index++);
		if (color == OP_COLOR_NONE) break;
		Card8 r, g, b;
		this->lookman->GetColorRGB (color, r, g, b);
		this->colors[color] = this->colman->FindColor (r, g, b);
	}
}


void
OPAdorner::SelectSysFont ()
{
	OP_METHOD (SelectSysFont, 0);
	this->port->SetFont (OPLookMan::Current ()->ReturnSysFont ());
}



/*
 *	Change a color definition.
 */

void
OPAdorner::DefineColor (OPLookColor color, OPColorPixel pixel)
{
	OP_METHOD (DefineColor, 2);
	OP_ARG (1, OPLookColor, color);
	OP_ARG (2, OPColorPixel, pixel);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		this->colors[color] = pixel;
	}
}


/*
 *	Return the pixel value for the specified color.
 */

OPColorPixel
OPAdorner::ReturnColor (OPLookColor color) const
{
	OP_METHOD (ReturnColor, 1);
	OP_ARG (1, OPLookColor, color);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		return this->colors[color];
	}
	
	return 0;
}


void
OPAdorner::DefineColor (OPAtomIn color, OPColorPixel pixel)
{
	OP_METHOD (DefineColor, 2);
	OP_ARG (1, OPAtomIn, color);
	OP_ARG (2, OPColorPixel, pixel);
	
	this->DefineColor (this->lookman->FindColor (color), pixel);
}

OPColorPixel
OPAdorner::ReturnColor (OPAtomIn color) const
{
	OP_METHOD (ReturnColor, 1);
	OP_ARG (1, OPAtomIn, color);
	
	return this->ReturnColor (this->lookman->FindColor (color));
}



void
OPAdorner::SetForeColor (OPLookColor color)
{
	OP_METHOD (SetForeColor, 1);
	OP_ARG (1, OPLookColor, color);
	OP_ASSERT (this->colors);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		OPColorPixel pixel = this->colors[color];
		this->port->SetForeColor (pixel);
	}
}


void
OPAdorner::SetBackColor (OPLookColor color)
{
	OP_METHOD (SetBackColor, 1);
	OP_ARG (1, OPLookColor, color);
	OP_ASSERT (this->colors);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		OPColorPixel pixel = this->colors[color];
		this->port->SetBackColor (pixel);
	}
}


void
OPAdorner::SetForeColor (OPAtomIn color)
{
	this->SetForeColor (this->lookman->FindColor (color));
}


void
OPAdorner::SetBackColor (OPAtomIn color)
{
	this->SetBackColor (this->lookman->FindColor (color));
}


void
OPAdorner::FillRect (OPRectIn rect, OPLookColor color, OPRelCoord margin)
{
	OP_METHOD (FillRect, 3);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPLookColor, color);
	OP_ARG (3, OPRelCoord, margin);
	OP_ASSERT (this->colors);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		OPColorPixel pixel = this->colors[color];
		this->port->SetForeColor (pixel);
		this->port->FillRect (rect, margin);
	}
}


void
OPAdorner::FillRect (OPRectIn rect, OPAtomIn color, OPRelCoord margin)
{
	OP_METHOD (FillRect, 3);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPAtomIn, color);
	OP_ARG (3, OPRelCoord, margin);
	OP_ASSERT (this->colors);
	
	this->FillRect (rect, this->lookman->FindColor (color), margin);
}


void
OPAdorner::DrawRectHigh (OPRectIn rect, OPRelCoord margin)
{
	OP_METHOD (DrawRectHigh, 2);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPRelCoord, margin);
	
//	this->port->SetForeColor (this->colors[OP_COLOR_WIDGET_FORE]);
//	this->port->FillRect (rect, margin);
	
	this->port->SetForeColor (this->colors[OP_COLOR_LGRAY]);
	this->port->DrawLineDX (rect.X1 (), rect.Y1 (), rect.DX ());
	this->port->DrawLineDY (rect.X1 (), rect.Y1 (), rect.DY ());
	
	this->port->SetForeColor (this->colors[OP_COLOR_WHITE]);
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y1 ()+1, rect.DX ()-2);
	this->port->DrawLineDY (rect.X1 ()+1, rect.Y1 ()+1, rect.DY ()-2);
	
	this->port->SetForeColor (this->colors[OP_COLOR_BLACK]);
	this->port->DrawLineDX (rect.X1 (), rect.Y2 (), rect.DX ());
	this->port->DrawLineDY (rect.X2 (), rect.Y1 (), rect.DY ());
	
	this->port->SetForeColor (this->colors[OP_COLOR_DGRAY]);
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y2 ()-1, rect.DX ()-2);
	this->port->DrawLineDY (rect.X2 ()-1, rect.Y1 ()+1, rect.DY ()-2);
}


void
OPAdorner::DrawRectLow (OPRectIn rect, OPRelCoord margin)
{
	OP_METHOD (DrawRectLow, 2);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPRelCoord, margin);
	
//	this->port->SetForeColor (this->colors[OP_COLOR_WIDGET_FORE]);
//	this->port->FillRect (rect, margin);
	
	this->port->SetForeColor (this->colors[OP_COLOR_LGRAY]);
	this->port->DrawLineDY (rect.X2 (), rect.Y1 (), rect.DY ());
	this->port->DrawLineDX (rect.X1 (), rect.Y2 (), rect.DX ());
	
	this->port->SetForeColor (this->colors[OP_COLOR_WHITE]);
	this->port->DrawLineDY (rect.X2 ()-1, rect.Y1 ()+1, rect.DY ()-2);
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y2 ()-1, rect.DX ()-2);
	
	this->port->SetForeColor (this->colors[OP_COLOR_BLACK]);
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y1 ()+1, rect.DX ()-2);
	this->port->DrawLineDY (rect.X1 ()+1, rect.Y1 ()+1, rect.DY ()-2);
	
	this->port->SetForeColor (this->colors[OP_COLOR_DGRAY]);
	this->port->DrawLineDX (rect.X1 (), rect.Y1 (), rect.DX ());
	this->port->DrawLineDY (rect.X1 (), rect.Y1 (), rect.DY ());
}


void
OPAdorner::DrawRectHigh (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OPRect box (x1, y1, x2, y2);
	this->DrawRectHigh (box);
}


void
OPAdorner::DrawRectLow (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OPRect box (x1, y1, x2, y2);
	this->DrawRectLow (box);
}


void
OPAdorner::DrawRect (OPRectIn rect, OPLookColor color, OPRelCoord margin)
{
	OP_METHOD (DrawRect, 3);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPLookColor, color);
	OP_ARG (3, OPRelCoord, margin);
	OP_ASSERT (this->colors);
	
	if ( (color >  OP_COLOR_NONE)
	  && (color <= OP_COLOR_MAX) ) {
		OPColorPixel pixel = this->colors[color];
		this->port->SetForeColor (pixel);
		this->port->DrawRect (rect, margin);
	}
}


void
OPAdorner::DrawRect (OPRectIn rect, OPAtomIn color, OPRelCoord margin)
{
	OP_METHOD (DrawRect, 3);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPAtomIn, color);
	OP_ARG (3, OPRelCoord, margin);
	OP_ASSERT (this->colors);
	
	this->DrawRect (rect, this->lookman->FindColor (color), margin);
}


void
OPAdorner::DrawFrame (OPRectIn rect, OPLookFrame frame, OPRelCoord margin)
{
	OP_METHOD (DrawFrame, 3);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPLookFrame, frame);
	OP_ARG (3, OPRelCoord, margin);

	OPCoord x  = rect.X1 () + margin;
	OPCoord y  = rect.Y1 () + margin;
	OPCoord dx = rect.DX () - 2 * margin;
	OPCoord dy = rect.DY () - 2 * margin;

	this->port->SetForeColor (this->colors[OP_COLOR_WHITE]);
	
	if (frame & OP_FRAME_TOP)    this->port->DrawLineDX (x, y, dx);
	if (frame & OP_FRAME_LEFT)   this->port->DrawLineDY (x, y, dy);
	if (frame & OP_FRAME_BOTTOM) this->port->DrawLineDX (x, y+dy-2, dx);
	if (frame & OP_FRAME_RIGHT)  this->port->DrawLineDY (x+dx-2, y, dy);

	this->port->SetForeColor (this->colors[OP_COLOR_BLACK]);

	if (frame & OP_FRAME_TOP)    this->port->DrawLineDX (x, y+1, dx);
	if (frame & OP_FRAME_LEFT)   this->port->DrawLineDY (x+1, y, dy);
	if (frame & OP_FRAME_BOTTOM) this->port->DrawLineDX (x, y+dy-1, dx);
	if (frame & OP_FRAME_RIGHT)  this->port->DrawLineDY (x+dx-1, y, dy);
}



void
OPAdorner::DrawFrame (OPRectIn rect, OPRelCoord margin)
{
	OP_METHOD (DrawFrame, 2);
	OP_ARG (1, OPRectIn, rect);
	OP_ARG (2, OPRelCoord, margin);

	OPCoord x  = rect.X1 () + margin;
	OPCoord y  = rect.Y1 () + margin;
	OPCoord dx = rect.DX () - 2 * margin;
	OPCoord dy = rect.DY () - 2 * margin;

	this->port->SetForeColor (this->colors[OP_COLOR_DGRAY]);
	
	this->port->DrawLineDX (x, y, dx-1);
	this->port->DrawLineDY (x, y, dy-1);
	this->port->DrawLineDX (x, y+dy-2, dx-1);
	this->port->DrawLineDY (x+dx-2, y, dy-1);

	this->port->SetForeColor (this->colors[OP_COLOR_WHITE]);

	this->port->DrawLineDX (x+1, y+1, dx-3);
	this->port->DrawLineDY (x+1, y+1, dy-3);
	this->port->DrawLineDX (x, y+dy-1, dx);
	this->port->DrawLineDY (x+dx-1, y, dy);
}



void
OPAdorner::DrawLineFrame (OPRectIn the_rect, OPRelCoord margin)
{
	OP_METHOD (DrawLineFrame, 2);
	OP_ARG (1, OPRectIn, the_rect);
	OP_ARG (2, OPRelCoord, margin);

	OPRect rect (the_rect, margin);
	
	this->port->SetForeColor (this->colors[OP_COLOR_DGRAY]);
	this->port->DrawLineDX (rect.X1 (), rect.Y1 (), rect.DX ());
	this->port->DrawLineDY (rect.X1 (), rect.Y1 (), rect.DY ());
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y1 ()+1, rect.DX ()-2);
	this->port->DrawLineDY (rect.X1 ()+1, rect.Y1 ()+1, rect.DY ()-2);
	
	this->port->SetForeColor (this->colors[OP_COLOR_WHITE]);
	this->port->DrawLineDX (rect.X1 (), rect.Y2 (), rect.DX ());
	this->port->DrawLineDY (rect.X2 (), rect.Y1 (), rect.DY ());
	
	this->port->SetForeColor (this->colors[OP_COLOR_LGRAY]);
	this->port->DrawLineDX (rect.X1 ()+1, rect.Y2 ()-1, rect.DX ()-2);
	this->port->DrawLineDY (rect.X2 ()-1, rect.Y1 ()+1, rect.DY ()-2);
}


void
OPAdorner::DrawIcon (OPCoord x, OPCoord y, OPLookIcon icon)
{
	OP_METHOD (DrawIcon, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPLookIcon, icon);
	
	const OPLookManRef&     look  = OPLookMan::Current ();
	const OPGrailPixmapRef& icons = look->ReturnIconPixmap (icon);
	
	if (icon == OP_ICON_NONE) {
		return;
	}
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	OPCoord ox = 0;
	OPCoord oy = 0;
	
	look->GetIconSize (icon, dx, dy);
	look->GetIconPos (icon, ox, oy);
	
	if (look->IsIconTransparent (icon)) {
		this->port->SetTransparency (look->ReturnIconTransparency (icon));
		this->port->SetTransparent (TRUE);
		this->port->CopyRect (icons, ox, oy, x, y, x+dx-1, y+dy-1);
		this->port->SetTransparent (FALSE);
	} else {
		this->port->CopyRect (icons, ox, oy, x, y, x+dx-1, y+dy-1);
	}
}


void
OPAdorner::DrawIcon (OPCoord x, OPCoord y, OPAtomIn icon)
{
	this->DrawIcon (x, y, OPLookMan::Current ()->FindIcon (icon));
}


/*
 *	Flush the graphical output according to the specified
 *	priority.
 */

void
OPAdorner::Flush (OPFlushPriority priority)
{
	OP_METHOD (Flush, 1);
	OP_ARG (1, OPFlushPriority, priority);
	
	switch (priority) {
		
		case OP_FLUSH_SYNC:
			this->port->Flush ();
			break;
		
		case OP_FLUSH_FAST:
			this->port->Flush ();
			break;
		
		default:
			break;
	}
}


void
OPAdorner::Update ()
{
}
