/*
 *	colorman.cxx
 *
 *	OPaC Grail color manager. This file implements the machine
 *	independent part, whereas "sd/../colman.cxx" implements the
 *	rest.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPColorMan);

/****************************************************************/

OPColorManRef	OPColorMan::system_man = 0;

/****************************************************************/


/*
 *	Save the color manager's state.
 */

void
OPColorMan::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::StoreInstance (storage);
	
	//	NIY
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the color manager's state.
 */

void
OPColorMan::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::RestoreInstance (storage);
	
	//	NIY
	
	storage >> OPStorageSync::mark;
}



/*
 *	Find the nearest matching pixel color for the specified
 *	RGB color. The components are in the 0..1 range.
 */

OPColorPixel
OPColorMan::FindColor (Real r, Real g, Real b) const
{
	OP_METHOD (FindColor, 3);
	OP_ARG (1, Real, r);
	OP_ARG (2, Real, g);
	OP_ARG (3, Real, b);
	
	Card16 rval = (Card16)(r * 256);
	Card16 gval = (Card16)(g * 256);
	Card16 bval = (Card16)(b * 256);
	if (rval > 255) rval = 255;
	if (gval > 255) gval = 255;
	if (bval > 255) bval = 255;
	return this->FindColor ( (Card8)(rval), (Card8)(gval), (Card8)(bval) );
}



/*
 *	Find the nearest matching pixel color for the specified
 *	RGB color. The components are in the 0..255 range.
 */

OPColorPixel
OPColorMan::FindColor (Card8 r, Card8 g, Card8 b) const
{
	OP_METHOD (FindColor, 3);
	OP_ARG (1, Card8, r);
	OP_ARG (2, Card8, g);
	OP_ARG (3, Card8, b);
	
	if ( (this->pal_size == 0)
	  || (this->pal_cube == 0) ) {
		return this->MakeTrueColor (r, g, b);
	}
	
	PaletteEntry* entry = & this->pal_entry[this->cube_pos];
	Count         count =   this->pal_cube;
	int           dist  = 256 * 10;
	Bool          gray  = (r == g) && (g == b) ? TRUE : FALSE;
	OPColorPixel  pixel = 0;
	
	while (count) {
		
		if (entry->match_enable == FALSE) {
			entry++;
			continue;
		}
		
		if (gray && (entry->gray == FALSE) ) {
			entry++;
			count--;
			continue;
		}
		
		int dr = (entry->r - r) * 3;	//	Y = 30% red
		int dg = (entry->g - g) * 6;	//	Y = 60% green
		int db = (entry->b - b) * 1;	//	Y = 10% blue
		
		if (dr < 0) dr = -dr;
		if (dg < 0) dg = -dg;
		if (db < 0) db = -db;
		
		int delta = dg;
		if (dr > delta) delta = dr;
		if (db > delta) delta = db;
		
		if (delta < dist) {
			
			dist  = delta;
			pixel = entry->pixel;
			
			if (dist == 0) break;
		}
		
		entry++;
		count--;
	}
	
	return pixel;
}



/*
 *	Find the nearest matching pixel color for the specified
 *	gray level. The level is in the range 0..1.
 */

OPColorPixel
OPColorMan::FindColor (Real i) const
{
	OP_METHOD (FindColor, 1);
	OP_ARG (1, Real, i);
	
	Card16 val = (Card16)(i * 256);
	if (val > 255) val = 255;
	return this->FindColor ((Card8)(val));
}



/*
 *	Find the nearest matching pixel color for the specified
 *	gray level. The level is in the range 0..255.
 */

OPColorPixel
OPColorMan::FindColor (Card8 i) const
{
	OP_METHOD (FindColor, 1);
	OP_ARG (1, Card8, i);
	
	return this->gray_mapping[i];
}



/*
 *	Return an array specifying which color should be used
 *	for a specified gray value.
 */

const OPColorPixel*
OPColorMan::ReturnGrayMapping () const
{
	OP_METHOD (ReturnGrayMapping, 0);
	return this->gray_mapping;
}



/*
 *	Build the gray mapping array for the "GetGrayMapping"
 *	method.
 */

void
OPColorMan::ComputeNearestColor ()
{
	OP_METHOD (ComputeNearestColor, 0);
	
	int          i     = 0;
	OPColorPixel pixel = 0;
	
	for (i = 0; i < 256; i++) {
		
		if (this->pal_size == 0) {
			
			Card8 rr = i;
			Card8 gg = i;
			Card8 bb = i;
			
			pixel = this->MakeTrueColor (rr, gg, bb);
		
		} else {
			
			PaletteEntry* entry = & this->pal_entry[0];
			Count         count =   this->pal_gray;
			int           dist  = 256;
			
			while (count) {
				
				if (entry->match_enable == FALSE) {
					entry++;
					continue;
				}
				
				int delta = ((int)(entry->r) - i);
				if (delta < 0) delta = -delta;
				
				if (delta < dist) {
					
					dist  = delta;
					pixel = entry->pixel;
					
					if (dist == 0) break;
				}
				
				entry++;
				count--;
			}
		}
		
		this->gray_mapping[i] = pixel;
	}
}



/*
 *	Return a pixel color from a color name. This is a kind
 *	of logical colors. It will be implemented using a data
 *	base object.
 */

Bool
OPColorMan::FindColor (OPAtomIn name, OPColorPixelOut pixel) const
{
	OP_METHOD (FindColor, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, OPColorPixelOut, pixel);
	
	//	NIY
	
	return FALSE;
}


/*
 *	Return the color components from a pixel value.
 */

Bool
OPColorMan::FindColor (OPColorPixel p, RealOut r, RealOut g, RealOut b) const
{
	OP_METHOD (FindColor, 4);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, RealOut, r);
	OP_ARG (3, RealOut, g);
	OP_ARG (4, RealOut, b);
	
	Card8 r8 = 0;
	Card8 g8 = 0;
	Card8 b8 = 0;
	
	//	Use system dependent function here.
	
	if (this->FindColor (p, r8, g8, b8)) {
		
		r = r8 / 255.0;
		g = g8 / 255.0;
		b = b8 / 255.0;
		
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Define a color name and its associated RGB values in
 *	the 0..1 range.
 */

Bool
OPColorMan::DefineColor (OPAtomIn name, Real r, Real g, Real b)
{
	OP_METHOD (DefineColor, 4);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, Real, r);
	OP_ARG (3, Real, g);
	OP_ARG (4, Real, b);
	
	return this->DefineColor (name, (Card8)(r * 255.0),
									(Card8)(g * 255.0),
									(Card8)(b * 255.0));
}



/*
 *	Define a color name and its associated RGB values in
 *	the 0..255 range.
 */

Bool
OPColorMan::DefineColor (OPAtomIn name, Card8 r, Card8 g, Card8 b)
{
	OP_METHOD (DefineColor, 4);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (2, Card8, r);
	OP_ARG (3, Card8, g);
	OP_ARG (4, Card8, b);
	
	//	NIY
	
	return FALSE;
}



/*
 *	Count the available colors (gray levels and RGB colors
 *	are being added).
 */

Count
OPColorMan::CountColors () const
{
	OP_METHOD (CountColors, 0);
	
	return (this->pal_gray + this->pal_cube > 0)
		   ? (this->pal_gray + this->pal_cube)
		   : ((Count)(1) << this->depth);
}



/*
 *	Return TRUE if the output device supports true color.
 *	This will be considered TRUE if the depth is greater
 *	than 8 bit/pixel.
 */

Bool
OPColorMan::IsTrueColor () const
{
	OP_METHOD (IsTrueColor, 0);
	return (this->depth > 8) ? TRUE : FALSE;
}



/*
 *	Return the color model of this output device.
 */

OPColorModel
OPColorMan::ReturnColorModel () const
{
	OP_METHOD (ReturnColorModel, 0);
	return this->color_model;
}



/*
 *	Return the gray model of this output device.
 */

OPGrayModel
OPColorMan::ReturnGrayModel () const
{
	OP_METHOD (ReturnGrayModel, 0);
	return this->gray_model;
}


/*****************************************************************************/

/*
 *	Return the default system color manager. This one has
 *	to be specified at OPaC's initialization.
 */

const OPColorManRef&
OPColorMan::ReturnSystemColorMan ()
{
	OP_MEMBER (ReturnSystemColorMan, 0);
	return OPColorMan::system_man;
}



/*
 *	Initialise the default system color manager. This will
 *	be accepted only once.
 */

void
OPColorMan::InitSystemColorMan (OPColorManIn manager)
{
	OP_MEMBER (InitSystemColorMan, 1);
	OP_ARG (1, OPColorManIn, manager);
	OP_ASSERT (OPColorMan::system_man.IsInvalid ());
	OPColorMan::system_man = manager;
}



void
OPColorMan::KillSystemColorMan ()
{
	OP_MEMBER (KillSystemColorMan, 0);
	OPColorMan::system_man = 0;
}


