/*
 *	cycarray.cxx
 *
 *	The Cyclic Array class is a modified version of the normal
 *	array, adding a modulo to the index.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/cycarray.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (cycarray.h);
OP_IMPLEMENT_CLASS (OPCyclicArray);

/****************************************************************/

/*
 *	
 */

OPCyclicArray::OPCyclicArray ()
{
	OP_METHOD (OPCyclicArray, 0);
	
	this->modulo = 0;
}


OPCyclicArray::~OPCyclicArray ()
{
	OP_METHOD (~OPCyclicArray, 0);
	
	//	Empty the array before the parent destructor has a chance of
	//	doing this, since it would not call the proper SetSize method.
	
	this->SetSize (0);
}


/*
 *	
 */

void
OPCyclicArray::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	storage << this->modulo;
	storage << OPStorageSync::mark;
	
	super::StoreInstance (storage);
}


/*
 *	Restore an array.
 */

void
OPCyclicArray::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	storage >> this->modulo;
	storage >> OPStorageSync::mark;
	
	super::RestoreInstance (storage);
}


/*
 *	Handle a possible gap at the beginning of the array (it
 *	appears as soon as a wrap-around occurs).
 */

void
OPCyclicArray::InternalGetGapSize (Count pos, CountOut gap, CountOut num )
{
	OP_METHOD (InternalGetGapSize, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, CountOut, gap);
	OP_ARG (3, CountOut, num);
	
	gap = 0;
	num = 0;
	
	//	There is a gap at the beginning of the array if the number of
	//	elements is larger than the modulo. Don't store any information
	//	about those elements.
	
	if ( (this->modulo)
	  && (this->num_element > this->modulo)
	  && (pos == 0) ) {
		gap = pos = this->num_element - this->modulo;
	}
	
	//	What is the size of the gap ?
	
	while ( (pos < this->num_element)
		 && (this->references[pos].IsInvalid ()) ) {
		pos++;
		gap++;
	}
	
	//	And what is the size of the chunk following the gap ?
	
	while ( (pos < this->num_element)
		 && (this->references[pos].IsValid ()) ) {
		pos++;
		num++;
	}
}

void
OPCyclicArray::InternalSetGapSize (Count pos, Count gap, Count num)
{
	OP_METHOD (InternalSetGapSize, 3);
	OP_ARG (1, Count, pos);
	OP_ARG (2, Count, gap);
	OP_ARG (3, Count, num);
	
	super::InternalSetGapSize (pos, gap, num);	
}


/*
 *	Define the modulo to use. This will empty the array.
 */

Bool
OPCyclicArray::SetModulo (Count modulo)
{
	OP_METHOD (SetModulo, 1);
	OP_ARG (1, Count, modulo);
	
	super::SetSize (0);
	this->modulo = modulo;
	
	return TRUE;
}

Count
OPCyclicArray::ReturnModulo () const
{
	OP_METHOD (ReturnModulo, 0);
	return this->modulo;
}


/*
 *	Overridden method used to change the size with a modulo
 *	for the element count.
 */

Bool
OPCyclicArray::SetSize (Count size)
{
	OP_METHOD (SetSize, 1);
	OP_ARG (1, Count, size);
	
	if (size == this->num_element) {
		return TRUE;
	}
	
	if (this->modulo == 0) {
		return super::SetSize (size);
	}
	
	//	Growing the cyclic array will enlarge the underlying array till the
	//	modulo size is reached.
	
	if (size > this->num_element) {
		
		if (this->num_element < this->modulo) {
			super::SetSize (this->modulo);
		}
		
		this->num_element = size;
		return TRUE;
	}
	
	//	Shrinking the cyclic array below the modulo size will reduce the
	//	size of the underlying array. Anyway, the references which are
	//	removed will be filled with nulls.
	
	Count num = this->num_element - size;
	Count pos = this->num_element;
	
	if (num > this->modulo) {
		num = this->modulo;
	}
	
	while (num--) {
		this->Access (--pos) = OPObjectNull;
	}
	
	if (size < this->modulo) {
		if (this->num_element > this->modulo) {
			this->num_element = this->modulo;
		}
		
		//	WARNING: This might not work properly in every case !!!
		
		super::SetSize (size);
	}
	
	this->num_element = size;
	return TRUE;
}



/*
 *	Overridden method used to access data in a cyclic way
 *	thanks to modulo arithmetic.
 */

OPObjectRef&
OPCyclicArray::Access (Count pos) const
{
	OP_METHOD (Access, 1);
	OP_ARG (1, Count, pos);
	
	static OPObjectRef null;
	
	//	We cannot return a valid element if the one which is asked for is
	//	not in the valid range [max-modulo...max[. The lower bound may be
	//	replaced by 0 if it gets negative.
	
	if ( (pos < this->num_element)
	  && ( (this->modulo == 0)
	    || (this->num_element < this->modulo)
		|| (this->num_element-this->modulo <= pos) ) ) {
		
		if (this->modulo) {
			pos %= this->modulo;
		}
		
		return this->references[pos];
	}
	
	//	If the previous caller modified the "l-value", null might no longer
	//	be null, so make sure this is the case !
	
	null = OPObjectNull;
	return null;
}


