/*
 *	database.cxx
 *
 *	This file implements the OPaC Data Base class.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/database.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (database.h);
OP_IMPLEMENT_CLASS (OPDataBase);

/****************************************************************/

/*
 *	A data base can be accesses as a generic container with just
 *	an index to the "Get" method. An extraction mode has to be
 *	specified :
 */

#define	EXTRACT_NONE			0				//	extract mode not defined
#define	EXTRACT_ENTITY			1				//	extract entities
#define	EXTRACT_ATTR_BY_NAME	2				//	extract attributes by their name
#define	EXTRACT_ATTR_BY_INDEX	3				//	extract attributes by their index

/****************************************************************/


/*
 *	Create an empty data base object.
 */

OPDataBase::OPDataBase ()
{
	OP_METHOD (OPDataBase, 0);
	
	OPMemZone* zone = this->ReturnMemZone ();
	
	this->dict            = OP_NEW_FROM_ZONE (zone, OPDict);
	this->array           = OP_NEW_FROM_ZONE (zone, OPHashTable);
	this->extr_mode       = EXTRACT_NONE;
	this->extr_sort       = OPAtomNull;
	this->extr_attr_name  = OPAtomNull;
	this->extr_attr_index = 0;
}



/*
 *	Destroy the data base object. This will remove every
 *	entity and/or model still stored in the object.
 */

OPDataBase::~OPDataBase ()
{
	OP_METHOD (~OPDataBase, 0);
	this->RemoveAll ();
}



/*
 *	Save the data base to the storage.
 */

void
OPDataBase::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->array
	        << this->dict
	        << this->extr_mode
	        << this->extr_sort
	        << this->extr_attr_name
	        << this->extr_attr_index;
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the data base from the storage.
 */

void
OPDataBase::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->array
	        >> this->dict
	        >> this->extr_mode
	        >> this->extr_sort
	        >> this->extr_attr_name
	        >> this->extr_attr_index;
	
	storage >> OPStorageSync::mark;
}



/*
 *	Return the total number of entities stored in the data
 *	base.
 */

Count
OPDataBase::TotalCount () const
{
	OP_METHOD (TotalCount, 0);
	OP_ASSERT (this->array.IsValid ());
	
	return this->array->TotalCount ();
}



/*
 *	Specify a generic extraction. We will have to interpret
 *	the command in order to be able to understand what kind
 *	of extraction we will have to perform...
 *
 *	These are the accepted extractions :
 *
 *	 - "ExtractEntity" : OPAtomIn sort
 *   - "ExtractAttribute" : OPAtomIn sort, OPAtomIn name
 *
 *	If the extraction category does not match a listed one,
 *	then FALSE is returned. If the types do not match, an
 *	exception will be thrown.
 */

Bool
OPDataBase::SpecifyExtraction (OPArrayIn x)
{
	OP_METHOD (SpecifyExtraction, 1);
	OP_ARG (1, OPArrayIn, x);
	
	static OPAtomRef x_entity = "ExtractEntity";
	static OPAtomRef x_attrib = "ExtractAttribute";
	
	Count size = x->TotalCount ();
	
	if ( (size == 0)
	  || (x[0].IsInvalid ())
	  || (x[0]->IsKindOf (OP_RUNTIME (OPAtom)) == FALSE) ) {
		return FALSE;
	}
	
	OPAtomRef how = x[0];
	
	if ( (how  == x_entity)
	  && (size == 2) ) {
		
		//	Only one argument : the "sort" name.

		this->SpecifyExtractEntity (x[1]);
		return TRUE;
	}

	if ( (how  == x_attrib)
	  && (size == 3) ) {
		
		//	Two arguments : the "sort" name and the attribute name.

		this->SpecifyExtractAttribute (x[1], x[2]);
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Specify an extraction mode for the generic "Get" method.
 *	This is needed, otherwise no extraction will be possible.
 */

void
OPDataBase::SpecifyExtractEntity (OPAtomIn sort)
{
	OP_METHOD (SpecifyExtractEntity, 1);
	OP_ARG (1, OPAtomIn, sort);
	
	this->extr_mode       = EXTRACT_ENTITY;
	this->extr_sort       = sort;
	this->extr_attr_name  = OPAtomNull;
	this->extr_attr_index = 0;
}

void
OPDataBase::SpecifyExtractAttribute (OPAtomIn sort, OPAtomIn name)
{
	OP_METHOD (SpecifyExtractAttribute, 2);
	OP_ARG (1, OPAtomIn, sort);
	OP_ARG (2, OPAtomIn, name);
	
	this->extr_mode       = EXTRACT_ATTR_BY_NAME;
	this->extr_sort       = sort;
	this->extr_attr_name  = name;
	this->extr_attr_index = 0;
}

void
OPDataBase::SpecifyExtractAttribute (OPAtomIn sort, Count index)
{
	OP_METHOD (SpecifyExtractAttribute, 2);
	OP_ARG (1, OPAtomIn, sort);
	OP_ARG (2, Count, index);
	
	this->extr_mode       = EXTRACT_ATTR_BY_INDEX;
	this->extr_sort       = sort;
	this->extr_attr_name  = OPAtomNull;
	this->extr_attr_index = index;
}



/*
 *	Return, depending on the selected extraction mode,
 *	either :
 *	
 *	 1. the nth entity...
 *	 
 *	 2. an attribute associated to the named entity...
 *	 
 *	 3. an attribute associated to the entity specified
 *      by an index...
 */

Bool
OPDataBase::Get (Count index, OPObjectOut object)
{
	OP_METHOD (Get, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPObjectOut, object);
	
	switch (this->extr_mode) {
		
		case EXTRACT_ENTITY:
			{
				OPEntityRef entity;
				if (this->GetEntity (this->extr_sort, index, entity)) {
					object = entity;
					return TRUE;
				}
			}
			break;
		
		case EXTRACT_ATTR_BY_NAME:
			return this->GetAttribute (this->extr_sort, index, this->extr_attr_name, object);
		
		case EXTRACT_ATTR_BY_INDEX:
			return this->GetAttribute (this->extr_sort, index, this->extr_attr_index, object);
		
		default:
			break;
	}
	
	return FALSE;
}



/*
 *	Return an entity specified by its position according
 *	to some sorting criterion (linear sorting based on an
 *	entity attribute).
 */

Bool
OPDataBase::GetEntity (OPAtomIn sort, Count pos, OPEntityOut entity)
{
	OP_METHOD (GetEntity, 3);
	OP_ARG (1, OPAtomIn, sort);
	OP_ARG (2, Count, pos);
	OP_ARG (3, OPEntityOut, entity);
	
	OPObjectRef key;
	OPObjectRef value;
	
	if (this->dict->GetFromKey (sort, value) == FALSE) {
		
		//	Well, nobody asked for that kind of sorting up to now. We will have to
		//	create a new dictionary using the named attribute as key.
		
		OPMemZone* zone = this->ReturnMemZone ();
		
		OPDictRef   dict = OP_NEW_FROM_ZONE (zone, OPDict);
		OPEntityRef entity;
		OPObjectRef attr;
		
		Count num = this->array->TotalCount ();
		Count pos = 0;
		
		//	Sort the current entities using the specified attribute as a key.
		
		while (num--) {
			this->array->Get (pos++, entity);
			entity->GetAttribute (sort, attr);
			dict->Insert (attr, entity);
		}
		
		//	Remember about this new sorted dictionary.
		
		this->dict->Insert (sort, dict);
		
		value = dict;
	}
	
	OPSetRef set = value;
	
	//	Never mind about the key, only the value is valuable for this kind of
	//	query. The key is explicitely contained in the entity.
	
	Bool ok = set->Get (pos, key, value);
	
	if (ok) {
		entity = value;
	}
	
	return ok;
}



/*
 *	Same as GetEntity followed by a GetAttribute.
 */

Bool
OPDataBase::GetAttribute (OPAtomIn sort, Count pos, OPAtomIn name, OPObjectOut attr)
{
	OP_METHOD (GetAttribute, 4);
	OP_ARG (1, OPAtomIn, sort);
	OP_ARG (2, Count, pos);
	OP_ARG (3, OPAtomIn, name);
	OP_ARG (4, OPObjectOut, attr);
	
	OPEntityRef entity;

	if (this->GetEntity (sort, pos, entity)) {
		return entity->GetAttribute (name, attr);
	}

	return FALSE;
}



/*
 *	Same as "GetEntity" followed by an "entity->GetAttribute".
 */

Bool
OPDataBase::GetAttribute (OPAtomIn sort, Count pos, Count index, OPObjectOut attr)
{
	OP_METHOD (GetAttribute, 4);
	OP_ARG (1, OPAtomIn, sort);
	OP_ARG (2, Count, pos);
	OP_ARG (3, Count, index);
	OP_ARG (4, OPObjectOut, attr);
	
	OPEntityRef entity;

	if (this->GetEntity (sort, pos, entity)) {
		return entity->GetAttribute (index, attr);
	}

	return FALSE;
}



/*
 *	Insert a new entity into the data base. This will be
 *	sorted according to the entity's attributes.
 */

Bool
OPDataBase::Insert (OPEntityIn entity)
{
	OP_METHOD (Insert, 1);
	OP_ARG (1, OPEntityIn, entity);
	
	Count pos = 0;
	Count num = this->dict->TotalCount ();
	
	OPObjectRef key;
	OPObjectRef value;
	
	if (this->array->Insert (entity) == 0) {
		return FALSE;
	}
	
	//	For each sorted list of the entities, insert the entity using the specified
	//	attribute as a key.
	
	while (num--) {
		
		this->dict->Get (pos++, key, value);
		
		//	The reference conversions will produce some exception if the types do
		//	not match. This should never happen !
		
		OPAtomRef atom = key;
		OPSetRef  set  = value;
		
		//	Got the set and the attribute name used to sort in the specified set.
		
		OPObjectRef attr;
		
		if (entity->GetAttribute (atom, attr)) {
			set->Insert (attr, entity);
		}
	}
	
	return TRUE;
}


/*
 *	Remove an entity from the data base.
 */

Bool
OPDataBase::Remove (OPEntityIn entity)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, OPEntityIn, entity);
	
	Count pos   = 0;
	Count num   = this->dict->TotalCount ();
	Bool  found = (num) ? TRUE : FALSE;
	
	OPObjectRef key;
	OPObjectRef value;
	
	if (this->array->Remove (entity) == FALSE) {
		return FALSE;
	}
	
	//	For each sorted list of the entities, remove the entity using the specified
	//	attribute as a key.
	
	while (num--) {
		
		this->dict->Get (pos++, key, value);
		
		OPSetRef set = value;
		found &= set->RemoveValue (entity);
	}
	
	return found;
}



/*
 *	Remove every entity from this data base.
 */

void
OPDataBase::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);
	
	Count num = this->dict->TotalCount ();
	
	OPObjectRef key;
	OPObjectRef value;
	
	this->array->RemoveAll ();
	
	//	For each sorted list of the entities, remove the entity using the specified
	//	attribute as a key.
	
	while (num--) {
		
		//	WARNING: A lock should be set up here to prevent that the dictionary
		//	changes without notice while doing the two following operations :
		
		this->dict->Get (0, key, value);
		this->dict->Remove (0);
		
		OPSetRef  set  = value;
		set->RemoveAll ();
	}
}

