/*
 *	dragpal.cxx
 *
 *	The drag palette is a view from which other views can be dragged. These
 *	automatically get cloned and stored into a "drag view".
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/carbon.h>
#include <OPaC/dragpal.h>
#include <OPaC/dragview.h>
#include <OPaC/event.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (dragpal.h);
OP_IMPLEMENT_CLASS (OPDragPalette);

/****************************************************************/


/*
 *	Initialise a new instance. Nothing special has to be
 *	done now.
 */

OPDragPalette::OPDragPalette ()
{
	OP_METHOD (OPDragPalette, 0);
	
	this->auto_copy  = TRUE;
	this->state     &= ~OP_VIEW_OK_DROP;
}



/*
 *	Delete an instance. Nothing special has to be done here.
 */

OPDragPalette::~OPDragPalette ()
{
	OP_METHOD (~OPDragPalette, 0);
}



/*
 *	Store a drag palette and all its subviews.
 */

void
OPDragPalette::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore a drag palette and all its subviews.
 */

void
OPDragPalette::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the drag palette for its deletion.
 */

void
OPDragPalette::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
}


/*
 *	Initialise the drag palette.
 */

void
OPDragPalette::Initialise (OPViewIn view)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, OPViewIn, view);
	
	if (view.IsValid ()) {
		
		OPViewRef parent;
		OPRect box;
		
		view->GetParent (parent);
		view->GetBox (box);
		
		this->SetParent (parent);
		this->SetOrigin (box.X1 () - 1, box.Y1 () - 1);
		this->SetSize (box.DX () + 2, box.DY () + 2);
		
		view->SetParent (this);
	}
}


/*
 *	The drag palette may want to check if it accepts some
 *	child or not. This is currently not used.
 */

Bool
OPDragPalette::WouldAcceptChild (OPViewIn child) const
{
	OP_METHOD (WouldAcceptChild, 1);
	OP_ARG (1, OPViewIn, child);
	return TRUE;
}



/*
 *	Whenever a new child is added to the palette, it is put
 *	into a special mode where detection is disabled.
 */

void
OPDragPalette::NotifyInsertChild (OPViewIn child)
{
	OP_METHOD (NotifyInsertChild, 1);
	OP_ARG (1, OPViewIn, child);
	OP_ASSERT (child.IsValid ());
	
	child->SetFlagOkDetect (FALSE);
	
	super::NotifyInsertChild (child);
}



/*
 *	When a view is being removed from the palette, we could
 *	consider restoring its previous state. We don't do it
 *	yet.
 */

void
OPDragPalette::NotifyRemoveChild (OPViewIn child)
{
	OP_METHOD (NotifyRemoveChild, 1);
	OP_ARG (1, OPViewIn, child);
	super::NotifyRemoveChild (child);
}



/*
 *	Same as the detect method, but check directly based on
 *	the bounding box, so we don't have any problems handling
 *	disabled views.
 */

static Bool
micro_detect (OPCoord x, OPCoord y, OPViewIn view)
{
	if ( (view.IsValid ())
	  && (view->ReturnBox ().IsInside (x, y)) ) {
		return TRUE;
	}

	return FALSE;
}


/*
 *	When the user presses a mouse button, we will start a
 *	dragging operation based on the clicked view.
 */

void
OPDragPalette::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (event->ReturnChange () == OP_EV_BUTTON_LEFT) {
		
		OPViewRef iter = this->child.tail;
		OPCoord   x    = event->ReturnX ();
		OPCoord   y    = event->ReturnY ();
		
		while (iter.IsValid ()) {
			if (micro_detect (x, y, iter)) {
				break;
			}
			iter->GetPrevSibling (iter);
		}
		
		if (iter.IsValid ()) {
			
			OPViewRef     self = this;
			OPDragViewRef drag = OP_NEW (OPDragView);
			OPViewRef     copy;
			
			if (this->auto_copy) {
				
				//	Make a carbon copy of the selected view, then put it into the
				//	drag view... It's fun to play with these objects :-)
				
				OPCarbonCopyRef carbon = OP_NEW (OPCarbonCopy);
				
				carbon->StartStoring ();
				carbon << iter;
				carbon->StartRestoring ();
				carbon >> copy;
				
				copy->SetParent (self);
				
			} else {
				
				//	Don't copy anything. Tear off the view...
				
				copy = iter;
			}
			
			OPCoord ox;
			OPCoord oy;
			
			copy->SetFlagOkDetect (TRUE);
			copy->GetOrigin (ox, oy);
			
			drag->SetAutoStop (TRUE);
			drag->SetAutoOops (FALSE);
			drag->Setup (copy, TRUE);
			drag->SetHotSpot (x-ox, y-oy);
			
			OPEvent::HandOverToWindow (drag);
		}

		return;
	}

	super::ButtonPress (event);
}



/*
 *	Define whether the dragged views will be copied or just
 *	torn away.
 */

void
OPDragPalette::SetAutoCopy (Bool copy)
{
	OP_METHOD (SetAutoCopy, 1);
	OP_ARG (1, Bool, copy);

	this->auto_copy = copy;
}

Bool
OPDragPalette::ReturnAutoCopy () const
{
	OP_METHOD (GetAutoCopy, 0);
	return this->auto_copy;
}


/*
 *	Avoid that objects get dropped back into the drag palette
 *	from where they originate...
 */

Bool
OPDragPalette::DragAndDrop (OPRectIn box, OPObjectIn initiator, OPDragDropProtocol ddp)
{
	OP_METHOD (DragAndDrop, 3);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPObjectIn, initiator);
	OP_ARG (3, OPDragDropProtocol, ddp);
	
	switch (ddp) {
		case OP_DDP_DRAGnDROP_CHECK:
			return FALSE;
		case OP_DDP_CONNECT_CHECK:
			return FALSE;
		case OP_DDP_UI_EDIT_CHECK:
			return FALSE;
	}
	
	return super::DragAndDrop (box, initiator, ddp);
}