/*
 *	genprovider.cxx
 *
 *	The Generic Provider is an abstact class which must be overridden
 *	in order to provide a collection of objects.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/boxview.h>
#include <OPaC/icon.h>
#include <OPaC/new.h>
#include <OPaC/objectpane.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>

/****************************************************************/

OP_ABSTRACT_CLASS (OPGenericProvider);

/****************************************************************/

OPGenericProvider::OPGenericProvider ()
{
	OP_METHOD (OPGenericProvider, 0);
}

OPGenericProvider::~OPGenericProvider ()
{
	OP_METHOD (~OPGenericProvider, 0);
}

void
OPGenericProvider::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
}

void
OPGenericProvider::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
}


/* 
 *	Build a complete view with the specified object. This is
 *	handled as a simple small icon cell with text.
 */

Bool
OPGenericProvider::GetCompleteView (Count index, OPCoord dx, OPCoord dy, OPControlOut cell)
{
	OP_METHOD (GetCompleteView, 4);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ARG (4, OPControlOut, cell);
	
	OPLookIcon  look_icon = OP_ICON_NONE;
	OPStringRef text;
	OPObjectRef object;
	OPCoord x = 0;
	OPCoord y = 0;
	
	OPBoxViewRef box  = OP_NEW (OPBoxView);
	OPIconRef    icon = OP_NEW (OPIcon);
	OPTextTagRef tag  = OP_NEW (OPTextTag);
	
	box->SetSize (dx, dy);
	box->EnableTitle (FALSE);
	box->EnableFrame (FALSE);
	
	this->GetObject (index, object);
	this->GetSmallIcon (index, look_icon);
	this->GetText (index, text);
	
	icon->Initialise (look_icon);
	icon->GetMinSize (x, y);
	icon->SetParent (box);
	icon->SetOrigin (0, 0);
	icon->SetSize (x, dy);
	icon->SetFlagOkDetect (FALSE);
	
	tag->SetTitle (text);
	tag->SetParent (box);
	tag->SetOrigin (x, 0);
	tag->SetSize (dx-x, dy);
	
	cell = box;
	
	cell->SetHiliteMode (TRUE);
	cell->SetShapeStyle (cell->ReturnShapeStyle () | OP_CTRL_AUTOHILITE);
	cell->SetAttribute ("EmbeddedObjectForConnect", object);
	
	return TRUE;
}

/****************************************************************/

/*
 *	Return the generic sub-tree provider (if any) for the
 *	specified object. This should be overridden if some
 *	meaningful (dynamic) browsing is required.
 */

Bool
OPGenericProvider::EnterSubtree (Count index, OPGenericProviderOut provider)
{
	OP_METHOD (EnterSubtree, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPGenericProviderOut, provider);
	
	OPObjectRef object;
	
	if ( (this->GetObject (index, object))
	  && (object.IsValid ()) ) {
		
		object->GetAttribute ("DefaultBrowserProvider", object);
		
		if (object.IsInvalid ()) {
			return FALSE;
		}
		
		static OPAtomRef name = "Provider";
		
		if (object->QueryData (name, provider)) {
			return TRUE;
		}
		
		if (object->IsKindOf (OP_RUNTIME (OPGenericProvider))) {
			provider = object;
			return TRUE;
		}
	}
	
	return FALSE;
}

Bool
OPGenericProvider::ReleaseSubtree (Count index)
{
	OP_METHOD (ReleaseSubtree, 1);
	OP_ARG (1, Count, index);
	
	return TRUE;
}

/****************************************************************/

