/*
 *	lookmancol.cxx
 *
 *	Look Manager - color handling. See lookman.cxx.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>

/****************************************************************/

/*
 *	The look manager relies on a color data base. This will be
 *	dynamic some day; currently, it is static !
 */

struct InternalColorDef
{
	OPLookColor		 color;						//	color ID
	const LatinChar* name;						//	color name
	Card8			 r,g,b;						//	RGB components
};

/****************************************************************/

OPLookColor	OPLookMan::col_conv[OP_COLOR_NUM];
OPAtomRef	OPLookMan::col_name[OP_COLOR_NUM];

/****************************************************************/

static InternalColorDef opac_color_def[] = {
  {	OP_COLOR_WHITE,			"White",		0xFF, 0xFF, 0xFF },
  { OP_COLOR_LGRAY,			"LightGray",	0xC0, 0xC0, 0xC0 },
  { OP_COLOR_DGRAY,			"DarkGray",		0x80, 0x80, 0x80 },
  { OP_COLOR_BLACK,			"Black",		0x00, 0x00, 0x00 },
  { OP_COLOR_BACKGROUND,	"Background",	0xC0, 0xC0, 0xC0 },
  { OP_COLOR_WIDGET_BACK,	"WidgetBack",	0x80, 0x80, 0x80 },
  { OP_COLOR_WIDGET_FORE,	"WidgetFore",	0xC0, 0xC0, 0xC0 },
  { OP_COLOR_PSY_DANGER,	"PsyDanger",	0xFF, 0x00, 0x00 },
  { OP_COLOR_PSY_WARNING,	"PsyWarning",	0xC0, 0x80, 0x00 },
  { OP_COLOR_PSY_SAFE,		"PsySafe",		0x00, 0xC0, 0x00 },
  { OP_COLOR_VIEW_NORMAL,	"ViewNormal",	0x00, 0x00, 0x00 },
  { OP_COLOR_VIEW_DIM,		"ViewDim",		0xA0, 0xA0, 0xA0 },
  { OP_COLOR_VIEW_ACTIVE,	"ViewActive",	0xE0, 0xE0, 0xE0 },
  { OP_COLOR_VIEW_DRAGGING,	"ViewDragging",	0xE0, 0xE0, 0xE0 },
  { OP_COLOR_VIEW_DROPPING,	"ViewDropping",	0xC0, 0xFF, 0xC0 },
  { OP_COLOR_VIEW_CURSOR,	"ViewCursor",	0x00, 0x00, 0x00 },
  { OP_COLOR_VIEW_HILITE,	"ViewHilite",	0xFF, 0xFF, 0xFF },
  { OP_COLOR_VIEW_CONNECT,	"ViewConnect",	0x00, 0x00, 0x00 },
  { OP_COLOR_HILITE_CLICK,	"HiliteClick",	0xFF, 0xFF, 0xFF },
  { OP_COLOR_HILITE_SELECT,	"HiliteSelect",	0x80, 0x80, 0x80 },
  { OP_COLOR_HILITE_ACTIVE, "HiliteActive", 0xE0, 0xE0, 0xE0 },
  { OP_COLOR_UI_SELECT,		"UISelect",		0xC0, 0x00, 0x00 },
  { OP_COLOR_UI_SEL_ACTIVE, "UISelActive",	0xFF, 0x00, 0x00 },
  { OP_COLOR_TOOLTIP_BACK,	"ToolTipBack",	0xFF, 0xFF, 0xC0 },
  { OP_COLOR_TOOLTIP_FRAME,	"ToolTipFrame", 0x00, 0x00, 0x00 },
  { OP_COLOR_TOOLTIP_TEXT,	"ToolTipText",	0x00, 0x00, 0x00 },
  { OP_COLOR_NONE,			0,				0x00, 0x00, 0x00 }
};

/****************************************************************/

void
OPLookMan::InitColors ()
{
	Count i;
	
	for (i = 0; i < OP_COLOR_NUM; i++) {
		OPLookMan::col_conv[i] = OP_COLOR_NONE;
		this->col_rgb[0][i] = 0;
		this->col_rgb[1][i] = 0;
		this->col_rgb[2][i] = 0;
	}
	
	InternalColorDef* def = opac_color_def;
	
	while (def->name) {
		
		OPLookMan::col_conv[def->color] = def->color;
		OPLookMan::col_name[def->color] = def->name;
		
		this->col_rgb[0][def->color] = def->r;
		this->col_rgb[1][def->color] = def->g;
		this->col_rgb[2][def->color] = def->b;
		
		def++;
	}
}

Bool
OPLookMan::GetColorName (OPLookColor color, OPAtomOut name) const
{
	OP_METHOD (GetColorName, 2);
	OP_ARG (1, OPLookColor, color);
	OP_ARG (2, OPAtomOut, name);
	
	if ( (color > OP_COLOR_MAX)
	  || (OPLookMan::col_conv[color] == OP_COLOR_NONE) ) {
		
		name = "";
		return FALSE;
	}
	
	name = OPLookMan::col_name[color];
	
	return TRUE;
}

Bool
OPLookMan::GetColorRGB (OPLookColor color, Card8Out r, Card8Out g, Card8Out b) const
{
	OP_METHOD (GetColorRGB, 4);
	OP_ARG (1, OPLookColor, color);
	OP_ARG (2, Card8Out, r);
	OP_ARG (3, Card8Out, g);
	OP_ARG (4, Card8Out, b);
	
	if ( (color > OP_COLOR_MAX)
	  || (OPLookMan::col_conv[color] == OP_COLOR_NONE) ) {
		return FALSE;
	}
	
	color = OPLookMan::col_conv[color];
	
	r = this->col_rgb[0][color];
	g = this->col_rgb[1][color];
	b = this->col_rgb[2][color];
	
	return TRUE;
}

OPLookColor
OPLookMan::FindColor (OPAtomIn name) const
{
	OP_METHOD (FindColor, 1);
	OP_ARG (1, OPAtomIn, name);
	
	for (Count i = 0; i < OP_COLOR_NUM; i++) {
		if (OPLookMan::col_name[i] == name) {
			return OPLookMan::col_conv[i];
		}
	}
	
	return OP_COLOR_NONE;
}

OPLookColor
OPLookMan::FindNthColor (Count index) const
{
	OP_METHOD (FindNthColor, 1);
	OP_ARG (1, Count, index);
	
	for (Count i = 0; i < OP_COLOR_NUM; i++) {
		if (OPLookMan::col_conv[i] != OP_COLOR_NONE) {
			if (index-- == 0) {
				return OPLookMan::col_conv[i];
			}
		}
	}
	
	return OP_COLOR_NONE;
}

/******************************************************************************/
