/*
 *	switch.cxx
 *
 *	Switch class implementation : it is used for both the radio buttons
 *	and the check-box buttons.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/switch.h>
#include <OPaC/icon.h>
#include <OPaC/texttag.h>
#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>

#include <OPaC/notmacro.h>
#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (switch.h);
OP_IMPLEMENT_CLASS (OPSwitch);

/****************************************************************/

OP_TARGET_ACTOR (OPSwitch, TurnOn);
OP_TARGET_ACTOR (OPSwitch, TurnOff);
OP_TARGET_ACTOR (OPSwitch, Toggle);

OP_NOTIFY_EVENT (OPSwitch, UserChange);
OP_NOTIFY_EVENT (OPSwitch, AutoChange);

OP_NOTIFY_CHANGE (OPSwitch, Value, UserChange);
OP_NOTIFY_CHANGE (OPSwitch, Value, AutoChange);

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPSwitch, "Value", ReturnValue, SetValue, "Switch value", Bool, Bool);
OP_PUBLISH_MODIFIER (OPSwitch, "Title", GetTitle, SetTitle, "Switch title", OPObject, OPString);
OP_PUBLISH_MODIFUNC (OPSwitch, "RadioButton", IsARadioButton, SetAsRadioButton, "Switch as radio button", Bool, Bool);

/****************************************************************/

OPAtomRef OPSwitch::name_clear_radio;

/****************************************************************/

/*
 *	We could have defined operators to do the "a+b-c" op
 *	on the icon description, but then GCC 2.5.8 would not
 *	be happy...
 */

static inline void
combine (OPLookIcon& a, OPLookIcon b, OPLookIcon c)
{
	int x = (int)(a) + (int)(b) - (int)(c);
	a = (OPLookIcon)(x);
}

/****************************************************************/

/*
 *	Initialise the switch. Default will be neither check-
 *	box, nor radio button.
 */

OPSwitch::OPSwitch ()
{
	OP_METHOD (OPSwitch, 0);
	
	this->title = 0;
	this->value = FALSE;
	
	this->shape |= OP_CTRL_AUTOACTIVE;
	this->shape &= ~ OP_CTRL_AUTOACTHILITE;
	this->shape &= ~ OP_CTRL_AUTOHILITE;
	
	this->do_press_on_action  = FALSE;
	this->do_hilite_on_action = TRUE;
	this->is_radio_button     = FALSE;
	this->is_check_box        = FALSE;

	OPSwitch::name_clear_radio = "ClearRadio";
}



/*
 *	Destroy the switch.
 */

OPSwitch::~OPSwitch ()
{
	OP_METHOD (~OPSwitch, 0);
	OP_ASSERT (this->title.IsValid () == FALSE);
}



/*
 *	Store the switch, its current state and its title.
 */

void
OPSwitch::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->value;
	storage << this->is_radio_button << this->is_check_box;
	storage << this->title;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the switch, its current state and its title.
 */

void
OPSwitch::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->value;
	storage >> this->is_radio_button >> this->is_check_box;
	storage >> this->title;
	storage >> OPStorageSync::mark;
}



/*
 *	This casting function is needed by the OPaC extension
 *	class OPxTitleWidget.
 */

OPView*
OPSwitch::ReturnObjectAsView ()
{
	return this;
}


/*
 *	Allocate a new title object, when this is needed or
 *	free the existing one.
 */

void
OPSwitch::AllocTitle (Bool alloc)
{
	OP_METHOD (AllocTitle, 1);
	OP_ARG (1, Bool, alloc);
	
	if ( (this->title.IsValid () == TRUE)
	  && (alloc == FALSE) ) {
		
		//	Free the current title...
		
		OPTextTagRef title = this->title;
		
		title->SetParent (OPViewNull);
		title->Delete ();
		
		this->title = 0;
	}
	
	if ( (this->title.IsValid () == FALSE)
	  && (alloc == TRUE) ) {
		
		//	Allocate a new title instance...
		
		this->title = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPTextTag);
		this->title->SetParent (this);
		this->title->SetFlagOkDetect (FALSE);
		this->title->SetFlagOkUIMove (FALSE);
		this->title->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		
		//	Define the title's position...
		
		this->ChangeGeometry (this->box);
	}
}


/*
 *	Delete the switch. This will also delete the title...
 */

void
OPSwitch::Delete ()
{
	OP_METHOD (Delete, 0);
	this->AllocTitle (FALSE);
	super::Delete ();
}



/*
 *	Initialise the switch as a check box.
 */

void
OPSwitch::InitAsCheckBox ()
{
	OP_METHOD (InitAsCheckBox, 0);
	
	this->is_radio_button = FALSE;
	this->is_check_box    = TRUE;
}



/*
 *	Initialise the switch as a radio button.
 */

void
OPSwitch::InitAsRadioButton ()
{
	OP_METHOD (InitAsRadioButton, 0);
	
	this->is_radio_button = TRUE;
	this->is_check_box    = FALSE;
}



/*
 *	Turn the switch on, same as SetValue (TRUE). This will
 *	automatically turn of other radio buttons.
 */

void
OPSwitch::TurnOn ()
{
	OP_METHOD (TurnOn, 0);
	
	if ( (this->is_radio_button)
	  && (this->value == FALSE) ) {
		
		OPSwitchRef self = this;
		this->parent->Broadcast (OPSwitch::name_clear_radio, self);
	}

	this->SetValue (TRUE);
}



/*
 *	Turn the switch off, same as SetValue (FALSE).
 */

void
OPSwitch::TurnOff ()
{
	OP_METHOD (TurnOff, 0);
	this->SetValue (FALSE);
}


/*
 *	Change the state of the switch.
 */

void
OPSwitch::Toggle ()
{
	OP_METHOD (Toggle, 0);
	
	if (this->value) {
		this->TurnOff ();
	} else {
		this->TurnOn ();
	}
}


/*
 *	Return TRUE if the switch is a radio button.
 */

Bool
OPSwitch::IsARadioButton () const
{
	OP_METHOD (IsARadioButton, 0);
	return this->is_radio_button;
}

Bool
OPSwitch::IsARadioButton (BoolOut yes) const
{
	OP_METHOD (IsARadioButton, 1);
	OP_ARG (1, BoolOut, yes);
	
	yes = this->is_radio_button;
	
	return TRUE;
}


/*
 *	Return TRUE if the switch is a radio button.
 */

Bool
OPSwitch::SetAsRadioButton (Bool yes)
{
	OP_METHOD (SetAsRadioButton, 1);
	OP_ARG (1, Bool, yes);

	if (yes) {
		this->InitAsRadioButton ();
	} else {
		this->InitAsCheckBox ();
	}

	return TRUE;
}



/*
 *	Return the switch's title.
 */

Bool
OPSwitch::GetTitle (OPStringOut title) const
{
	OP_METHOD (GetTitle, 1);
	OP_ARG (1, OPStringOut, title);
	
	if (this->title.IsValid () == FALSE) {
		title = "";
		return TRUE;
	}
	
	this->title->GetText (title);
	return TRUE;
}



/*
 *	Define the title. Don't redraw anything.
 */

Bool
OPSwitch::SetTitle (OPStringIn title)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, title);
	
	this->AllocTitle (TRUE);
	this->title->SetText (title);
	
	return TRUE;
}



/*
 *	Define the value. Don't redraw anything.
 */

Bool
OPSwitch::SetValue (Bool value)
{
	OP_METHOD (SetValue, 1);
	OP_ARG (1, Bool, value);
	
	this->value = value;
	this->Report (notify_AutoChange);
	
	return TRUE;
}



/*
 *	Return the switch's value.
 */

Bool
OPSwitch::ReturnValue () const
{
	OP_METHOD (ReturnValue, 0);
	return this->value;
}



/*
 *	Return the switch's minimal size. This will be the
 *	icon's size added to the text's size.
 */

void
OPSwitch::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPLookIcon icon  = (this->is_radio_button) ? OP_ICON_RADIO_0_NORMAL
											   : OP_ICON_CHECK_0_NORMAL;
	
	OPCoord dx1 = 0, dy1 = 0;
	OPCoord dx2 = 0, dy2 = 0;
	
	if (this->title.IsValid ()) {
		this->title->GetMinSize (dx1, dy1);
	}
	
	OPLookMan::Current ()->GetIconSize (icon, dx2, dy2);
	
	dy = (dy1 > dy2 ? dy1 : dy2);
	dx = dx1 + dx2 + 6;
}



/*
 *	Draw the active frame around the switch's text.
 */

void
OPSwitch::DrawActiveFrame (OPGrailPortIn port, OPAdornerIn adorner,
							OPCoord margin)
{
	OP_METHOD (DrawActiveFrame, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPAdornerIn, adorner);
	OP_ARG (3, OPCoord, margin);
	
	if (this->title.IsValid ()) {
		
		OPRect rect = this->title->ReturnBox ();
		
		adorner->DrawRect (rect, OP_COLOR_BLACK);
	}
}


/*
 *	Draw the switch.
 */

void
OPSwitch::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		OPLookIcon icon;
		OPCoord    dx = 0, dy = 0;
		OPCoord    offset_x = 0, offset_y = 0;
		
		icon  = (this->is_radio_button) ? OP_ICON_RADIO_0_NORMAL
										: OP_ICON_CHECK_0_NORMAL;
		
		//	The icon may be either :
		//
		//	- normal (enabled, not pressed, not highlighted).
		//	- dimmed (not enabled, not pressed, not highlighted)
		//	- pressed (enabled, not highilighted)
		//	- hilited (enabled, not pressed)
		
		if (this->value) {
			combine (icon, OP_ICON_RADIO_1_NORMAL, OP_ICON_RADIO_0_NORMAL);
		}
		
		if ((this->state & OP_VIEW_IS_ENABLED) == 0) {
			combine (icon, OP_ICON_RADIO_0_DIM, OP_ICON_RADIO_0_NORMAL);
		} else if (this->is_being_pressed) {
			combine (icon, OP_ICON_RADIO_0_PRESS, OP_ICON_RADIO_0_NORMAL);
		} else if (this->is_being_hilited) {
			combine (icon, OP_ICON_RADIO_0_HILITE, OP_ICON_RADIO_0_NORMAL);
		}
		
		OPLookMan::Current ()->GetIconSize (icon, dx, dy);
		
		offset_y = (this->box.DY () - dy) / 2;
		offset_x = OPLookMan::Current ()->ReturnWidgetFrameSize ();
		
		//	The text (title) will be drawn automagically by the TextTag instance.
		
		adorner->DrawIcon (this->box.X1 () + offset_x, this->box.Y1 () + offset_y, icon);
	}
}



/*
 *	There has been some action (either mouse click or key
 *	pressed) which will change the switch's state.
 */

void
OPSwitch::Action (Bool clicked)
{
	OP_METHOD (Action, 1);
	OP_ARG (1, Bool, clicked);
	
	if (this->is_radio_button) {
		OPSwitchRef self = this;
		this->parent->Broadcast (OPSwitch::name_clear_radio, self);
		this->parent->Display ();
	}

	this->SetValue (this->value ? FALSE : TRUE);
	
	if (clicked) {
		this->Report (notify_UserChange, OP_REPORT_NEW);
	}
	
	super::Action (clicked);
}



/*
 *	Set the switch's size. This will move and change the
 *	size of the title.
 */

void
OPSwitch::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	
	OPCoord odx = 0;
	OPCoord ody = 0;
	
	OPLookIcon icon  = (this->is_radio_button) ? OP_ICON_RADIO_0_NORMAL
											   : OP_ICON_CHECK_0_NORMAL;
	
	OPLookMan::Current ()->GetIconSize (icon, odx, ody);
	
	OPCoord nx = this->box.X1 () + 6 + odx;
	OPCoord ny = this->box.Y1 ();
	
	if (this->title.IsValid ()) {
		this->title->SetOrigin (nx, ny);
		this->title->SetSize (this->box.DX () - 6 - odx, this->box.DY ());
	}
}



/*
 *	Handle broadcast message "ClearRadio" only, others are
 *	simple forwarded to the super-class.
 */

void
OPSwitch::Broadcast (OPAtomIn message, OPObjectIn param)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, message);
	OP_ARG (2, OPObjectIn, param);
	
	if ( (this->is_radio_button)
	  && (message == OPSwitch::name_clear_radio) ) {
		this->SetValue (FALSE);
	}
	
	super::Broadcast (message, param);
}



