/*
 *	textline.cxx
 *
 *	The Text Line is used to represent and edit a line of text.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/event.h>
#include <OPaC/window.h>
#include <OPaC/app.h>
#include <OPaC/new.h>
#include <OPaC/targetmacro.h>
#include <OPaC/textline.h>

#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (textline.h);
OP_IMPLEMENT_CLASS (OPTextLine);

/****************************************************************/

OP_NOTIFY_EVENT (OPTextLine, UserTextChange);
OP_NOTIFY_EVENT (OPTextLine, AutoTextChange);
OP_NOTIFY_EVENT (OPTextLine, OnClick);

OP_NOTIFY_CHANGE (OPTextLine, Text, UserTextChange);
OP_NOTIFY_CHANGE (OPTextLine, Text, AutoTextChange);
OP_NOTIFY_CHANGE (OPTextLine, Text, Changed);

/****************************************************************/

OP_TARGET_ACTOR (OPTextLine, Clear);
OP_TARGET_ACTOR (OPTextLine, SetFullHilite);
OP_TARGET_ACTOR (OPTextLine, SetEmptyHilite);

/****************************************************************/

OPCoord OPTextLine::look_3D_effect  = 0;
OPCoord OPTextLine::look_frame_size = 0;
Count   OPTextLine::look_cursor_off = 0;
Count   OPTextLine::look_cursor_on  = 0;

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPTextLine, "CursorPosition", ReturnCursorPos, SetCursorPos, "Cursor position", Card32, Card32);
OP_PUBLISH_FUNCTION (OPTextLine, "IsHilited", IsHilited, "Text is hilited", Bool, Bool);
OP_PUBLISH_MODIFIER (OPTextLine, "Text", GetText, SetText, "Text", OPObject, OPString);
OP_PUBLISH_MODIFUNC (OPTextLine, "IsFrameless", ReturnFramelessMode, SetFramelessMode, "Text is frameless", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPTextLine, "IsTransparent", ReturnTransparentMode, SetTransparentMode, "Text is transparent", Bool, Bool);

/****************************************************************/

#define	TEXT_BORDER			2
#define TEXT_BORDER_WIDTH	2
#define	TEXT_BORDER_HEIGHT	1

#define	TEXT_SCROLL			16
#define	TEXT_SCROLL_RIGHT	4

/****************************************************************/

/*
 *	Create an empty text line.
 */
	
OPTextLine::OPTextLine ()
{
	OP_METHOD (OPTextLine, 0);
	
	this->state |= (OP_VIEW_IS_ENABLED | OP_VIEW_IS_ALIVE);
	
	//	Override some definitions made by the OPControl class :
	
	this->do_hilite_on_action = FALSE;
	this->do_press_on_action  = FALSE;
	
	this->text_string = "";
	
	this->cursor_pos   = 0;
	this->hilite_pos_1 = 0;
	this->hilite_pos_2 = 0;
	this->temp_pos     = 0;
	
	this->text_dx    = 0;
	this->cursor_x   = 0;
	this->hilite_x_1 = 0;
	this->hilite_x_2 = 0;
	this->start_x    = 0;
	
	this->cursor_ascender  = 0;
	this->cursor_descender = 0;
	
	this->cursor_blink   = FALSE;
	this->cursor_enable  = TRUE;
	this->is_dragging    = FALSE;
	this->is_frameless   = FALSE;
	this->is_transparent = FALSE;
	
	const OPLookManRef& look = OPLookMan::Current ();
	
	OPTextLine::look_3D_effect  = look->Return3DEffectWidth ();
	OPTextLine::look_frame_size = look->ReturnWidgetFrameSize ();
	OPTextLine::look_cursor_off = look->ReturnCursorOffTime ();
	OPTextLine::look_cursor_on  = look->ReturnCursorOnTime ();
}



/*
 *	Free a text line. This need not do anything special.
 */

OPTextLine::~OPTextLine ()
{
	OP_METHOD (~OPTextLine, 0);
}



/*
 *	Delete a text line. This will free the link to the
 *	string object.
 */

void
OPTextLine::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->text_string = OPStringNull;
	super::Delete ();
}



/*
 *	Store the current text line. This won't store the
 *	cursor geometry, since it will be restored by Awake !
 */

void
OPTextLine::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->text_string;
	storage << this->cursor_pos;
	storage << this->hilite_pos_1 << this->hilite_pos_2;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the text line. The cursor geometry will be
 *	computed by "Awake..." !
 */

void
OPTextLine::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->text_string;
	storage >> this->cursor_pos;
	storage >> this->hilite_pos_1 >> this->hilite_pos_2;
	storage >> OPStorageSync::mark;
}



/*
 *	Awaking from the storage : this computes the new
 *	cursor geometry.
 */

void
OPTextLine::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
	this->UpdateCursorGeometry ();
}

void
OPTextLine::AwakeFromApp (OPAppIn app)
{
	OP_METHOD (AwakeFromApp, 1);
	OP_ARG (1, OPAppIn, app);
	
	super::AwakeFromApp (app);
	this->UpdateCursorGeometry ();
}

void
OPTextLine::AwakeFromObject (OPObjectIn object)
{
	OP_METHOD (AwakeFromObject, 1);
	OP_ARG (1, OPObjectIn, object);
	
	super::AwakeFromObject (object);
	this->UpdateCursorGeometry ();
}



/*
 *	This method will be called whenever the text has been
 *	edited or modified. If user is set to TRUE, then the
 *	modification is the result of a user action.
 */

void
OPTextLine::NotifyTextChange (Bool user)
{
	OP_METHOD (NotifyTextChange, 1);
	OP_ARG (1, Bool, user);
	
	Count      id   = (user) ? OP_REPORT_NEW : OP_REPORT_AUTO;
	OPAtomRef& atom = (user) ? notify_UserTextChange : notify_AutoTextChange;
	
	this->Report (atom, id);
	this->ReportChange ();
}


/*
 *	Set the text stored in the line. The cursor will be
 *	placed at the end of it.
 */

Bool
OPTextLine::SetText (OPStringIn text)
{
	OP_METHOD (SetText, 1);
	OP_ARG (1, OPStringIn, text);
	
	if (text.IsValid ()) {
		this->text_string = text;
	} else {
		this->text_string = "";
	}

	this->SetCursorPos (this->text_string->CountCharacters ());
	this->NotifyTextChange (FALSE);
	this->Display ();
	
	return TRUE;
}


/*
 *	Clear the text: remove any text which might have been typed
 *	into the text line.
 */

void
OPTextLine::Clear ()
{
	OP_METHOD (Clear, 0);
	
	this->text_string = "";
	
	this->SetCursorPos (0);
	this->NotifyTextChange (FALSE);
	this->Display ();
}


void
OPTextLine::SetFullHilite ()
{
	OP_METHOD (SetFullHilite, 0);
	
	this->SetHiliteRange (0, this->text_string->CountCharacters ());
}


void
OPTextLine::SetEmptyHilite ()
{
	OP_METHOD (SetEmptyHilite, 0);
	
	this->SetHiliteRange (0, 0);
}



/*
 *	Return the current text stored in the line.
 */

Bool
OPTextLine::GetText (OPStringOut text) const
{
	OP_METHOD (GetText, 1);
	OP_ARG (1, OPStringOut, text);
	
	text = "";
	
	if (this->text_string.IsValid ()) {
		text->SetString (this->text_string);
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Return TRUE if the text has been fully or partially
 *	hilited.
 */

Bool
OPTextLine::IsHilited () const
{
	OP_METHOD (IsHilited, 0);
	return (this->hilite_pos_1 == this->hilite_pos_2) ? FALSE : TRUE;
}



/*
 *	Update the cursor's geometry, using current system
 *	font information.
 */

void
OPTextLine::UpdateCursorGeometry ()
{
	OP_METHOD (UpdateCursorGeometry, 0);
	
	Count len = this->text_string->CountCharacters ();
	
	this->text_dx     = this->ReturnTextIntervalWidth (0, len);
	this->cursor_x    = this->ReturnTextIntervalWidth (0, this->cursor_pos);
	this->hilite_x_1  = this->ReturnTextIntervalWidth (0, this->hilite_pos_1);
	this->hilite_x_2  = this->hilite_x_1;
	this->hilite_x_2 += this->ReturnTextIntervalWidth (this->hilite_pos_1, this->hilite_pos_2);
	
 	this->cursor_ascender  = this->ReturnTextAscender ();
	this->cursor_descender = this->ReturnTextDescender ();
}



/*
 *	Redraw just the cursor. This needs some bounding box
 *	computation, but it won't take too much time...
 */

void
OPTextLine::RedrawCursor ()
{
	OP_METHOD (RedrawCursor, 0);
	
	OPCoord frame = this->ReturnFrameSize ();
	
	OPCoord x1 = this->box.X1 () + frame + TEXT_BORDER_WIDTH + this->cursor_x + this->start_x;
	OPCoord y1 = this->box.Y1 () + frame + TEXT_BORDER_HEIGHT;
	OPCoord x2 = x1 + 1;
	OPCoord y2 = this->box.Y2 () - frame - TEXT_BORDER_HEIGHT;
	
	OPRect redraw (x1, y1, x2, y2);
	
	this->Display (redraw);
}



/*
 *	Return the width of a piece of text. This will use the
 *	system font's geometry.
 */

OPCoord
OPTextLine::ReturnTextIntervalWidth (Card32 pos_1, Card32 pos_2) const
{
	OP_METHOD (ReturnTextIntervalWidth, 2);
	OP_ARG (1, Card32, pos_1);
	OP_ARG (2, Card32, pos_2);
	
	Count len = this->text_string->CountCharacters ();
	
	if (pos_2 > len) {
		pos_2 = (Card32)(len);
	}
	
	if ( (pos_1 >= len)
	  || (pos_1 >= pos_2)
	  || (len == 0) ) {
		return 0;
	}
	
	//	Return size of interesting part of the string...
	
	return this->ReturnFont ()->ReturnWidth (this->text_string, pos_1, pos_2 - pos_1);
}



/*
 *	Return the font used to draw the characters in the
 *	text line.
 */

const OPGrailFontRef&
OPTextLine::ReturnFont () const
{
	OP_METHOD (ReturnFont, 0);
	return OPLookMan::Current ()->ReturnSysFont ();
}



/*
 *	Return the text's height over the base line. This is
 *	called the ascender and is usually a positive value.
 */

OPRelCoord
OPTextLine::ReturnTextAscender () const
{
	OP_METHOD (ReturnTextAscender, 0);
	return this->ReturnFont ()->ReturnAscender ();
}



/*
 *	Return the text's depth under the base line. This is
 *	called the descender and is usually a negative value.
 */

OPRelCoord
OPTextLine::ReturnTextDescender () const
{
	OP_METHOD (ReturnTextDescender, 0);
	return this->ReturnFont ()->ReturnDescender ();
}



/*
 *	Return position within line according to the mouse
 *	position.
 */

Card32
OPTextLine::ComputePos (OPEventIn event) const
{
	OP_METHOD (ComputePos, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPRelCoord delta;
	Card32     len = (Card32)(this->text_string->CountCharacters ());
	Card32     pos = 0;
	
	delta  = event->ReturnX () - this->box.X1 ();
	delta -= this->ReturnFrameSize ();
	delta -= TEXT_BORDER_WIDTH;
	delta -= this->start_x;
	
	if (delta <= 0) {
		return 0;
	}
	
	for (pos = 0; pos < len; pos++) {
		
		//	Compute width of one character.
		
		OPCoord    width = this->ReturnTextIntervalWidth (pos, pos + 1);
		OPRelCoord half  = (OPRelCoord)(width / 2);
		
		//	If before half of this character, position before. Otherwise, the cursor
		//	will be positioned after the character...
		
		if (delta < half) {
			break;
		}
		
		delta -= width;
	}
	
	return pos;
}



/*
 *	Move the cursor to the specified position. This will
 *	redraw the view if the cursor moved.
 */

Bool
OPTextLine::SetCursorPos (Card32 pos)
{
	OP_METHOD (SetCursorPos, 1);
	OP_ARG (1, Card32, pos);
	
	if (pos == this->cursor_pos) {
		return TRUE;
	}
	
	Count len = this->text_string->CountCharacters ();
	
	if (pos > len) {
		pos = len;
	}
	
	OPCoord old_x  = this->cursor_x;
	OPCoord new_x  = this->ReturnTextIntervalWidth (0, pos);
	
	this->cursor_blink = TRUE;
	this->cursor_x     = new_x;
	this->cursor_pos   = pos;
	
	this->UpdateCursorGeometry ();
	
	Bool    f_draw = FALSE;
	OPCoord border = this->ReturnFrameSize () + TEXT_BORDER;
	OPCoord min_x  = 0;
	OPCoord max_x  = this->box.DX () - 2 * border;
	OPCoord comp_x = this->start_x + new_x;
	
	if (comp_x < (min_x + TEXT_SCROLL)) {
		this->start_x += min_x - comp_x;
		this->start_x += TEXT_SCROLL;
		f_draw = TRUE;
	}

	if (comp_x > (max_x - TEXT_SCROLL)) {
		this->start_x -= comp_x - max_x;
		this->start_x -= TEXT_SCROLL;
		f_draw = TRUE;
	}
	
	if (this->start_x < max_x - this->text_dx - TEXT_SCROLL_RIGHT) {
		this->start_x = max_x - this->text_dx - TEXT_SCROLL_RIGHT;
		f_draw = TRUE;
	}
	
	if (this->start_x > 0) {
		this->start_x = 0;
		f_draw = TRUE;
	}

	if (f_draw) {
		this->Display ();
		return TRUE;
	}
	
	//	Just redraw the part where the cursor vanished and the part where it will
	//	appear soon.
	
	OPCoord x  = this->box.X1 () + this->ReturnFrameSize ()
				 + this->start_x + TEXT_BORDER_WIDTH;
	
	OPCoord x1 = x + ((old_x < new_x) ? old_x : new_x) + 0;
	OPCoord x2 = x + ((old_x > new_x) ? old_x : new_x) + 1;
	
	OPRect redraw (x1, this->box.Y1 (), x2, this->box.Y2 ());
	
	this->Display (redraw);
	return TRUE;
}



/*
 *	Define the highlighted part of the text line.
 */

Bool
OPTextLine::SetHiliteRange (Card32 pos_1, Card32 pos_2)
{
	OP_METHOD (SetHiliteRange, 2);
	OP_ARG (1, Card32, pos_1);
	OP_ARG (2, Card32, pos_2);
	
	if ( (this->hilite_pos_1 == pos_1)
	  && (this->hilite_pos_2 == pos_2) ) {
		return TRUE;
	}
	
	this->hilite_pos_1 = pos_1;
	this->hilite_pos_2 = pos_2;
	
	this->UpdateCursorGeometry ();

	//	Scroll the text if the cursor moved outside the visible
	//	range.
	
	OPCoord border = this->ReturnFrameSize () + TEXT_BORDER;
	OPCoord min_x  = 0;
	OPCoord max_x  = this->box.DX () - 2 * border;
	OPCoord comp_x = this->start_x + this->cursor_x;
	
	if (comp_x < (min_x + TEXT_SCROLL)) {
		this->start_x += min_x - comp_x;
		this->start_x += TEXT_SCROLL;
	}

	if (comp_x > (max_x - TEXT_SCROLL)) {
		this->start_x -= comp_x - max_x;
		this->start_x -= TEXT_SCROLL;
	}
	
	if (this->start_x < max_x - this->text_dx - TEXT_SCROLL_RIGHT) {
		this->start_x = max_x - this->text_dx - TEXT_SCROLL_RIGHT;
	}
	
	if (this->start_x > 0) {
		this->start_x = 0;
	}
	
	this->Display ();
	return TRUE;
}



/*
 *	Return current cursor position. This will always be in
 *	the range [0..length].
 */

Card32
OPTextLine::ReturnCursorPos () const
{
	OP_METHOD (ReturnCursorPos, 0);
	return this->cursor_pos;
}



/*
 *	Return the starting offset of the hiliting.
 */

Card32
OPTextLine::ReturnHiliteBegin () const
{
	OP_METHOD (ReturnHiliteBegin, 0);
	return this->hilite_pos_1;
}



/*
 *	Return the ending offset of the hiliting.
 */

Card32
OPTextLine::ReturnHiliteEnd () const
{
	OP_METHOD (ReturnHiliteEnd, 0);
	return this->hilite_pos_2;
}

/****************************************************************/

Bool
OPTextLine::SetFramelessMode (Bool yes)
{
	OP_METHOD (SetFramelessMode, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_frameless = yes;
	
	return TRUE;
}

Bool
OPTextLine::SetTransparentMode (Bool yes)
{
	OP_METHOD (SetTransparentMode, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_transparent = yes;
	
	return TRUE;
}

Bool
OPTextLine::ReturnFramelessMode () const
{
	OP_METHOD (ReturnFramelessMode, 0);
	return this->is_frameless;
}

Bool
OPTextLine::ReturnTransparentMode () const
{
	OP_METHOD (ReturnTransparentMode, 0);
	return this->is_transparent;
}


OPCoord
OPTextLine::ReturnFrameSize () const
{
	OP_METHOD (ReturnFrameSize, 0);
	
	return this->is_frameless
		   ? 0
		   : OPTextLine::look_3D_effect + OPTextLine::look_frame_size;
}

/****************************************************************/

/*
 *	Always accept to become the main responder view.
 */

Bool
OPTextLine::WouldAcceptMain () const
{
	OP_METHOD (WouldAcceptMain, 0);
	return TRUE;
}



/*
 *	Become main responder : this will start up a blinking
 *	cursor...
 */

void
OPTextLine::BecomeMain (Bool clicked)
{
	OP_METHOD (BecomeMain, 1);
	OP_ARG (1, Bool, clicked);
	
	OPTextLineRef self  = this;
	Card32        ticks = this->cursor_blink ? OPTextLine::look_cursor_on
											 : OPTextLine::look_cursor_off;
	
	this->EnableTick ();
	
 	this->cursor_ascender  = this->ReturnTextAscender ();
	this->cursor_descender = this->ReturnTextDescender ();
	
	if (clicked == FALSE) {
		this->hilite_pos_1 = 0;
		this->hilite_pos_2 = (Card32)(this->text_string->CountCharacters ());
		this->cursor_pos   = this->hilite_pos_2;
		this->UpdateCursorGeometry ();
		this->Display ();
	}
	
	if (this->cursor_blink == FALSE) {
		this->cursor_blink = TRUE;
		this->RedrawCursor ();
	}
	
	super::BecomeMain (clicked);
}



/*
 *	Resign main responder : stop the blinking cursor.
 */

void
OPTextLine::ResignMain ()
{
	OP_METHOD (ResignMain, 0);
	
	OPTextLineRef self = this;
	this->DisableTick ();
	
	if (this->cursor_blink == TRUE) {
		this->cursor_blink = FALSE;
		this->RedrawCursor ();
	}
	
	super::ResignMain ();
}



/*
 *	The "Action" method gets called after the "ActionOff"
 *	method if the user validated the action.
 */

void
OPTextLine::Action (Bool clicked)
{
	OP_METHOD (Action, 1);
	OP_ARG (1, Bool, clicked);
	
	OPTextLineRef self = this;
	this->Report (notify_OnClick, OP_REPORT_NEW);
	this->window->SetMainResponder (self, clicked);
}



/*
 *	Return the minimal size which will be accepted for
 *	any text line. This depends on the text height.
 */

void
OPTextLine::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	const OPGrailFontRef& font = this->ReturnFont ();
	
	OPRelCoord min_width  = font->ReturnWidth ("O");
	OPRelCoord min_height = font->ReturnAscender () - font->ReturnDescender ();
	
	OPCoord frame = this->ReturnFrameSize ();
	
	dx = (OPCoord)(min_width) + 2 * (frame + TEXT_BORDER_WIDTH + TEXT_SCROLL);
	dy = (OPCoord)(min_height) + 2 * (frame + TEXT_BORDER_HEIGHT);
}



/*
 *	Draw the text line.
 */

void
OPTextLine::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		Count   len    = this->text_string->CountCharacters ();
		OPCoord frame  = this->ReturnFrameSize ();
		
		OPCoord x  = this->box.X1 () + frame + this->start_x + TEXT_BORDER_WIDTH;
		OPCoord y  = this->box.Y1 () + frame + TEXT_BORDER_HEIGHT + this->cursor_ascender;
		OPCoord dy = 0;
		
		if (this->is_frameless == FALSE) {
			adorner->DrawLineFrame (this->box, OPTextLine::look_frame_size);
		}
		if (this->is_transparent == FALSE) {
			adorner->FillRect (this->box, OP_COLOR_WHITE, frame);
		}
		
		adorner->SelectSysFont ();
		
		OPRect new_clip (this->box, frame + TEXT_BORDER);
		OPRect old_clip;
		
		port->GetClipRect (old_clip);
		
		new_clip *= old_clip;
		
		port->SetClipRect (new_clip);
		
		if ( (this->hilite_pos_1 == this->hilite_pos_2)
		  || ( ((this->state & OP_VIEW_IS_ACTIVE) == 0)
			&& (this->is_dragging == FALSE) ) ) {
			
			//	No highlighting to be done : this is easy. We just draw the string
			//	in one piece.
			
			adorner->SetForeColor (OP_COLOR_BLACK);
			port->ShowText (x, y, this->text_string, 0, len);
			
			//	Display cursor...
			
			if ( (this->cursor_blink)
			  && (this->cursor_enable)
			  && (this->state & OP_VIEW_IS_ACTIVE)
			  && (this->window->IsMainResponderOk ())
			  && (this->window->IsActive ()) ) {
				
				y  = (OPCoord)(y - this->cursor_ascender);
				dy = (OPCoord)(this->cursor_ascender - this->cursor_descender);
				
				adorner->SetForeColor (OP_COLOR_VIEW_CURSOR);
				port->DrawLineDY (x + this->cursor_x, y + 1, dy - 2);
			}
			
		} else {
			
			//	There is some highlighting. Possibly display the string in three
			//	separate parts. Also draw the highlighting rectangle...
			
			OPRect hilite (x + this->hilite_x_1, y - this->cursor_ascender + 1,
						   x + this->hilite_x_2 - 1, y - this->cursor_descender - 2);
			
			adorner->FillRect (hilite, OP_COLOR_HILITE_SELECT);
			adorner->SetForeColor (OP_COLOR_VIEW_HILITE);
			
			port->ShowText (x + this->hilite_x_1, y, this->text_string,
							this->hilite_pos_1, this->hilite_pos_2 - this->hilite_pos_1);
			
			adorner->SetForeColor (OP_COLOR_BLACK);
			
			if (this->hilite_pos_1) {
				port->ShowText (x, y, this->text_string, 0, this->hilite_pos_1);
			}
			
			if (this->hilite_pos_2 != len) {
				port->ShowText (x + this->hilite_x_2, y, this->text_string,
								this->hilite_pos_2, len - this->hilite_pos_2);
			}
		}
		
		port->SetClipRect (old_clip);
	}
}



/*
 *	Pressed a button in the view.
 */

void
OPTextLine::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked == FALSE)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		OPTextLineRef self = this;
		Card32        pos  = this->ComputePos (event);
		
		this->temp_pos    = pos;
		this->is_dragging = TRUE;
		
		this->window->SetDraggingView (self);
		this->SetHiliteRange (0, 0);
		this->SetCursorPos (pos);
		this->ActionOn ();
	}
}



/*
 *	Released a button in the view. This will stop any
 *	dragging.
 */

void
OPTextLine::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
	if ( (this->state & OP_VIEW_IS_ENABLED)
	  && (this->is_being_clicked)
	  && (event->ReturnChange () == OP_EV_BUTTON_LEFT) ) {
		
		OPTextLineRef nobody = 0;
		
		this->cursor_enable = TRUE;
		this->is_dragging   = FALSE;
		this->cursor_pos    = this->temp_pos;
		
		this->window->SetDraggingView (nobody);
		this->ActionOff ();
		this->Action (TRUE);
	}
}



/*
 *	Dragging in a text line will select a piece of text.
 */

void
OPTextLine::DragNotify (OPEventIn event)
{
	OP_METHOD (DragNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (this->is_being_clicked) {
		
		Card32 pos  = this->ComputePos (event);
		Card32 tpos = this->cursor_pos;
		
		this->temp_pos = pos;
		
		if ( (this->cursor_enable == TRUE)
		  && (this->cursor_pos == pos) ) {
			return;
		}
		
		Card32 pos_1 = (pos < this->cursor_pos) ? pos : this->cursor_pos;
		Card32 pos_2 = (pos > this->cursor_pos) ? pos : this->cursor_pos;
		
		if ( (pos_1 == this->hilite_pos_1)
		  && (pos_2 == this->hilite_pos_2) ) {
			return;
		}
		
		this->cursor_enable = FALSE;
		this->cursor_pos = pos;
		this->SetHiliteRange (pos_1, pos_2);
		this->cursor_pos = tpos;
		this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
	}
}



/*
 *	When entering a text line, we change the cursor to a
 *	so-called "I beam", which is useful for clicking into
 *	text lines.
 */

void
OPTextLine::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPLookMan::Current ()->SelectCursor (OP_CURSOR_I_BEAM);
}



/*
 *	On exit, just restore the default arrow.
 */

void
OPTextLine::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);
	
	OPLookMan::Current ()->SelectCursor (OP_CURSOR_NORMAL);
}


/*
 *	Handle incoming characters. The following cases are
 *	considered :
 *
 *	- Cursor Left, Cursor Right, Home and End.
 *	- Shift Cursor Left, Shift Cursor Right, Shift Home
 *	  and Shift End
 *	- Backspace and Delete.
 *	- Any alphabetical character.
 */

void
OPTextLine::KeyDown (OPEventIn event)
{
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);
	
	if (event->ReturnKey ().type == OP_KEY_CURSOR) {
		
		//	Moving around using the cursor...
		
		Bool   ok    = FALSE;
		Card32 pos   = this->cursor_pos;
		Card32 old   = pos;
		Count  len   = this->text_string->CountCharacters ();
		Card32 shift = (event->ReturnKey().modifiers & OP_EV_MOD_MASK);
		
		switch (event->ReturnKey ().code) {
			case OP_CURSOR_LEFT:  ok = TRUE; if (pos) pos--;		break;
			case OP_CURSOR_RIGHT: ok = TRUE; if (pos < len) pos++;	break;
			case OP_CURSOR_HOME:  ok = TRUE; pos = 0;				break;
			case OP_CURSOR_END:   ok = TRUE; pos = (Card32)(len);	break;
		}
		
		if ( (ok == TRUE) && (shift == OP_EV_SHIFT) ) {
			
			//	Shift being pressed will highlight part of the text.
			
			Card32 pos_1 = this->hilite_pos_1;
			Card32 pos_2 = this->hilite_pos_2;
			
			if (pos_1 == pos_2) {
				pos_1 = old;
				pos_2 = pos;
			} else if (old == pos_1) {
				pos_1 = pos;
			} else {
				pos_2 = pos;
			}
			
			if (pos_1 > pos_2) {
				int p = pos_1;
				pos_1 = pos_2;
				pos_2 = p;
			}
			
			this->cursor_pos = pos;
			this->SetHiliteRange (pos_1, pos_2);
			
		} else if (ok) {
			
			if (this->hilite_pos_1 != this->hilite_pos_2) {
				
				//	Shift not being pressed, but selection has been made. This will
				//	remove the selection.
				
				this->hilite_pos_1 = 0;
				this->hilite_pos_2 = 0;
				this->cursor_pos   = pos;
				
				this->UpdateCursorGeometry ();
				this->Display ();
				this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
				
			} else {
				this->SetCursorPos (pos);
			}
		}
		
	} else if ( (event->ReturnKey ().type == OP_KEY_BACKSPACE)
			 || (event->ReturnKey ().type == OP_KEY_DELETE) ) {
		
		Card32 pos = this->cursor_pos;
		Card32 len = 1;
		
		if (this->hilite_pos_1 != this->hilite_pos_2) {
			
			pos = this->hilite_pos_1;
			len = this->hilite_pos_2 - this->hilite_pos_1;
			
			this->hilite_pos_1 = 0;
			this->hilite_pos_2 = 0;
			
			this->UpdateCursorGeometry ();
			
		} else if (event->ReturnKey ().type == OP_KEY_BACKSPACE) {
			
			//	Backspace will delete the previous character, not the current
			//	one !
			
			if (pos == 0) {
				return;
			}
			
			pos--;
		}
		
		this->text_string->DeleteString (pos, len);
		this->SetCursorPos (pos);
		this->Display ();
		this->NotifyTextChange (TRUE);
		this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
		
	} else if ( (event->ReturnKey ().type == OP_KEY_ALPHANUM)
			 && (event->ReturnKey ().value) ) {
		
		if (this->hilite_pos_1 != this->hilite_pos_2) {
			
			//	The selected part will be replaced by the new character.
			
			Card32 pos = this->hilite_pos_1;
			Card32 len = this->hilite_pos_2 - this->hilite_pos_1;
			
			this->text_string->DeleteString (pos, len);
			
			this->cursor_pos   = this->hilite_pos_1;
			this->hilite_pos_1 = 0;
			this->hilite_pos_2 = 0;
			
			this->UpdateCursorGeometry ();
		}
		
		LatinChar str[2] = { (LatinChar)(event->ReturnKey ().value), 0 };
		
		this->text_string->InsertString (this->cursor_pos, str);
		this->SetCursorPos (this->cursor_pos + 1);
		this->NotifyTextChange (TRUE);
		this->Display ();
		this->ReturnAdorner ()->Flush (OP_FLUSH_SYNC);
		
	} else {
	
		//	Key not handled : let the super class handle the event !
		
		super::KeyDown (event);
	}
}



/*
 *	Every once in a while, change the cursor's state and
 *	redraw the part which changed since the last time.
 */

void
OPTextLine::Tick ()
{
	static int recursion = 0;
	
	recursion++;
	
	Card32 ticks = this->cursor_blink ? OPTextLine::look_cursor_on
									  : OPTextLine::look_cursor_off;
	
	if (recursion == 1) {
		if (this->is_app_ticking) {
			this->app->InsertTickResponder (this, ticks);
		}
	}
	
	if (this->hilite_pos_1 == this->hilite_pos_2) {
		this->cursor_blink = (this->cursor_blink) ? FALSE : TRUE;
		this->RedrawCursor ();
	}
	
	recursion--;
}



/*
 *	The Look Manager notifies the view instances whenever
 *	the user changed its look ! The view need not redraw
 *	itself now, since the window will do it automatically.
 */

void
OPTextLine::Broadcast (OPAtomIn name, OPObjectIn param)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, name);
	OP_ARG (1, OPObjecIn, param);
	
	super::Broadcast (name, param);
	
	if (name == this->atom_look_man_changed) {
	
		OPLookManRef look = param;
		
		OPTextLine::look_3D_effect  = look->Return3DEffectWidth ();
		OPTextLine::look_frame_size = look->ReturnWidgetFrameSize ();
		OPTextLine::look_cursor_off = look->ReturnCursorOffTime ();
		OPTextLine::look_cursor_on  = look->ReturnCursorOnTime ();
	
		this->UpdateCursorGeometry ();
	}
}
