/*
 *	texttagb.cxx
 *
 *	Base Text Tag class. This is used by OPTextTag, OPLed7Display
 *	and others...
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/texttag.h>
#include <OPaC/new.h>
#include <OPaC/string.h>
#include <OPaC/targetmacro.h>
#include <OPaC/pubmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPTextTagBase);

OP_TARGET_CLASS (OPTextTagBase);
OP_TARGET_ARG (OPStringRef, text);
OP_TARGET_CALL (SetText (text));
OP_TARGET_END ();

/****************************************************************/

OP_PUBLISH_MODIFIER (OPTextTagBase, "Caption", GetText, SetText, "Text's title", OPObject, OPString);

/****************************************************************/

/*
 *	Create an empty text tag. No text is currently defined.
 */

OPTextTagBase::OPTextTagBase ()
{
	OP_METHOD (OPTextTagBase, 0);
	
	this->state &= ~ (OP_VIEW_OK_DETECT			//	may not be detected (click)
					| OP_VIEW_OK_MAIN			//	may not become main responder
					| OP_VIEW_IS_ALIVE);		//	may not become active
	
	this->text = OPStringNull;
}


/* 
 *	Destroy a text tag. This is easy.
 */

OPTextTagBase::~OPTextTagBase ()
{
	OP_METHOD (~OPTextTagBase, 0);
	
	this->text = OPStringNull;
}



/*
 *	Delete a text tag : prepare for automatic destruction
 *	by the garbage collector.
 */

void
OPTextTagBase::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->text = OPStringNull;
	
	super::Delete ();
}



/*
 *	Store the text tag.
 */

void
OPTextTagBase::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->text;
	storage << OPStorageSync::mark;
}



/*
 *	Restore the text tag : string, font and mode plus the
 *	view's definition.
 */

void
OPTextTagBase::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->text;
	storage >> OPStorageSync::mark;
}


/*
 *	Return a pointer to the object as a view. This is needed
 *	by class OPxTitleWidget.
 */

OPView*
OPTextTagBase::ReturnObjectAsView ()
{
	OP_METHOD (ReturnObjectAsView, 0);
	return this;
}


/*
 *	Define the text to display within the text tag. This
 *	will not redraw the text tag...
 */

Bool
OPTextTagBase::SetText (OPStringIn text)
{
	OP_METHOD (SetText, 1);
	OP_ARG (1, OPStringIn, text);
	
	this->text = text;
	
	return TRUE;
}


/*
 *	Same as "SetText", used by class OPxTitleWidget.
 */

Bool
OPTextTagBase::SetTitle (OPStringIn text)
{
	OP_METHOD (SetTitle, 1);
	OP_ARG (1, OPStringIn, text);
	
	return this->SetText (text);
}



/*
 *	Return the text displayed by the text tag.
 */

Bool
OPTextTagBase::GetText (OPStringOut text) const
{
	OP_METHOD (GetText, 1);
	OP_ARG (1, OPStringOut, text);
	
	text = "";
	
	if (this->text.IsValid ()) {
		text->SetString (this->text);
		return TRUE;
	}
	
	return FALSE;
}

