/*
 *	values.cxx
 *
 *	Encapsulate values (numbers, integers, booleans, reals, etc.)
 *	so that they can be handled as standard OPaC objects.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/convert.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/string.h>
#include <OPaC/values.h>

#include <string.h>

OP_IMPLEMENTATION (values.h);
OP_IMPLEMENT_CLASS (OPValue);

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPValue, "Card32", ReturnCard32, SetCard32, "Cardinal value", Card32, Card32);
OP_PUBLISH_MODIFUNC (OPValue, "Int32",  ReturnInt32,  SetInt32,  "Integer value",  Int32, Int32);
OP_PUBLISH_MODIFUNC (OPValue, "Bool",   ReturnBool,   SetBool,   "Boolean value",  Bool, Bool);
OP_PUBLISH_MODIFUNC (OPValue, "Real",   ReturnReal,   SetReal,   "Real value",     Real, Real);

/****************************************************************/

/*
 *	
 */

OPValue::OPValue ()
{
	OP_METHOD (OPValue, 0);
	
	this->type = OP_TYPE_NONE;
	this->auto_format = OPStringNull;
}



/*
 *
 */

OPValue::~OPValue ()
{
	OP_METHOD (~OPValue, 0);
	
}



void
OPValue::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);

	this->RefreshValue ();
	
	storage << this->type;
	storage << this->auto_format;
	storage << OPStorageSync::mark;

	switch (this->type) {
		case OP_TYPE_CARD:
		case OP_TYPE_ENUM:
			storage << this->value.the_card32;
			break;
		
		case OP_TYPE_INT:
			storage << this->value.the_int32;
			break;
		
		case OP_TYPE_BOOL:
			storage << this->value.the_bool;
			break;
		
		case OP_TYPE_REAL:
			storage << this->value.the_real;
			break;
		
		case OP_TYPE_OBJECT:
		case OP_TYPE_TEXT:
		case OP_TYPE_ARRAY:
			storage << this->object;
			break;
	}
	
	storage << OPStorageSync::mark;
}

void
OPValue::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	this->object = 0;
	
	storage >> this->type;
	storage >> this->auto_format;
	storage >> OPStorageSync::mark;
	
	switch (this->type) {
		case OP_TYPE_CARD:
		case OP_TYPE_ENUM:
			storage >> this->value.the_card32;
			break;
		
		case OP_TYPE_INT:
			storage >> this->value.the_int32;
			break;
		
		case OP_TYPE_BOOL:
			storage >> this->value.the_bool;
			break;
		
		case OP_TYPE_REAL:
			storage >> this->value.the_real;
			break;
		
		case OP_TYPE_OBJECT:
		case OP_TYPE_TEXT:
		case OP_TYPE_ARRAY:
			storage >> this->object;
			break;
	}
	
	storage >> OPStorageSync::mark;
	
	this->ChangedValue ();
}



/*
 *	Refresh the value. This is useful for subclasses
 *	which modify the value "on the fly". Override this
 *	method to do something meaningful.
 */

void
OPValue::RefreshValue ()
{
	OP_METHOD (RefreshValue, 0);
}


/*
 *	Provide some means to handle a change of the internal
 *	value for smart subclasses. Override this method to
 *	do something meaningful.
 */

void
OPValue::ChangedValue ()
{
	OP_METHOD (ChangedValue, 0);
}


/*
 *	Return a hashed form of the value so that it can be
 *	stored in a hash table.
 */

Count
OPValue::ComputeHashValue (Count num) const
{
	OP_METHOD (ComputeHashValue, 1);
	OP_ARG (1, Count, num);
	
	return super::ComputeHashValue (num);
}


static int compare_real (Real a, Real b)
{
	if (a == b) {
		return OP_COMP_THIS_SAME;
	} else if (a < b) {
		return OP_COMP_THIS_BEFORE;
	} else {
		return OP_COMP_THIS_AFTER;
	}
}

/*
 *	Return result of comparison.
 */

int
OPValue::CompareWith (OPObjectIn value) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, OPObjectIn, value);
	
	if (value.IsValid ()) {
		
		if (value->IsKindOf (OP_RUNTIME (OPValue))) {
			
			OPValueRef x = value;
			
			Real a = this->ToReal ();
			Real b = x->ToReal ();
			
			return compare_real (a, b);
			
		} else {
			return super::CompareWith (value);
		}
	}

	return OP_COMP_THIS_AFTER;
}

int
OPValue::CompareWith (Card32 x) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, Card32, x);
	
	return compare_real (this->ToReal (), (Real)(x));
}

int
OPValue::CompareWith (Int32 x) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, Int32, x);
	
	return compare_real (this->ToReal (), (Real)(x));
}

int
OPValue::CompareWith (Bool x) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, Bool, x);
	
	return compare_real (this->ToReal (), (Real)(x));
}

int
OPValue::CompareWith (Real x) const
{
	OP_METHOD (CompareWith, 1);
	OP_ARG (1, Real, x);
	
	return compare_real (this->ToReal (), (Real)(x));
}



Real
OPValue::ToReal () const
{
	OP_METHOD (ToReal, 0);
	
	//	This is an infringement to the C++ ethics... but the user has
	//	not to know about the internal changes which might happen...
	
	//	NB: There are other const_cast<> expressions below.
	
	OPValue* that = const_cast<OPValue*> (this);
	that->RefreshValue ();
	
	switch (this->type) {
		case OP_TYPE_CARD:
			return (Real)(this->value.the_card32);
		case OP_TYPE_INT:
			return (Real)(this->value.the_int32);
		case OP_TYPE_BOOL:
			return (Real)(this->value.the_bool);
		case OP_TYPE_REAL:
			return (Real)(this->value.the_real);
	}
	
	return 0;
}


Bool
OPValue::SetCard32 (Card32 x)
{
	OP_METHOD (SetCard32, 1);
	OP_ARG (1, Card32, x);
	
	this->type = OP_TYPE_CARD;
	this->value.the_card32 = x;
	this->object = 0;
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetInt32 (Int32 x)
{
	OP_METHOD (SetInt32, 1);
	OP_ARG (1, Int32, x);
	
	this->type = OP_TYPE_INT;
	this->value.the_int32 = x;
	this->object = 0;
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetBool (Bool x)
{
	OP_METHOD (SetBool, 1);
	OP_ARG (1, Bool, x);
	
	this->type = OP_TYPE_BOOL;
	this->value.the_bool = x;
	this->object = 0;
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetReal (Real x)
{
	OP_METHOD (SetReal, 1);
	OP_ARG (1, Real, x);
	
	this->type = OP_TYPE_REAL;
	this->value.the_real = x;
	this->object = 0;
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetObject (OPObjectIn x)
{
	OP_METHOD (SetObject, 1);
	OP_ARG (1, OPObjectIn, x);
	
	this->type = OP_TYPE_OBJECT;
	this->object = x;
	
	if (x.IsValid ()) {
		if (x->IsKindOf (OP_RUNTIME (OPString))) {
			this->type = OP_TYPE_TEXT;
		} else if (x->IsKindOf (OP_RUNTIME (OPArray))) {
			this->type = OP_TYPE_ARRAY;
		} else if (x->IsKindOf (OP_RUNTIME (OPValue))) {
			OPValue* value = (OPValue*)x.ReturnObject ();
			this->type        = value->type;
			this->auto_format = value->auto_format;
			this->value       = value->value;
			this->object      = value->object;
		}
	}
	
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetText (OPTextIn x)
{
	OP_METHOD (SetText, 1);
	OP_ARG (1, OPTextIn, x);
	
	this->type = OP_TYPE_TEXT;
	this->object = x;
	this->ChangedValue ();
	
	return TRUE;
}

Bool
OPValue::SetArray (OPArrayIn x)
{
	OP_METHOD (SetArray, 1);
	OP_ARG (1, OPArrayIn, x);
	
	this->type = OP_TYPE_ARRAY;
	this->object = x;
	this->ChangedValue ();
	
	return TRUE;
}


Bool
OPValue::SetAutoFormat (OPStringIn format)
{
	OP_METHOD (SetAutoFormat, 1);
	OP_ARG (1, OPStringIn, format);
	
	this->auto_format = format;
	
	return TRUE;
}

const OPStringRef&
OPValue::ReturnAutoFormat () const
{
	OP_METHOD (ReturnAutoFormat, 0);
	
	return this->auto_format;
}




Bool
OPValue::ConvertToString (OPStringOut text, OPStringIn format) const
{
	OP_METHOD (ConvertToString, 2);
	OP_ARG (1, OPStringOut, text);
	OP_ARG (2, OPStringIn, format);
	
	OPStringRef fmt = "";
	OPValue* that = const_cast<OPValue*> (this);
	that->RefreshValue ();
	
	switch (this->type) {
		case OP_TYPE_CARD:
			OPConvert::CardToString (this->value.the_card32, fmt, text);
			return TRUE;
		case OP_TYPE_INT:
			OPConvert::IntToString (this->value.the_int32, fmt, text);
			return TRUE;
		case OP_TYPE_BOOL:
			OPConvert::BoolToString (this->value.the_bool, fmt, text);
			return TRUE;
		case OP_TYPE_REAL:
			OPConvert::RealToString (this->value.the_real, fmt, text);
			return TRUE;
		case OP_TYPE_OBJECT:
		case OP_TYPE_TEXT:
			if (this->object.IsValid () == FALSE) {
				text = "";
				return TRUE;
			} else if (this->object->IsKindOf (OP_RUNTIME (OPString))) {
				OPStringRef string = this->object;
				return string->GetString (text);
			} else if (this->object->IsKindOf (OP_RUNTIME (OPAtom))) {
				OPAtomRef atom = this->object;
				text = atom;
				return TRUE;
			}
			return FALSE;
	}

	return FALSE;
}

Bool
OPValue::ConvertToAtom (OPAtomOut atom, OPStringIn format) const
{
	OP_METHOD (ConvertToAtom, 2);
	OP_ARG (1, OPTextOut, atom);
	OP_ARG (2, OPStringIn, format);
	
	OPStringRef text;
	
	if (this->ConvertToString (text, format)) {
		atom = text;
		return TRUE;
	}
	
	return FALSE;
}


static Card32
convert_text_to_Card32 (OPTextIn text)
{
	OPStringRef string = text->ReturnString ();
	Card32 x = 0;
	OPConvert::StringToCard (string, x);
	return x;
}

static Int32
convert_text_to_Int32 (OPTextIn text)
{
	OPStringRef string = text->ReturnString ();
	Int32 x = 0;
	OPConvert::StringToInt (string, x);
	return x;
}

static Real
convert_text_to_Real (OPTextIn text)
{
	OPStringRef string = text->ReturnString ();
	Real x = 0;
	OPConvert::StringToReal (string, x);
	return x;
}

static Bool
convert_text_to_Bool (OPTextIn text)
{
	LatinChar buffer[6];
	LatinChar* ptr = buffer;
	OPStringRef string = text->ReturnString ();
	
	string->ExtractStringCopy (buffer, 0, 5);
	
	while (*ptr) {
		if ((*ptr >= 'A') && (*ptr <= 'Z')) *ptr += 'a' - 'A';
		ptr++;
	}
	
	if (strncmp (buffer, "false", 5) == 0) {
		return FALSE;
	} else if (strncmp (buffer, "true", 4) == 0) {
		return TRUE;
	}
	
	Card32 x = 0;
	OPConvert::StringToCard (string, x);
	
	return x ? TRUE : FALSE;
}


#define	implement_type(T)														\
																				\
	OPValue* that = const_cast<OPValue*> (this);								\
	that->RefreshValue ();														\
																				\
	switch (this->type) {														\
		case OP_TYPE_CARD:														\
			return (T)(this->value.the_card32);									\
		case OP_TYPE_INT:														\
			return (T)(this->value.the_int32);									\
		case OP_TYPE_BOOL:														\
			return (T)(this->value.the_bool);									\
		case OP_TYPE_REAL:														\
			return (T)(this->value.the_real);									\
		case OP_TYPE_TEXT:														\
			return convert_text_to_##T (this->object);							\
	}																			\
	return 0

Card32
OPValue::ReturnCard32 () const
{
	OP_METHOD (ReturnCard32, 0);
	implement_type (Card32);
}

Int32
OPValue::ReturnInt32 () const
{
	OP_METHOD (ReturnInt32, 0);
	implement_type (Int32);
}

Bool
OPValue::ReturnBool () const
{
	OP_METHOD (ReturnBool, 0);
	implement_type (Bool);
}

Real
OPValue::ReturnReal () const
{
	OP_METHOD (ReturnReal, 0);
	implement_type (Real);
}

const OPObjectRef&
OPValue::ReturnObject () const
{
	OP_METHOD (ReturnObject, 0);
	
	if ( (this->type == OP_TYPE_OBJECT)
	  || (this->type == OP_TYPE_TEXT)
	  || (this->type == OP_TYPE_ARRAY) ) {
		return this->object;
	}
	
	return OPObjectNull;
}

const OPObjectRef&
OPValue::ReturnText () const
{
	OP_METHOD (ReturnText, 0);
	
	if (this->type == OP_TYPE_TEXT) {
		return this->object;
	}
	
	return OPObjectNull;
}

const OPObjectRef&
OPValue::ReturnArray () const
{
	OP_METHOD (ReturnArray, 0);
	
	if (this->type == OP_TYPE_ARRAY) {
		return this->object;
	}
	
	return OPObjectNull;
}



OPSimpleType
OPValue::ReturnType () const
{
	OP_METHOD (ReturnValueType, 0);
	return this->type;
}


void
OPValueRef::NewValue ()
{
	(*this) = OP_NEW (OPValue);
}

