/*
 *	buildedit.cxx
 *
 *	Edition tools for the interface builder.
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "buildedit.h"
#include "bitsetspy.h"
#include "conviewer.h"
#include "dlinkviewer.h"

#include <OPaC/boxview.h>
#include <OPaC/button.h>
#include <OPaC/cellgrid.h>
#include <OPaC/database.h>
#include <OPaC/datarep.h>
#include <OPaC/message.h>
#include <OPaC/new.h>
#include <OPaC/object.h>
#include <OPaC/objectpane.h>
#include <OPaC/objectwell.h>
#include <OPaC/pubmacro.h>
#include <OPaC/scrgrid.h>
#include <OPaC/selgrid.h>
#include <OPaC/switch.h>
#include <OPaC/targetmacro.h>
#include <OPaC/tablist.h>
#include <OPaC/tabview.h>
#include <OPaC/textcaption.h>
#include <OPaC/textline.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

APP_IMPLEMENT_CLASS (OPBuildEdit);

OP_TARGET_ACTOR (OPBuildEdit, Update);
OP_TARGET_ACTOR (OPBuildEdit, EditOK);

OP_TARGET_CLASS (OPBuildEdit);
OP_TARGET_ARG (OPDataReportRef, reporter);
OP_TARGET_CALL (AddReporter (reporter));
OP_TARGET_END ();

/****************************************************************/

#define	OP_TYPE_O_TEXTTAG	(OP_TYPE_USER+0)
#define	OP_TYPE_O_ICON		(OP_TYPE_USER+1)
#define	OP_TYPE_O_VIEW		(OP_TYPE_USER+2)
#define	OP_TYPE_BITSET		(OP_TYPE_USER+3)

/****************************************************************/

OPBuildEdit::OPBuildEdit ()
{
}

OPBuildEdit::~OPBuildEdit ()
{
}

/****************************************************************/

/*
 *	Find the simple type for a given object field (the type and
 *	has been obtained through "GetPublicData", with the specified
 *	index).
 */

OPSimpleType
OPBuildEdit::FindSimpleType (OPTypeDef type, OPObjectIn object, Card32 index)
{
	OP_MEMBER (FindSimpleType, 3);
	OP_ARG (1, OPTypeDef, type);
	OP_ARG (2, OPObjectIn, object);
	OP_ARG (3, Card32, index);
	
	if (type->simple == OP_TYPE_OBJECT) {
		
		OPRuntimeRef runtime = *((const OPRuntimeRef*)(type->runtime));
		OPObjectRef  data;
		
		//	If the type describes an object, check to see whether it is a text
		//	and, if it is, return the TEXT type instead of the OBJECT type.
		
		if (runtime->IsKindOf (OP_RUNTIME (OPText))) {
			return OP_TYPE_TEXT;
		} else if (runtime->IsKindOf (OP_RUNTIME (OPTextTagBase))) {
			return OP_TYPE_O_TEXTTAG;
		} else if (runtime->IsKindOf (OP_RUNTIME (OPIcon))) {
			return OP_TYPE_O_ICON;
		} else if (runtime->IsKindOf (OP_RUNTIME (OPView))) {
			return OP_TYPE_O_VIEW;
		}
		
		//	Look if the real data is something else than a generic object, as
		//	defined by the type. Perhaps "OPObject" hides a string...
		
		if ( (object->QueryData (index, data))
		  && (data.IsValid ()) ) {
			
			if (data->IsKindOf (OP_RUNTIME (OPText))) {
				return OP_TYPE_TEXT;
			} else if (data->IsKindOf (OP_RUNTIME (OPTextTagBase))) {
				return OP_TYPE_O_TEXTTAG;
			} else if (data->IsKindOf (OP_RUNTIME (OPIcon))) {
				return OP_TYPE_O_ICON;
			} else if (data->IsKindOf (OP_RUNTIME (OPView))) {
				return OP_TYPE_O_VIEW;
			} else {
				//	Some other objects might be handled in a special way.
			}
		}
		
	} else if (type->simple == OP_TYPE_CARD) {
		
		const OPaC_BitsetDataDef* bitset_def = 0;
		Card32                    bitset_num = 0;
		
		if (object->GetPublicDataEx (index, 0, bitset_def, bitset_num)) {
			return OP_TYPE_BITSET;
		}
	}
	
	return type->simple;
}

/****************************************************************/

Bool
OPBuildEdit::HasAttachedTarget (OPObjectIn edit, OPAtomIn find)
{
	Count num   = edit->CountAttached ();
	Count index = 0;
	
	while (index < num) {
		
		OPAtomRef   event;
		OPAtomRef   actor;
		OPObjectRef observer;
		Card32      id;
		
		edit->GetAttached (index, event, actor, observer, id);
		
		if ( (find == event)
		  && (observer->IsKindOf (OP_RUNTIME (OPDataReport)) == FALSE) ) {
			return TRUE;
		}
		
		index++;
	}
	
	return FALSE;
}

Bool
OPBuildEdit::HasAttachedDataTarget (OPObjectIn edit, OPAtomIn find, OPAtomIn notifier)
{
	Count num   = edit->CountAttached ();
	Count index = 0;
	
	while (index < num) {
		
		OPAtomRef   event;
		OPAtomRef   actor;
		OPObjectRef observer;
		Card32      id;
		
		edit->GetAttached (index, event, actor, observer, id);
		
		if ( (notifier == event)
		  && (observer->IsKindOf (OP_RUNTIME (OPDataReport))) ) {
			OPDataReportRef reporter = observer;
			if ( (reporter->ReturnQueryName () == find)
			  && (reporter->IsVisible ()) ) {
				return TRUE;
			}
		}
		
		index++;
	}
	
	return FALSE;
}


void
OPBuildEdit::MakeEventList (OPDBScrollListIn list, OPObjectIn edit)
{
	list->RemoveAll ();
	
	OPRuntimeRef runtime = edit->ReturnRuntime ();
	
	while (runtime != OP_RUNTIME (OPObject)) {
		
		OPAtomRef name;
		Count     count = 0;
		
		while (runtime->GetNotify (count++, name)) {
			
			Bool dot = HasAttachedTarget (edit, name);
			
			list->Insert (name, name, edit, name,
						  OP_ICON_NONE, OP_ICON_NONE,
						  dot ? OP_ICON_SLIDOT_NORMAL : OP_ICON_NONE);
		}
		
		runtime = runtime->ReturnSuper ();
	}
	
	list->Update ();
}

void
OPBuildEdit::MakeTargetList (OPDBScrollListIn list, OPObjectIn edit)
{
	list->RemoveAll ();
	
	OPRuntimeRef runtime = edit->ReturnRuntime ();
	
	while (runtime != OP_RUNTIME (OPObject)) {
		
		OPTargetType       type   = 0;
		const OPaC_Target* target = 0;
		
		OPAtomRef name;
		Count     count = 0;
		
		while (runtime->GetTarget (count++, name, type, target)) {
			list->Insert (name, name, edit, name);
		}
		runtime = runtime->ReturnSuper ();
	}
	
	list->Update ();
}

void
OPBuildEdit::MakeDataSourceList (OPDBScrollListIn list, OPObjectIn edit)
{
	list->RemoveAll ();
	
	OPAtomRef   name;
	OPStringRef desc;
	OPTypeDef   type;
	
	Bool  is_read_ok;
	Bool  is_write_ok;
	Count count = 0;
	
	while (edit->GetPublicData (count++, name, desc, type, is_read_ok, is_write_ok)) {
		if (is_read_ok) {
			OPAtomRef notifier;
			Count     nth = 0;
			
			while (edit->GetPublicDataNotifier (name, nth++, notifier)) {
				
				OPStringRef text = name;
				
				text->AppendString ("/");
				text->AppendString (notifier);
				text->SetAttribute ("Notifier", notifier);
				
				Bool dot = HasAttachedDataTarget (edit, name, notifier);
				
				list->Insert (text, name, edit, text,
							  OP_ICON_NONE, OP_ICON_NONE,
							  dot ? OP_ICON_SLIDOT_NORMAL : OP_ICON_NONE);
			}
		}
	}
	
	list->Update ();
}

void
OPBuildEdit::MakeDataTargetList (OPDBScrollListIn list, OPObjectIn edit)
{
	list->RemoveAll ();
	
	OPAtomRef   name;
	OPStringRef desc;
	OPTypeDef   type;
	
	Bool  is_read_ok;
	Bool  is_write_ok;
	Count count = 0;
	
	while (edit->GetPublicData (count++, name, desc, type, is_read_ok, is_write_ok)) {
		if (is_write_ok) {
			list->Insert (name, name, edit, name);
		}
	}
	
	list->Update ();
}


void
OPBuildEdit::MakeAttachedList (OPViewIn parent, OPObjectIn edit, OPStringIn title, OPViewOut view)
{
	OPDBScrollListRef list = OP_NEW (OPDBScrollList);
	
	list->SetParent (parent);
	list->SetOrigin (0, 0);
	list->SetSize (160, 120);
	list->SetupDataBase (title);
	
	Count num   = edit->CountAttached ();
	Count index = 0;
	
	while (index < num) {
		
		OPAtomRef   event;
		OPAtomRef   actor;
		OPObjectRef observer;
		Card32      id;
		
		edit->GetAttached (index, event, actor, observer, id);
		
		OPStringRef text = "";
		
		text->AppendString (event);
		text->AppendString (" => ");
		text->AppendString (observer->ReturnRuntime ()->ReturnClassName ());
		text->AppendString ("::");
		text->AppendString (actor);
		
		list->Insert (text, text, edit, text);
		index++;
	}
	
	list->Update ();
	view = list;
}

static void
MakeAttributeList (OPViewIn parent, OPObjectIn edit, OPViewOut view)
{
	OPListRef data = OP_NEW (OPList);
	Count num   = edit->CountAttributes ();
	Count index = 0;
	
	data->SetInsertMode (FALSE);	
	
	while (index < num) {
		
		OPObjectRef key;
		OPObjectRef value;
		OPTextRef   name;
		
		edit->GetAttribute (index, key, value);
		
		OPStringRef text;
		
		if (key->IsKindOf (OP_RUNTIME (OPText))) {
			name = key;
			text = name->ReturnString ();
		} else {
			text = "[obj]";
		}
		
		data->Insert (text);
		index++;
	}
	
	OPCoord dx = 200;
	OPCoord dy = 24;
	OPRect  box;
	
	box.SetOrigin (0, 0);
	box.SetSize (dx+2, dy*5+2);
	
	OPScrollGridRef grid;
	
//	OPBuildEdit::MakeGridFromStringList (data, grid, parent, box, dx, dy);
	
	view = grid;
}

/****************************************************************/

/*
 *	Add reporter object to the specified object. This uses
 *	dynamic notifications.
 */

void
OPBuildEdit::DoAddReporter (OPObjectIn signal, OPDataReportIn reporter)
{
	OP_MEMBER (DoAddReporter, 2);
	OP_ARG (1, OPObjectIn, signal);
	OP_ARG (2, OPDataReportIn, reporter);
	
	static Bool init_needed = TRUE;
	static OPMessageRef m_AddReporter;
	
	if (init_needed) {
		
		m_AddReporter = OP_NEW_FROM_ZONE (OPaC_temp_zone, OPMessage);
		m_AddReporter->SetMethod ("AddReporter");
		
		init_needed = FALSE;
	}
	
	signal <- m_AddReporter << reporter;
}

/*
 *	Add a reporter to the current edition object. This is
 *	needed so that the reporters can be removed when the
 *	edition dialog is being closed.
 */

void
OPBuildEdit::AddReporter (OPDataReportIn reporter)
{
	OP_METHOD (AddReporter, 1);
	OP_ARG (1, OPDataReportIn, reporter);
	
	if (this->reporters.IsInvalid ()) {
		this->reporters = OP_NEW (OPHashTable);
	}
	
	this->reporters->Insert (reporter);
}


/*
 *	Prepare the object for its destruction.
 */

void
OPBuildEdit::Delete ()
{
	OP_METHOD (Delete, 0);
	
 	this->Cleanup ();
	
	super::Delete ();
}


/*
 *	Remove every temporary reporter which had been
 *	created for the edition.
 */

void
OPBuildEdit::Cleanup ()
{
	OP_METHOD (Cleanup, 0);
	
 	if (this->reporters.IsValid ()) {
		
		OPObjectRef reporter;
		
		while (this->reporters->Get (0, reporter)) {
			reporter->Delete ();
			this->reporters->Remove (0);
		}
		
		this->reporters->Delete ();
		this->reporters = 0;
	}
}

/****************************************************************/

/*
 *	Produce the view object which will be inserted into the
 *	edition grid. There are plenty of different cases to be
 *	handled here :
 *
 *	- Simple text (string) properties.
 *	- Numeric (integer or real) properties.
 *	- Boolean properties.
 *	- ...
 */

void
OPBuildEdit::EditTextProperty (OPObjectIn me, OPObjectIn edit, OPAtomIn name, OPViewOut view)
{
	OP_MEMBER (EditTextProperty, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, OPViewOut, view);
	
	OPTextLineRef   text    = OP_NEW (OPTextLine);
	OPDataReportRef data_r1 = OP_NEW (OPDataReport);
	OPDataReportRef data_r2 = OP_NEW (OPDataReport);
	
	text->SetSize (200, 24);
	
	data_r1->Initialise (edit, "Changed", name, text, "Text", "Display", FALSE);
	data_r2->Initialise (text, "Changed", "Text", edit, name, "Display", FALSE);
	
	OPBuildEdit::DoAddReporter (me, data_r1);
	OPBuildEdit::DoAddReporter (me, data_r2);
	
	OPObject::NewReportID ();
	data_r1->Changed (edit, 0);
	
	view = text;
}

void
OPBuildEdit::EditTextTagProperty (OPObjectIn me, OPObjectIn edit, OPAtomIn name, OPViewOut view)
{
	OP_METHOD (EditTextTagProperty, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, OPViewOut, view);
	
	OPTextLineRef   text    = OP_NEW (OPTextLine);
	OPDataReportRef data_r1 = OP_NEW (OPDataReport);
	OPDataReportRef data_r2 = OP_NEW (OPDataReport);
	
	text->SetSize (200, 24);
	
	OPStringRef path = name;
	path->AppendString (".Caption");
	
	data_r1->DefineQueryPath (path);
	data_r1->DefineTarget (text, "Text");
	data_r1->DefineAction ("Display");
	data_r1->SetVisible (FALSE);
	data_r2->DefineQuery ("Text");
	data_r2->DefineTargetPath (edit, path);
	data_r2->DefineAction ("Display");
	data_r2->SetVisible (FALSE);
	
	text->AttachObserver (data_r2, "UserTextChange", "Changed");
	edit->AttachObserver (data_r1, "Changed");
	
	OPBuildEdit::DoAddReporter (me, data_r1);
	OPBuildEdit::DoAddReporter (me, data_r2);
	
	OPObject::NewReportID ();
	data_r1->Changed (edit, 0);
	
	view = text;
}

void
OPBuildEdit::EditNumericPropery (OPObjectIn me, OPObjectIn edit, OPAtomIn name, OPViewOut view)
{
	OP_METHOD (EditNumericProperty, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, OPViewOut, view);
	
	OPTextLineRef   text    = OP_NEW (OPTextLine);
	OPDataReportRef data_r1 = OP_NEW (OPDataReport);
	OPDataReportRef data_r2 = OP_NEW (OPDataReport);
	
	text->SetSize (80, 24);
	
	data_r1->DefineQuery (name);
	data_r1->DefineTarget (text, "Text");
	data_r2->DefineQuery ("Text");
	data_r2->DefineTarget (edit, name);
	
	text->AttachObserver (data_r2, "UserTextChange", "Changed");
	text->AttachObserver (me,      "UserTextChange", "Update");
	
	OPBuildEdit::DoAddReporter (me, data_r1);
	OPBuildEdit::DoAddReporter (me, data_r2);
	
	OPObject::NewReportID ();
	data_r1->Changed (edit, 0);
	
	view = text;
}

void
OPBuildEdit::EditBoolProperty (OPObjectIn me, OPObjectIn edit, OPAtomIn name, OPViewOut view)
{
	OP_METHOD (EditBoolProperty, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, OPViewOut, view);
	
	OPSwitchRef     flag    = OP_NEW (OPSwitch);
	OPDataReportRef data_r1 = OP_NEW (OPDataReport);
	OPDataReportRef data_r2 = OP_NEW (OPDataReport);
	
	flag->SetSize (200, 24);
	flag->InitAsCheckBox ();
	flag->SetTitle ("yes");
	
	data_r1->DefineQuery (name);
	data_r1->DefineTarget (flag, "Value");
	data_r2->DefineQuery ("Value");
	data_r2->DefineTarget (edit, name);
	
	flag->AttachObserver (data_r2, "OnClick", "Changed");
	flag->AttachObserver (me,      "OnClick", "Update");
	
	OPBuildEdit::DoAddReporter (me, data_r1);
	OPBuildEdit::DoAddReporter (me, data_r2);
	
	OPObject::NewReportID ();
	data_r1->Changed (edit, 0);
	
	view = flag;
}


void
OPBuildEdit::EditViewProperty (OPObjectIn me, OPObjectIn edit, OPAtomIn name, OPViewOut view)
{
	OP_METHOD (EditViewProperty, 3);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, OPViewOut, view);
	
	OPObjectWellRef well    = OP_NEW (OPObjectWell);
	OPDataReportRef data_r1 = OP_NEW (OPDataReport);
	OPDataReportRef data_r2 = OP_NEW (OPDataReport);
	
	data_r1->DefineQuery (name);
	data_r1->DefineTarget (well, "Object");
	data_r2->DefineQuery ("Object");
	data_r2->DefineTarget (edit, name);
	
	well->AttachObserver (data_r2, "ObjectChange", "Changed");
	well->AttachObserver (me,      "ObjectChange", "Update");
	
	OPBuildEdit::DoAddReporter (me, data_r1);
	OPBuildEdit::DoAddReporter (me, data_r2);
	
	OPObject::NewReportID ();
	data_r1->Changed (edit, 0);
	
	OPCoord dx;
	OPCoord dy;
	
	well->GetMinSize (dx, dy);
	well->SetSize (dx, dy);
	
	view = well;
}


void
OPBuildEdit::EditBitsetProperty (OPObjectIn me, OPObjectIn edit, OPAtomIn name, Count index, OPViewOut view)
{
	OP_METHOD (EditBitsetProperty, 4);
	OP_ARG (1, OPObjectIn, edit);
	OP_ARG (2, OPAtomIn, name);
	OP_ARG (3, Count, index);
	OP_ARG (4, OPViewOut, view);
	
	const   MAX_BITSET   = 8;
	Card32  count_bitset = 0;
	OPCoord ox           = 110;
	
	const OPaC_BitsetDataDef* bitset_def[MAX_BITSET] = { 0 };
	const OPaC_BitsetDataDef* def = 0;
	Card32 bitset_num[MAX_BITSET] = { 0 };
	Card32 num;
	
	while (edit->GetPublicDataEx (index, count_bitset,
								  bitset_def[count_bitset],
								  bitset_num[count_bitset])) {
		count_bitset++;
	}
	
	if (count_bitset == 0) {
		return;
	}
	
	OPCellGridRef big_grid = OP_NEW (OPCellGrid);
	
	big_grid->SetIsometricMode (FALSE);
	big_grid->SetArraySize (1, 1);
	big_grid->SetDrawFrameHigh (FALSE);
	
	OPCoord dx;
	OPCoord dy;
	
	for (Count j = 0; j < count_bitset; j++) {
		
		def = bitset_def[j];
		num = bitset_num[j];
		
		if ( (num) && (def->type == OP_BITSET_DATA_MASK) ) {
			
			OPCellGridRef  grid = OP_NEW (OPCellGrid);
			OPBoxViewRef   box  = OP_NEW (OPBoxView);
			OPBitsetSpyRef spy  = OP_NEW (OPBitsetSpy);
			
			spy->SetRefObject (edit, index, def->value);
			
			box->SetTitle (def->name);
			
			grid->SetParent (box);
			grid->SetOrigin (4, 16);
			grid->SetArraySize (1, 1);
			grid->SetIsometricMode (FALSE);
			grid->SetDrawFrameHigh (FALSE);
			
			num--;
			def++;
			
			while ( (num)
				 && ( (def->type == OP_BITSET_DATA_VALUE)
				   || (def->type == OP_BITSET_DATA_BIT) ) ) {
				
				OPSwitchRef flag = OP_NEW (OPSwitch);
				
				if (def->type == OP_BITSET_DATA_VALUE) {
					flag->InitAsRadioButton ();
					spy->AddSwitch (flag, def->value);
				} else {
					flag->InitAsCheckBox ();
					spy->AddSwitchBit (flag, def->value);
				}
				
				flag->SetTitle (def->name);
				grid->InsertCellInColumn (flag);
				
				num--;
				def++;
			}
			
			grid->SetMargin (2);
			grid->UpdateGridGeometry ();
			grid->GetSize (dx, dy);
			
			dx += 2*4;
			dy += 4;
			
			box->SetSize (dx, dy);
			box->SetDesiredMinSize (dx, dy);
			
			big_grid->InsertCellInLine (box);
		}
	}
	
	big_grid->SetMargin (2);
	big_grid->UpdateGridGeometry ();
	big_grid->GetSize (dx, dy);
	
	view = big_grid;
}



/*
 *
 */

Bool
OPBuildEdit::EditProperties (OPObjectIn edit)
{
	OP_METHOD (EditProperties, 1);
	OP_ARG (1, OPObjectIn, edit);
	
	Count num   = edit->CountPublicData ();
	Count iter  = 0;
	Count total = 0;
	
	//	Find how many properties can really be used for the dynamic
	//	edition of the specified object.
	
	for (iter = 0; iter < num; iter++) {
		OPAtomRef name;
		OPAtomRef notifier;
		OPStringRef desc;
		OPTypeDef type;
		Bool is_read_ok;
		Bool is_write_ok;
		edit->GetPublicData (iter, name, desc, type, is_read_ok, is_write_ok);
		if (is_read_ok && is_write_ok) total++;
	}
	
	if (total == 0) {
		return FALSE;
	}
	
	OPCellGridRef grid = OP_NEW (OPCellGrid);
	
	grid->SetIsometricMode (FALSE);
	grid->SetArraySize (2, total);
	
	//	Handle each public data, one after the other.
	
	for (iter = 0; iter < num; iter++) {
		
		OPObjectRef  data;					//	data (object) of the field
		OPAtomRef    name;					//	name of public data field
		OPStringRef  desc;					//	description of attribute
		Bool         is_read_ok;			//	TRUE => value can be read
		Bool         is_write_ok;			//	TRUE => value can be modified
		OPTypeDef    type = 0;				//	type description
		OPSimpleType what = OP_TYPE_NONE;	//	type description (if it is a simple type)
		
		edit->GetPublicData (iter, name, desc, type, is_read_ok, is_write_ok);
		
		if (is_read_ok == FALSE) continue;
		if (is_write_ok == FALSE) continue;
		
		what = FindSimpleType (type, edit, iter);
		
		OPViewRef    view    = OPViewNull;
		OPTextTagRef caption = OP_NEW (OPTextTag);
		OPBoxViewRef boxview = OP_NEW (OPBoxView);
		
		switch (what) {
			case OP_TYPE_TEXT:
				EditTextProperty (this, edit, name, view);
				break;
			case OP_TYPE_O_TEXTTAG:
				EditTextTagProperty (this, edit, name, view);
				break;
			case OP_TYPE_BOOL:
				EditBoolProperty (this, edit, name, view);
				break;
			case OP_TYPE_INT:
			case OP_TYPE_CARD:
			case OP_TYPE_REAL:
				EditNumericPropery (this, edit, name, view);
				break;
			case OP_TYPE_O_VIEW:
				EditViewProperty (this, edit, name, view);
				break;
			case OP_TYPE_BITSET:
				EditBitsetProperty (this, edit, name, iter, view);
				break;
		}
		
		desc->AppendString (" :  ");
		
		caption->SetTitle (desc);
		caption->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
		
		OPCoord dx = 0;
		OPCoord dy = 24;
		
		if (view.IsValid ()) {
			view->SetParent (boxview);
			view->SetOrigin (0, 0);
			view->GetSize (dx, dy);
		}
		
		boxview->EnableFrame (FALSE);
		boxview->EnableTitle (FALSE);
		boxview->SetDesiredMinSize (dx, dy);
		
		grid->InsertCellInColumn (caption, 0);
		grid->InsertCellInColumn (boxview, 1);
	}
	
	grid->SetMargin (2);
	grid->UpdateGridGeometry ();
	
	OPButtonRef     button = OP_NEW (OPButton);
	OPWindowRef     window = OP_NEW (OPWindow);
	OPScrollGridRef scroll = OP_NEW (OPScrollGrid);
	
	button->Initialise (10, 265, 100, 24, "OK", window);
	button->AttachObserver (this, "OnClick", "EditOK");
	
	window->SetOrigin (420, 20);
	window->SetSize (420, 300);
	window->SetStyle (OP_WIN_TITLE | OP_WIN_MOVE_OK);
	window->SetTitle ("Object Inspector");
	
	OPTabViewRef  tab_view = OP_NEW (OPTabView);
	OPTabListRef  tab_list = OPTabListNull;
	OPListRef     list = OP_NEW (OPList);
	OPTextTagRef  text;
	
	tab_view->SetParent (window);
	tab_view->SetOrigin (10, 10);
	tab_view->SetSize (400, 250);
	tab_view->SetActiveTab (1);
	tab_view->SetTabNum (5);
	
	list->SetInsertMode (FALSE);
	
	text = OP_NEW (OPTextTag);	text->SetText ("Properties");	list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Events");		list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Targets");		list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Fields");		list->Insert (text);
	text = OP_NEW (OPTextTag);	text->SetText ("Attributes");	list->Insert (text);
	
	tab_view->GetTabs (tab_list);
	tab_list->SetTabs (list);
	tab_view->GetViews (list);
	list->SetInsertMode (FALSE);

	OPRect box (10, 10, 380, 210);
	
	scroll->SetParent (window);
	scroll->Initialise (grid, box);
	scroll->SetParent (OPViewNull);
	list->Insert (scroll);
	
	OPViewRef view;
	
	OPBuildConViewerRef cview = OP_NEW (OPBuildConViewer);
	
	cview->SetParent (window);
	cview->SetSize (396, 180);
	cview->Initialise (edit, "Published events", "Attached actions");
	cview->SetParent (OPViewNull);
	list->Insert (cview);
	
	OPDBScrollListRef dbsl = OP_NEW (OPDBScrollList);
	
	dbsl->SetParent (window);
	dbsl->SetSize (188, 120);
	dbsl->SetupDataBase ("Published targets");
	dbsl->Update ();
	MakeTargetList (dbsl, edit);
	dbsl->SetParent (OPViewNull);
	list->Insert (dbsl);
	
	OPDataLinkViewerRef dview = OP_NEW (OPDataLinkViewer);
	
	dview->SetParent (window);
	dview->SetSize (396, 180);
	dview->Initialise (edit, "Published fields", "Attached fields");
	dview->SetParent (OPViewNull);
	list->Insert (dview);
	
	window->Open ();
	window->Display ();
	window->ShowNormal ();
	
	this->object = edit;
	this->window = window;
	
	return TRUE;
}

/****************************************************************/

Bool
OPBuildEdit::Update ()
{
	OP_METHOD (Update, 0);
	
	if ( (this->object.IsValid ())
	  && (this->object->IsKindOf (OP_RUNTIME (OPView))) ) {
		OPViewRef view = this->object;
		view->Display ();
	}
	
	return TRUE;
}

Bool
OPBuildEdit::EditOK ()
{
	OP_METHOD (EditOK, 0);
	
	OPBuildEditRef self = this;
	
	this->window->Hide ();
	this->window->Close ();
	this->window->Delete ();
	this->window = OPWindowNull;
	
	this->Delete ();
	
	return TRUE;
}

